/*
 itable.c
  inverse table generation

   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
*/

#include "mgraf.h"

#include <string.h>

/* ---- code */

#define ITAB_CMPRES			8
#define GI_RES				6
#define GI_ONE				(1<<GI_RES)
#define GI_MASK				(GI_ONE-1)
#define GI_TOLERANCE			(GI_ONE/4)
#define MAX_GI_LENGTH			(GI_TOLERANCE*3>>1)

typedef struct tc_entry/*(8)*/
{
	/* 'i' features two half/quarter-error bits (for color correction)
	  bit 14: dim green (postive error)
	  bit 15: gray (for matching ~zero/non-zero saturation)
	*/
	int16		r, g, b, i;
} tc_entry;

/* over 2KB in size */
typedef struct tc_table
{
	tc_entry	t[256];
	int16		c;
	uint8		g[(GI_ONE+MAX_GI_LENGTH+11)&~3]; /* gray intensity seek table */
} tc_table;

/* over 6KB in size */
typedef struct tc_split
{
	tc_table	r, g, b;
} tc_split;

static void sort_tc_table(tc_table *t,
	int (f)(const void *a,
		const void *b))
{
	qsort(t->t,t->c,sizeof(tc_entry),f);

	{ /* gray intensity seek table with clipping */
		const tc_entry	*a = t->t;
		uint8		*w = t->g+GI_TOLERANCE; /* already zeroed w. calloc */
		int		i, j = 1, v, c;

		i = 0;
		do
			if ((c=(v=((const uint8*)&a->i)[!BIG_ENDIAN]&GI_MASK)-j)>0)
			{
				j = v;
				do
					*w++ = i;
				while(--c);
			}
		while(++a,++i<255);
		if ((c=-(int)w&3)!=0)
			do
				*w++ = 0xFE;
			while(--c);
		if ((c=sizeof(t->g)+t->g-w>>2)!=0)
			do
				*((int32*)w)++ = 0xFEFEFEFEu;
			while(--c);
	}
}

static int compare8_gray(const void *a,
	const void *b)
{ /* invert gray intensity so index has precedence, adding last-most black first */
	return ((((const tc_entry*)b)->i&GI_MASK<<8)^0xFF00)
		-((((const tc_entry*)a)->i&GI_MASK<<8)^0xFF00);
}

static void fillGrayTable(uint8 *d,
	const uint16 *s,
	int c,
	int16 *t)
{
	int16		*a;
	int		x, i, j;
	/* first, fill the remap */
	x = 0;
	do
		if (*(a=t+(s[x]>>6))<0) /* don't overwrite */
			*a = x;
	while(++x<c);
	/* second, bridge the gaps, rounding */
	x = j = 0;
	do
		if ((i=t[++x])>=0)
		{
			c = (x-j+1)>>1;
			memset(d+j,t[j],c);
			j+= c;
			memset(d+j,i,x-j);
			j = x;
		}
	while(x<1023);
	/* the rest is white, already filled by the caller */
}

#define MAX_SEPARATION_THRESHOLD	0xA0A

int makeITables(const cTab64 *ct,
	uint8 **itab,
	int bpp,
	int r,
	boolean forceColor)
{
	tc_entry	cya;
	tc_split	*ta;
	int		count;
	boolean		isColor;

	if (r<3)
		r = 3;
	else if (r>5) /* 8 = 16MB */
		r = 5; /* can't handle 6 */
	if ((ta=calloc(1,sizeof(*ta)))==NULL)
		return -2;
	isColor = 0, count = 0, cya.i = 0;
	if (ct && (count=MIN(ct->count,256))!=0)
	{ /* build the RGB tables */
		const cSpec	*s = ct->colors;
		t_count		i = 0, cyr = IMAX(t_count);

		do
		{
			tc_entry	a;
			uint32		vr, vg, vb, v, g;

			vr = s->rgb.red, v = vr>>8;
			vg = s->rgb.green, vb = s->rgb.blue;
			/* ugly brown workaround to 15bpp red/green oscillation */
			/* additional red checks fix some game colors... a bit hacky */
			if (vr>vg && vg>vb && v<=106 && v!=40 && v!=63 && v!=95 && v!=90)
			{
				a.r = (vr-(vr>>r>>3))>>16-ITAB_CMPRES;
				a.g = (vg-(vg>>r>>3))>>16-ITAB_CMPRES;
				a.b = (vb-(vb>>r>>3))>>16-ITAB_CMPRES;
			} else
			{
				a.r = (vr-(vr>>r))>>16-ITAB_CMPRES;
				a.g = (vg-(vg>>r))>>16-ITAB_CMPRES;
				a.b = (vb-(vb>>r))>>16-ITAB_CMPRES;
			}
			v = (vr+vg+vg)<<2, g = vr+v, g+= vg, g+= vb, g+= vb; /* 5+9+2>>4 */
			/* gray intensity - make sure black & white gis are true */
			if ((v=g>>(4+8-GI_RES)&GI_MASK<<8)==0)
			{
				if (g)
					v = 1<<8;
			} else if (v==(GI_MASK<<8) && g!=0xFFFF0)
				v = (GI_ONE-2)<<8;
			v|= ct->flags&0x8000 ? i:s->value;
			if ((a.r^a.g)|(a.g^a.b))
				isColor = 1;
			else
				v|= 0x8000;
			if (a.g>a.r && a.g>a.b && a.g<=(a.r>a.b ? 3<<(ITAB_CMPRES-3)
					:1<<(ITAB_CMPRES-2)))
				v|= 0x4000;
			a.i = v;

			if (vr+MAX_SEPARATION_THRESHOLD>vg && vr+MAX_SEPARATION_THRESHOLD>vb)
				ta->r.t[ta->r.c++] = a;
			if (vg+MAX_SEPARATION_THRESHOLD>vr && vg+MAX_SEPARATION_THRESHOLD>vb)
				ta->g.t[ta->g.c++] = a;
			if (vb+MAX_SEPARATION_THRESHOLD>vr && vb+MAX_SEPARATION_THRESHOLD>vg)
				ta->b.t[ta->b.c++] = a;
			g = (2<<ITAB_CMPRES)-a.b-a.g+a.r;
			if (g<cyr)
				cyr = g, cya = a;
		} while(++s,++i<count);
	} else
	{
		memset(ta,0,sizeof(*ta));
		count = 1u<<bpp;
		ct = (cTab64*)&ta->g;
		((cTab64*)&ta->g)->flags = 0x8000, ((cTab64*)&ta->g)->count = count;
		ctMakeGrays(((cTab64*)&ta->g)->colors,count);
		((uint8*)&ta->r.t->i)[BIG_ENDIAN] = count-1;
	}
	sort_tc_table(&ta->r,compare8_gray);
	if (!isColor && !forceColor)
	{ /* gray itable */
		uint16		*d;
		const cSpec	*s;
		uint32		u;
		t_count		v;

		if ((d=realloc(*itab,1024+(ct->flags&0x8000 ? count*2:256*2)))==NULL)
			{free(ta); return -2;}
		*itab = (uint8*)d;
		memset(d,((const uint8*)&ta->r.t[count-1].i)[BIG_ENDIAN],1024);
		/* direct table */
		d+= 512, s = ct->colors, v = count;
		if (ct->flags&0x8000)
			do
				*d++ = formGray(s->rgb.red,s->rgb.green,s->rgb.blue,u)>>1;
			while(++s,--v);
		else
			do
				d[s->value] = formGray(s->rgb.red,s->rgb.green,s->rgb.blue,u)>>1;
			while(++s,--v);
		v = ((const uint8*)&ta->r.t->i)[BIG_ENDIAN]; /* closest black entry */
		/* ta, used as scratch, needs to be >4KB */
		memset(ta,-1,2048);
		*(int16*)ta = v;
		fillGrayTable(*itab,(const uint16*)*itab+512,count,(int16*)ta);
		free(ta);
		return 0;
	}
	if (ta->g.c<256)
	{ /* as green has precedence over blue, make sure the cyan is in the green table */
		t_count		i;

		for(i=0; i<ta->g.c; ++i)
			if (ta->g.t[i].i==cya.i)
				goto ao;
		ta->g.t[ta->g.c++] = cya;
	}
ao:	sort_tc_table(&ta->g,compare8_gray);
	sort_tc_table(&ta->b,compare8_gray);

	/* This is slower than Apple's MakeITable, but it produces better results.
	  MakeITable which doesn't use full table scans but a completion method
	  with the separation table and a temporary 16-bit fill table.
	*/
	{ /* color itable */
		uint8		*d;
		unsigned	vr, vg, vb, cn, lm;

		if ((d=realloc(*itab,(1u<<r<<r<<r)
				+(ct->flags&0x8000 ? 512:count*2)))==NULL)
			{free(ta); return -2;}
		*itab = d, *d++ = (uint8)ta->r.t->i;
		vb = cn = 1u<<ITAB_CMPRES>>r, lm = vg = vr = 0;
		goto loop;
		do
		{
			vg = 0;
			do
			{
				lm = (((vr+vg+vg)<<2)+vg+vr)>>1, vb = 0;
				do
				{
					tc_table	*t;
					const tc_entry	*a, *e;
					t_count		i, y, z, gf;

				loop:	z = (lm+vb)>>(3+ITAB_CMPRES-GI_RES); /* gi index */
					/* it's best the leave this order alone */
					if (vr>=vg && vr>=vb)
						t = &ta->r;
					else if (vg>=vr && vg>=vb)
						t = &ta->g;
					else
						t = &ta->b;
					a = t->t+t->g[z], e = t->t+t->g[z+MAX_GI_LENGTH]; /* range */
					/* some counter-bias precalculations */
					gf = (vr-vg)|(vg-vb) ? 0:-0x8000;

					z = IMAX(z), i = 0;
					do
					{
						int16		x;

						if ((x=(t_count)vb-a->b)<0)
							x =-x, ++x;
						if (x>=z)
							continue;
						if ((y=(t_count)vr-a->r)<0)
							y =-y, y+= 2;
						x+= y;
						if ((y=(t_count)vg-a->g)<0)
						{
							y =-y, y+= 2;
							if ((t_count)(a->i&0x4000))
								y+= y>>1;
						}
						x+= y;
						if (x>=z)
							continue;
						if ((t_count)((gf^a->i)&0x8000) && vg>9
								&& !(t_count)(a->i&0x4000))
						{ /* should be a color/gray */
							x+= 2, x+= x>>2;
							if (x>=z)
								continue;
						}
						i = a->i;
						if ((z=x)==0)
							break;
					} while(++a<=e);
					*d++ = i;
				} while(vb+=cn,vb&=(1<<ITAB_CMPRES)-1);
			} while(vg+=cn,vg&=(1<<ITAB_CMPRES)-1);
		} while(vr+=cn,vr&=(1<<ITAB_CMPRES)-1);
		
		{ /* direct table */
			const cSpec	*s = ct->colors;

			if (ct->flags&0x8000)
			{
				do
					*((uint16*)d)++ = formDirect(&s->rgb,r,r);
				while(++s,--count);
			} else
			{
				memset(d,0,512);
				do
					((uint16*)d)[s->value] = formDirect(&s->rgb,r,r);
				while(++s,--count);
			}
		}
	}
	free(ta);
	return r;
}
