/*
 mcnv.c
  pixel depth, endian and scaled conversion

   Created: 2014.04.22
   Revised: 2014.08.15

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.23
  Split from mblt.
2014.04.26
  Finished mcnv scaling.
2014.05.06
  Fixed dithering (mostly), split into RGB and Gray.
2014.08.15
  calloc instead of malloc appears to fix direct dithering.
*/

#include <string.h>
#include "mgraf.h"
#include "mblt.h"

/* ---- constants */

#define directSrc	8
#define directBSwap	1
#define direct32b	2
#define directDst	4
#define directScaled	20
#define indexedScaled	4
/* non-direct dest. */
#define dstGray		4
#define dstInverse	8

/* ---- typedefs */

/* replaces 'cache'/'hx'/'rx'; this is also 40 bytes at most */
typedef struct ditherParam/*(24+)*/
{
	int32		r, g, b;
	int		i;
	long int	gt;
	rgbRoutine	cc;
	rgb48		*a;
} ditherParam;

#define DITHERVAR(f)	((ditherParam*)f)

/* ---- macros */

#define rfdi()		(void)(*d=s[u>>16],u+=f->rx)
#define rxdi(x)		{i=u>>16; if(i>=mp) goto end; d[x]=s[i],u+=f->rx;}
#define rfdx(h)		{j=u>>16,*d=s[j>>h]>>(~j<<3-h&7),u+=f->rx;}
#define rxdx(h,x)	{j=u>>16,i=j>>h; if(i>=mp) goto end; d[x]=s[i]>>(~j<<3-h&7),u+=f->rx;}

/* ---- code */

void *rGenRemappingTable(pixMap *m,
	const cTab64 *ct,
	uint8 bpp,
	uint8 flag,
	boolean is32b)
{
	rgbRoutine	crtn;
	uint32		*t, *d, i, cm, rb;
	int		c;
	long int	gt;

	if ((t=calloc(is32b ? 4:2,ct && ct->flags&0x8000 ? MAX(ct->count,256):256))==NULL)
		return NULL;
	d = t;//+c;
	if (flag&remapForce30)
		rb = gt = 10, i = 31, c = ccDirect; //(c|ccDirect)&~ccIndexed;
	else
	{
		if ((c=pmGetRGBCC(m,&gt))<0)
			return NULL;
		rb = m->minCmp ? m->minCmp:10;
		if (flag&c&ccIndexed) /* indirect table supression (arithmetic modes) */
			c&=~ccNative, i = rb;
		else
			i = m->bpp;
	}
	cm =-(flag&1)&((2u<<(i-1))-1);
	crtn = rgbCCRoutine[c], i = rb;
	if (ct && (c=ct->count)!=0)
	{
		const cSpec	*cs = ct->colors;

		if (ct->flags&0x8000)
			if (flag&remapReverse)
				if (is32b)
				{
					d+= c;
					do
						*--d = crtn(i,gt,&cs->rgb)^cm;
					while(++cs,--c);
				} else
				{
					((uint16*)d)+= c;
					do
						--((uint16*)d), *(uint16*)d = crtn(i,gt,&cs->rgb)^cm;
					while(++cs,--c);
				}
			else if (is32b)
				do
					*d = crtn(i,gt,&cs->rgb)^cm;
				while(++d,++cs,--c);
			else
				do
					*((uint16*)d)++ = crtn(i,gt,&cs->rgb)^cm;
				while(++cs,--c);
		else if (is32b)
			do
				d[cs->value] = crtn(i,gt,&cs->rgb)^cm;
			while(++cs,--c);
		else
			do
				((uint16*)d)[cs->value] = crtn(i,gt,&cs->rgb)^cm;
			while(++cs,--c);
	} else
	{
		rgb48		cl;
		int		dx, x;

		c = 1u<<bpp, dx = (65535u<<16)/(c-1), x =~0u;
		if (is32b)
			do
			{
				cl.red = cl.green = cl.blue = x>>16;
				*d = crtn(i,gt,&cl)^cm;
			} while(++d,x-=dx,--c);
		else
			do
			{
				cl.red = cl.green = cl.blue = x>>16;
				*((uint16*)d)++ = (uint16)(crtn(i,gt,&cl)^cm);
			} while(x-=dx,--c);
	}
	return t;
}

/* ---- direct conversion */

/* maybe these optimized converters could be used...
 nah - they would slow down setup further, and'd require more code than this
*/

/*uint32 i15to24(uint16 v)
{
	return (uint8)(v<<3|v>>2)|(v<<6&0xF800)|(v<<1&0x700)
		|(v<<9&0xF80000)|(v<<4&0x70000);
}
uint16 i24to15(uint32 v)
{
	v>>= 3;
	return (v&0x1F)|(v>>3&0x3E0)|(v>>6&0x7C00);
}
*/

static DEF_CNV_RTN(pxroDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return cnvDirect(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroBDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return cnvDirect(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return cnvDirect(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroBDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return cnvDirect(v,u,t,r)^f->cm;
}

/* ---- direct through gray */

static DEF_CNV_RTN(pxroGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return cnvGray(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroBGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return cnvGray(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return cnvGray(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroBGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return cnvGray(v,u,t,r)>>1^f->cm;
}

/* ---- direct through itable */

static DEF_CNV_RTN(pxroIDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIBDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIBDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

/* ---- direct through gray itable */

static DEF_CNV_RTN(pxroIGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIBGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIBGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

/* ---- same-depth direct */

static DEF_CNV_RTN(pxro32)
	{return ((const uint32*)f->s)[o]^f->cm;}

static DEF_CNV_RTN(pxroB32)
{
	uint32		v = ((const uint32*)f->s)[o];

	return _reverse32(v)^f->cm;
}

static DEF_CNV_RTN(pxro16)
	{return ((const uint16*)f->s)[o]^f->cm;}

static DEF_CNV_RTN(pxroB16)
{
	uint16		v = ((const uint16*)f->s)[o];

	return _reverse16(v)^f->cm;
}

/* ---- indexed through table */

/* this one's usually overridden with the pattern method */
static DEF_CNV_RTN(pxro8)
	{return ((uint32*)f->t)[((const uint8*)f->s)[o]];}

static DEF_CNV_RTN(pxroCache4)
{
	int		u = o>>2, v;

	if (u!=f->hx)
	{ /* 2-byte alignment */
		const uint8	*s = (const uint8*)f->s+u*2;

		f->hx = u;
		v = *s;
		*f->cache = v>>4;
		f->cache[1] = v&15;
		v = s[1];
		f->cache[2] = v>>4;
		f->cache[3] = v&15;
	}
	return ((uint32*)f->t)[f->cache[o&3]];
}

static DEF_CNV_RTN(pxroCache2)
{
	int		u = o>>2;

	if (u!=f->hx)
	{
		f->hx = u, u = ((const uint8*)f->s)[u];
		*f->cache = u>>6;
		f->cache[1] = u>>4&3;
		f->cache[2] = u>>2&3;
		f->cache[3] = u&3;
	}
	return ((uint32*)f->t)[f->cache[o&3]];
}

static DEF_CNV_RTN(pxroCache1)
{
	int		u = o>>3;

	if (u!=f->hx)
	{
		f->hx = u, u = ((const uint8*)f->s)[u];
		*f->cache = u>>7;
		f->cache[1] = u>>6&1;
		f->cache[2] = u>>5&1;
		f->cache[3] = u>>4&1;
		f->cache[4] = u>>3&1;
		f->cache[5] = u>>2&1;
		f->cache[6] = u>>1&1;
		f->cache[7] = u&1;
	}
	return ((uint32*)f->t)[f->cache[o&7]];
}

/* ---- scaled */

/* unlike QD, scaling down here doesn't merge pixels */
/* for that, the furthest index for indexed modes; interpolate-integral for direct */

#define hscali(typ,h)	\
	uint32 u=o>>3, i; \
	o&=7; \
	if(u!=f->hx) { \
		const typ *s=f->s; int32 *d=f->cache, mp=f->mp>>h; \
		f->hx=u, u=u*8*f->rx+(f->rx>>1); \
		rfdi(); \
		rxdi(1) \
		rxdi(2) \
		rxdi(3) \
		rxdi(4) \
		rxdi(5) \
		rxdi(6) \
		rxdi(7) \
	}
#define hscalx(h)	\
	uint32 u=o>>3, i, j; \
	o&=7; \
	if(u!=f->hx) { \
		const uint8 *s=f->s; int32 *d=f->cache, mp=f->mp; \
		f->hx=u, u=u*8*f->rx+(f->rx>>1); \
		rfdx(h); \
		rxdx(h,1) \
		rxdx(h,2) \
		rxdx(h,3) \
		rxdx(h,4) \
		rxdx(h,5) \
		rxdx(h,6) \
		rxdx(h,7) \
	}

static DEF_CNV_RTN(pxrsCache32)
{
	hscali(uint32,2);
end:	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsBCache32)
{
	hscali(uint32,2);
end:	return _reverse32(f->cache[o])^f->cm;
}

static DEF_CNV_RTN(pxrsCache16)
{
	hscali(uint16,1);
end:	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsBCache16)
{
	hscali(uint16,1);
end:	return _reverse16(f->cache[o])^f->cm;
}

/* ---- scaled direct conversion */

static DEF_CNV_RTN(pxrsDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = f->cache[o];
	return cnvDirect(i,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsBDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = _reverse32(f->cache[o]);
	return cnvDirect(i,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = f->cache[o];
	return cnvDirect(i,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsBDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = _reverse16(f->cache[o]);
	return cnvDirect(i,u,t,r)^f->cm;
}

/* ---- scaled direct through gray */

static DEF_CNV_RTN(pxrsGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = f->cache[o];
	return cnvGray(i,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsBGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = _reverse32(f->cache[o]);
	return cnvGray(i,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = f->cache[o];
	return cnvGray(i,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsBGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = _reverse16(f->cache[o]);
	return cnvGray(i,u,t,r)>>1^f->cm;
}

/* ---- direct through itable */

static DEF_CNV_RTN(pxrsIDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = f->cache[o];
	return f->itb[cnvDirect(i,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIBDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = _reverse32(f->cache[o]);
	return f->itb[cnvDirect(i,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = f->cache[o];
	return f->itb[cnvDirect(i,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIBDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = _reverse16(f->cache[o]);
	return f->itb[cnvDirect(i,u,t,r)^f->cm];
}

/* ---- scaled direct through gray itable */

static DEF_CNV_RTN(pxrsIGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = f->cache[o];
	return f->itb[cnvGray(i,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIBGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32,2);
end:	i = _reverse32(f->cache[o]);
	return f->itb[cnvGray(i,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = f->cache[o];
	return f->itb[cnvGray(i,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIBGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16,1);
end:	i = _reverse16(f->cache[o]);
	return f->itb[cnvGray(i,u,t,r)>>7^f->cm];
}

/* ---- scaled indexed */

static DEF_CNV_RTN(pxrsCache8)
{
	hscali(uint8,0);
end:	return ((uint32*)f->t)[f->cache[o]];
}

static DEF_CNV_RTN(pxrsCache4)
{
	hscalx(1);
end:	return ((uint32*)f->t)[f->cache[o]&15];
}

static DEF_CNV_RTN(pxrsCache2)
{
	hscalx(2);
end:	return ((uint32*)f->t)[f->cache[o]&3];
}

static DEF_CNV_RTN(pxrsCache1)
{
	hscalx(3);
end:	return ((uint32*)f->t)[f->cache[o]&1];
}

/* ---- dithering (still in development) */

/*
 Robert W. Floyd/Louis Steinberg dithering:
	actual = p[0]
	closest = closest_match(actual)
	p[0] = closest
	error = actual-closest
	p[ 1][0]+= error*7/16
	p[-1][1]+= error*3/16
	p[ 0][1]+= error*5/16
	p[ 1][1]+= error*1/16

 Crafting it here:
   Take the src/dst lines, convert to high precision (30-bit).
   All graphics ops are done on that buffer line.
   Dither with the above method.
   However, calculating differences for the next line requires extra bits;
  the colors are converted into signed pseudo-rgb48 colors.

   The modern approach alternates in direction (serpentine scanning),
  which is a bit too complex the way the implementation here is driven, expected.
*/

static DEF_CNV_RTN(pxroRGBDither)
{ /* convert to dst. pixel, then back to 30-bit for subtraction */
  /* already 10 per-channel; expanding won't be necessary for rgbRoutine;
    shouldn't be dithering to 10:11:10 dest. anyway
  */
	rgb48		r, *a;
	rgb30		v = ((const rgb30*)f->s)[o];
	uint32		u, x;
	int32		dr = v>>20&1023, dg = v>>10&1023, db = v&1023;
	/* linear perception of the channels only applies when they're isolated;
	  ease up on positive green error when green's the dimmest;
	  otherwise, the pattern will often be brighter than correct
	*/
	u = (dg*16+DITHERVAR(f)->g)<<2;
	if (dr>dg*2 || db>dg*2 && (int)u>1)
		u-= (int)u>>5;
	if (u>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	/* convert to destination pixel, pinning the components */
	dg = (int)u>>6;
	r.green = (uint16)u;
	if ((u=(dr*16+DITHERVAR(f)->r)<<2)>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	dr = (int)u>>6;
	r.red = (uint16)u;
	if ((u=(db*16+DITHERVAR(f)->b)<<2)>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	db = (int)u>>6;
	r.blue = (uint16)u;
	x = DITHERVAR(f)->cc(DITHERVAR(f)->i,DITHERVAR(f)->gt,&r);

	{ /* convert back */
		uint32		*t = f->t, b;

		if (f->cc-directSrc+0u<dstInverse)
			u = cnvDirect(x,u,t,b); /* rgbRoutines don't bswap */
		else
			u = t[x];
	}

	/* get the 'next line' of differences */
	a = DITHERVAR(f)->a+o;

	/* can't do any merging tricks (red/blue in one) due to the multiplication, high prec. */
	/* error = actual-closest */
	dr-= u>>20&1023;
	dg-= u>>10&1023;
	db-= u&1023;
	/* p[ 1][0]+= error*7/16+p[1][1] */
	DITHERVAR(f)->r = dr*7+(int16)a[2].red;
	DITHERVAR(f)->g = dg*7+(int16)a[2].green;
	DITHERVAR(f)->b = db*7+(int16)a[2].blue;

	a->red+= dr*3, a->green+= dg*3, a->blue+= db*3; 	/* p[-1][1]+= error*3/16 */
	a[1].red+= dr*5, a[1].green+= dg*5, a[1].blue+= db*5;	/* p[ 0][1]+= error*5/16 */
	a[2].red = dr, a[2].green = dg, a[2].blue = db;		/* p[ 1][1]+= error*1/16 */
	return x;
}

static DEF_CNV_RTN(pxroGrayDither)
{
	rgb30		v = ((const rgb30*)f->s)[o];
	uint32		u, x;
	int32		*a, dr = v>>20&1023, dg = v>>10&1023, db = v&1023;
	//(((dr*306+dg*601+db*117)>>6)
	if ((dg=((int)formGray(dr,dg,db,u)*8+DITHERVAR(f)->g)/16)>0x3FFu)
		dg = (int)~dg>>31&0x3FF;
	x = ((const uint8*)DITHERVAR(f)->gt)[dg];
	u = ((const uint32*)f->t)[x];

	/* get the 'next line' of differences */
	a = (int32*)DITHERVAR(f)->a+o;

	dg-= u&1023;	/* error = actual-closest */
	DITHERVAR(f)->g = dg*7+a[2]; /* p[ 1][0]+= error*7/16+p[1][1] */
	*a+= dg*3; 	/* p[-1][1]+= error*3/16 */
	a[1]+= dg*5;	/* p[ 0][1]+= error*5/16 */
	a[2] = dg;	/* p[ 1][1]+= error*1/16 */
	return x;
}

mlvParam *rInitDithering(pixMap *dm)
{
	static pixMap	m30 =
		{1, 1, {0, 0}, 4, 32, pTypeLSB|pTypeDirect, 30, 10};
	mlvParam	*f;
	long int	gt;
	int		cc, u;

	if (!dm)
		return NULL;
	u = sizeof(mlvParam)+(dm->w+2)*(sizeof(rgb30)+sizeof(rgb48));
	if ((cc=pmGetRGBCC(dm,&gt))<0
			|| (f=calloc(u,1))==NULL)
		return NULL;
	/* the easy way to get the dest. pixels */
	u = rInitCnv(&f->d2m,&m30,dm->pTable,dm->pSize<<8|(dm->minCmp ? dm->minCmp:dm->bpp)
		|(dm->pType&(pTypeLSB|pTypeLGreen)),0x10000,mSrcPat*2,0);
	if (u<0)
		goto fail;
	if (rInitCnv(&f->m2d,dm,NULL,0x200A|!BIG_ENDIAN<<pTypeLSBBit,0x10000,mSrcPat*2,0)<0)
	{
		rEndCnv(&f->d2m);
	fail:	free(f);
		return NULL;
	}
	DITHERVAR(&f->m2d)->i = dm->minCmp;
	DITHERVAR(&f->m2d)->gt = gt;
	DITHERVAR(&f->m2d)->cc = rgbCCRoutine[cc];
	f->m2d.rtn = dm->minCmp ? pxroRGBDither:pxroGrayDither;
	f->m2d.s = (void*)(f+1);
	DITHERVAR(&f->m2d)->a = (rgb48*)((rgb30*)(f+1)+dm->w);
	return f;
}

/* ---- setup */

static cnvRoutine	f_cnvx[] =
{
	pxroCache1, pxroCache2, pxroCache4, pxro8,
	pxrsCache1, pxrsCache2, pxrsCache4, pxrsCache8,

	pxro16, pxroB16, pxro32, pxroB32,
	pxroDirect16, pxroBDirect16, pxroDirect32, pxroBDirect32,
	pxroGray16, pxroBGray16, pxroGray32, pxroBGray32,
	pxroIDirect16, pxroIBDirect16, pxroIDirect32, pxroIBDirect32,
	pxroIGray16, pxroIBGray16, pxroIGray32, pxroIBGray32,

	pxrsCache16, pxrsBCache16, pxrsCache32, pxrsBCache32,
	pxrsDirect16, pxrsBDirect16, pxrsDirect32, pxrsBDirect32,
	pxrsGray16, pxrsBGray16, pxrsGray32, pxrsBGray32,
	pxrsIDirect16, pxrsIBDirect16, pxrsIDirect32, pxrsIBDirect32,
	pxrsIGray16, pxrsIBGray16, pxrsIGray32, pxrsIBGray32
};

static boolean equalColors(const cTab64 *cta,
	const cTab64 *ctb)
{
	return cta->flags&ctb->flags&0x8000 && cta->count==ctb->count
		&& !memcmp(cta->colors,ctb->colors,ctb->count);
}

int rInitCnv(cnvParam *f,
	pixMap *m,
	const cTab64 *ct,
	int fmt,
	ufixed scale,
	uint8 mode,
	uint8 flag)
{ /* depth conversion or remap needed */
	cnvParam	*cf;
	uint32		*t;
	int		cc, i, j;

	f->rtn = NULL;
	if ((uint16)fmt<0x900)
	{ /* make table */
		int		c, pmode;
		boolean		is32b;

		is32b = 1;
		if ((uint8)fmt>4 && scale==0x10000)
			is32b = !(m->pSize&31) || flag&remapForce30, mode-= mSrcPat;
		cc = pDepthX[fmt>>8&31];
		/* equal set? */
		if (m->bpp==(fmt>>8&0xFF) && m->pTable && (pmode=mode&15)<mTransparent
				&& equalColors(m->pTable,ct))
			goto nix;
		if ((t=rGenRemappingTable(m,ct,fmt&31,flag,is32b))==NULL)
			return -2;
		f->ac = 1+is32b, f->t = t;
		if (scale!=0x10000)
			cc+= indexedScaled;
		else if ((uint8)fmt==m->bpp && ((pmode=mode&15)<mTransparent || m->bpp==1))
		{ /* if src/dst depths are the same & the table comes out an identity matrix
		    (a no-change table), then dispose of the table and pipe it normally */
			i = 0, c = 1u<<(uint8)fmt;
			if (is32b)
				do
					if (*t++!=i)
						goto sett;
				while(++i<c);
			else
				do
					if (*((uint16*)t)++!=(uint16)i)
						goto sett;
				while(++i<c);
			free(f->t);
			f->t = NULL, f->ac = 0;
		nix:	if (pmode==mTransparent)
				mode = mSrcOr;
			mode&=~0x30;
		}
	sett:	f->cm = 0, cf = f->t;
		goto spos;
	}
	cc = ((fmt>>12&2)|((fmt&pTypeLSB)>>pTypeLSBBit^!BIG_ENDIAN))+(directSrc+4);
	if (m->pSize<9)
	{
		if (!(flag&~1))
			cc+= dstInverse;
		if (!m->pTable || !m->minCmp)
		{
			if (!(flag&remapForce30))
				cc+= dstGray;
			if (!m->iTable && pmMakeITables(m,0)<=0)
				return -2;
			f->itb = m->iTable;
		}
	} else if (!(flag&remapForce30) && !(uint16)(fmt^formPxFmt(m)))
		cc-= 4;
	if (scale!=0x10000)
		cc+= directScaled;
	if (flag&remapForce30)
		i = j = 10;
	else
		i = m->minCmp, j = i+(m->pType>>pTypeLGreenBit&1);
	if (cc-directSrc<4u)
		goto sinv;
	if ((t=calloc(4*_cvdSize,1))==NULL)
		return -2;
	f->t = t, f->ac = 4;
	{
		int		si, di;

		t[_cvsi] = si = fmt&0xFF;
		t[_cvsj] = di = si+(fmt>>pTypeLGreenBit&1);
		t[_cvdi] = si*2;
		t[_cvdj] = t[_cvdj]*2;
		t[_cvki] =~(~0u>>si);
		t[_cvkj] =~(~0u>>di);
		t[_cvub] = 32-si;
		t[_cvug] = t[_cvub]-di;
		t[_cvur] = t[_cvug]-si;
	}
	t[_cvhb] = 32-i;
	t[_cvhg] = t[_cvhb]-j;
	t[_cvhr] = t[_cvhg]-i;
	t[_cvmg] =~(~0u<<j)<<i;
	t[_cvmr] =~(~0u<<i)<<i<<j;
sinv:	f->cm = flag&1 ? i ? (1u<<(i*3+(m->pType>>pTypeLGreenBit&1)))-1:1023:0;
spos:	f->cc = cc;
	f->rtn = f_cnvx[cc];
	f->hx =~0, f->rx = scale;
	return mode;
}

void rEndCnv(cnvParam *f)
{
	if (!f->ac)
		return;
	free(f->t); f->t = NULL;
	f->ac = 0;
}

/* ---- basic conversion */

void b1to8(uint8 *d,
	const uint8 *s,
	int c)
{
	if (c<=0)
		return;
	do
	{
		uint8		v = *s++;

		*d = v>>7&1;
		d[1] = v>>6&1;
		d[2] = v>>5&1;
		d[3] = v>>4&1;
		d[4] = v>>3&1;
		d[5] = v>>2&1;
		d[6] = v>>1&1;
		d[7] = v&=1;
	} while(d+=8,--c);
}
