/*
 mctl.c
  control objects (buttons, sliders, etc.)

   Revised: 2014.11.26
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
*/

#if _MSC_EXTENSIONS
 #include <windows.h>
#else
 #include <string.h>
 extern unsigned long GetTickCount(void);
#endif
#define __MWIND_PRIVATE
#include "mtext.h"
#include "mevent.h"
#include "mwind.h"

#include "../ftrig.h"

/* ---- globals */

const rgb48	_defMCtlColors[3] =
{
	{0,0,0}, /* frame color */
	{0,0,0}, /* title color */
	{52428,52428,52428} /* content */
};

uint16		ctlWheelScrollCount = 64;

/* ---- code */

static inline void rgb48lerp(rgb48 *out,
	const rgb48 *a,
	const rgb48 *b,
	ifix16 m)
{
	int		v;

	v = ((b->red-a->red)*m>>8)+a->red, out->red = v>65535u ? ~v>>31:v;
	v = ((b->green-a->green)*m>>8)+a->green, out->green = v>65535u ? ~v>>31:v;
	v = ((b->blue-a->blue)*m>>8)+a->blue, out->blue = v>65535u ? ~v>>31:v;
}

static inline void rgb48amp(rgb48 *out,
	const rgb48 *a,
	ifix16 m)
{
	int		v;

	v = a->red*m>>8, out->red = MIN(v,65535);
	v = a->green*m>>8, out->green = MIN(v,65535);
	v = a->blue*m>>8, out->blue = MIN(v,65535);
}

static inline void rgb48add(rgb48 *out,
	const rgb48 *a,
	uint16 m)
{
	int		v;

	v = a->red+m, out->red = MIN(v,65535);
	v = a->green+m, out->green = MIN(v,65535);
	v = a->blue+m, out->blue = MIN(v,65535);
}

static void wDelay(uint16 tickCount)
{
	int32		tick;

	tick = GetTickCount();
	do
		global_idle_proc();
	while(GetTickCount()-tick<tickCount);
}

static int ctlGeneral(ctlObj *ctl,
	int msg,
	const void *data)
{
	pt2d16		pt, adja;
	int32		oldvalue;
	int		type;

	if (ctl->hilite==255 || !ctl->visible) /* disabled or invisible */
		return 0;
	if ((msg&~1)==wcMsgTest)
		return ptInRect16(&ctl->bounds,*(const pt2d16*)data);
	if (data)
		pt = *(const pt2d16*)data;
	else
		get_mouse_loc(&pt.x,&pt.y);
	*(int32*)&adja = 0, oldvalue = ctl->value;
	if (ctl->trk)
	{
		*(int32*)&adja = ctl->rtn(ctl,wcMsgMarkMeas,(void*)oldvalue);
		adja.x = pt.x-adja.x, adja.y = pt.y-adja.y;
	}
	type = ctl->rtn(ctl,wcMsgType,0);
	if ((msg&~1)==wcMsgClick)
	{
		pt2d16		oldpt, pt;
		int		dn, part;

		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		dn = part = ctl->rtn(ctl,wcMsgTest,&pt);
		oldpt.x = oldpt.y = IMIN(int16);
		while(mouse_still_down())
		{
			if (dn!=ctl->hilite)
				wHiliteControl(ctl,dn);
			if (ctl->trk && (dn || (type==ctlTypeScrollBar && part==kControlLabelPart)))
			{
				pt.x-= adja.x, pt.y-= adja.y;
				if (part!=kControlLabelPart || *(int32*)&pt!=*(int32*)&oldpt)
				{
					int32		value = dn-1<254u
						? ctl->rtn(ctl,wcMsgDragMeas,&pt):oldvalue;

					ctl->trk(ctl,dn,value);
				}
				oldpt = pt;
			}
			get_mouse_loc(&pt.x,&pt.y);
			dn = ctl->rtn(ctl,part==kControlLabelPart ? wcMsgWTest:wcMsgTest,&pt);
			if (dn!=part)
				dn = dn-1<254u && type==ctlTypeScrollBar && part==kControlLabelPart;
		}
		if (ctl->hilite)
		{ /* if within, reset hilite and return part */
			wHiliteControl(ctl,0);
			wDelay(NATIVE_TICKS_PER_SECOND/24);
			return part;
		} else if (type==ctlTypeScrollBar && part==1 && ctl->value!=oldvalue)
			wSetControlValue(ctl,oldvalue);
		return 0;
	}
	if (msg==wcMsgDispose)
	{
		free(ctl->data);
		ctl->data = NULL;
	}
	return -8;
}

static void ctlBoxText(ctlObj *ctl,
	const rgb48 *colors,
	int o)
{
	if (*ctl->title)
	{
		const bmfMemory	*font;
		grafPort	*p = &ctl->wind->port;
		const bmfMemory	*oldfm = p->txFM;
		uint8		oldsty = p->txStyle;

		RGBFore(p,colors+wCTitle);
		RGBBack(p,colors+wCContent);
		p->txFM = font = ffMatch(fmFonts[ctl->txFont],ctl->txSize,ctl->txStyle);
		p->txStyle = ctl->txStyle;
		rDrawText(p,(const char*)ctl->title+1,*ctl->title,ctl->bounds.top
			+((ctl->bounds.bottom-ctl->bounds.top+(font->ascent>>1)
			+font->descent)>>1),(ctl->bounds.left+o)<<16);
		p->txFM = oldfm, p->txStyle = oldsty;
	}
}

#define btnRnd			4

#define btnEmb1			(btnRnd-2+(btnRnd<5)+(btnRnd==6))
#define btnEmb2			(btnRnd/2)

static DEF_CTL_RTN(pushBtnProc)
{
	rect2dq		b;
	penState	pen;
	grafPort	*p;
	int		w, dn = 0;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		rect16		oldcl;
		rgb48		oldfg, oldbk, cont;
		const rgb48	*color, *emb;
		const bmfMemory	*font;
		uint8		oldtm;

		if (!ctl->visible || !p)
			return 0;
		oldcl = p->clip, oldtm = p->txMode;
		oldfg = p->fgColor, oldbk = p->bkColor;
		FixRect(&b,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
		color = ctl->data ? ctl->data:_defMCtlColors;
		rPenNormal(p,&pen);
		p->penw = p->penh =-1;
		/* fill and frame */
		dn = ctl->hilite-1<254u;
		cont = p->bkColor = color[wCContent];
		if (dn)
			cont.red>>= 1, cont.green>>= 1, cont.blue>>= 1;
		RGBFore(p,p->map->bpp==1 && dn ? &rgb_black:&cont);
		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		p->penw = p->penh = 0x10000;
		rgb48lerp(&p->fgColor,&p->fgColor,color+wCContent,.5*256);
		if (ctl->hilite!=255)
			if (dn)
			{
				p->clip.left = MAX(ctl->bounds.left+btnRnd/2,oldcl.left);
				p->clip.top = MAX(ctl->bounds.top+btnRnd/2,oldcl.top);
			} else
			{
				p->clip.right = MIN(ctl->bounds.right-btnRnd/2,oldcl.right);
				p->clip.bottom = MIN(ctl->bounds.bottom-btnRnd/2,oldcl.bottom);
			}
		rgb48lerp(&p->fgColor,&p->fgColor,color,.5*256);
		RGBFore(p,&p->fgColor);
		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		if (ctl->hilite!=255)
		{
			p->clip = oldcl;
			if (dn)
			{
				p->clip.right = MIN(ctl->bounds.right-btnRnd/2,oldcl.right);
				p->clip.bottom = MIN(ctl->bounds.bottom-btnRnd/2,oldcl.bottom);
			} else
			{
				p->clip.left = MAX(ctl->bounds.left+btnRnd/2,oldcl.left);
				p->clip.top = MAX(ctl->bounds.top+btnRnd/2,oldcl.top);
			}
			rgb48lerp(&p->fgColor,&p->fgColor,color,.5*256);
			RGBFore(p,&p->fgColor);
			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		}
		p->penw = p->penh = 0xC000;
		RGBFore(p,color);
		p->clip = oldcl;
		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		p->txMode = mTransparent|mTextGray;
		if (ctl->hilite!=255)
		{
			rgb48		lt;

			p->penw = p->penh = 1<<16;
			lt.red = MIN(color[2].red,52428)+13107;
			lt.green = MIN(color[2].green,52428)+13107;
			lt.blue = MIN(color[2].blue,52428)+13107;
			emb = dn ? &lt:color;
		  /* br emboss - 1 */
			rgb48lerp(&p->fgColor,emb,&cont,.5*256);
			RGBFore(p,&p->fgColor);
			FixRect(&b,ctl->bounds.left-1,ctl->bounds.top-1,
				ctl->bounds.right-1,ctl->bounds.bottom-1);
			p->clip.left = MAX(ctl->bounds.left+btnEmb1,oldcl.left);
			p->clip.top = MAX(ctl->bounds.top+btnEmb1,oldcl.top);
			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		  /* br emboss - 2 */
			rgb48lerp(&p->fgColor,&p->fgColor,&cont,.5*256);
			RGBFore(p,&p->fgColor);
			FixRect(&b,ctl->bounds.left-2,ctl->bounds.top-2,
				ctl->bounds.right-2,ctl->bounds.bottom-2);
			p->clip.left = MAX(ctl->bounds.left+btnEmb2,oldcl.left);
			p->clip.top = MAX(ctl->bounds.top+btnEmb2,oldcl.top);
			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		  /* tl emboss - 1 */
			emb = dn ? color:&lt;
			rgb48lerp(&p->fgColor,emb,&cont,.5*256);
			RGBFore(p,&p->fgColor);
			FixRect(&b,ctl->bounds.left+1,ctl->bounds.top+1,
				ctl->bounds.right+1,ctl->bounds.bottom+1);
			p->clip = oldcl;
			p->clip.right = MIN(ctl->bounds.right-btnEmb1,oldcl.right);
			p->clip.bottom = MIN(ctl->bounds.bottom-btnEmb1,oldcl.bottom);
			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		  /* tl emboss - 2 */
			rgb48lerp(&p->fgColor,&p->fgColor,&cont,.75*256);
			RGBFore(p,&p->fgColor);
			if (!p->fgSolid)
			{
				rgb48lerp(&p->fgColor,&p->fgColor,&cont,.5*256);
				RGBFore(p,&p->fgColor);
			}
			FixRect(&b,ctl->bounds.left+2,ctl->bounds.top+2,
				ctl->bounds.right+2,ctl->bounds.bottom+2);
			w = btnEmb2+(dn && btnRnd<5);
			p->clip.right = MIN(ctl->bounds.right-w,oldcl.right);
			p->clip.bottom = MIN(ctl->bounds.bottom-w,oldcl.bottom);
			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);

			p->txMode = mTransparent;
		}

		RGBFore(p,p->map->bpp==1 && dn ? &rgb_white:color+1);
		p->clip.top = MAX(oldcl.top,ctl->bounds.top);
		p->clip.left = MAX(oldcl.left,ctl->bounds.left);
		p->clip.bottom = MIN(oldcl.bottom,ctl->bounds.bottom);
		p->clip.right = MIN(oldcl.right,ctl->bounds.right);
		if ((font=ffMatch(fmFonts[ctl->txFont],ctl->txSize,ctl->txStyle))!=NULL)
		{
			const bmfMemory	*oldfm = p->txFM;
			uint8		oldsty = p->txStyle;

			w = rTextWidth(font,(const char*)ctl->title+1,*ctl->title,
				NULL,0,0,ctl->txStyle);
			p->txFM = font, p->txStyle = ctl->txStyle;
			rDrawText(p,(const char*)ctl->title+1,*ctl->title,
				((ctl->bounds.bottom-ctl->bounds.top+(font->ascent>>1)+font->descent)
				>>1)+ctl->bounds.top,(ctl->bounds.right+ctl->bounds.left-w)<<15);
			p->txFM = oldfm, p->txStyle = oldsty;
		}
		rRestorePen(p,&pen,1);
		p->txMode = oldtm;
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		p->clip = oldcl;
		return 1;
	} else if ((msg&~1)==wcMsgTest)
	{ /* rough simulation of a "within round-rect" test */
		pt2d16		loc = *(pt2d16*)data;
		int		h, i;

		if ((w=loc.x-ctl->bounds.left)<0)
			return 0;
		if ((i=ctl->bounds.right-1-loc.x)<0)
			return 0;
		if (i<w)
			w = i;
		if ((h=loc.y-ctl->bounds.top)<0)
			return 0;
		if ((i=ctl->bounds.bottom-1-loc.y)<0)
			return 0;
		if (i<h)
			h = i;
		return w!=h || w>1 || h>1;
	} else if (msg==wcMsgType)
		return ctlTypePushBtn;
	else
		return ctlGeneral(ctl,msg,data);
}

/*
 FF00FF FFFFFF EEEEEE DDDDDD CCCCCC BBBBBB AAAAAA 999999
 888888 777777 666666 555555 333333 9F9F9F E6E6E6 000000
*/
static ifix16	amplist[10] =
{
	0xCCCC00/0xCCCC+.999,
	0xBBBB00/0xCCCC+.999,
	0xAAAA00/0xCCCC+.999,
	0x999900/0xCCCC+.999,
	0x888800/0xCCCC+.999,
	0x777700/0xCCCC+.999,
	0x666600/0xCCCC+.999,
	0x555500/0xCCCC+.999,
	0x333300/0xCCCC+.999
};

/* 16-color palette for radioBtn/checkBox image base */
int _imgContColors(cTab64 *ct,
	const rgb48 *cont,
	const rgb48 *back,
	boolean disabled)
{
	cSpec		*cs = ct->colors;
	int		i;

	ct->seed = 0;
	ct->count = 16;
	ct->flags = 0x8000;
	if (back)
	{
		cs->rgb = *back;
		rgb48amp(&cs[13].rgb,back,0x9F9F00/0xFFFF+.999);
		rgb48amp(&cs[14].rgb,back,0xDFDF00/0xFFFF+.999);
	}
	cs[15].rgb.red = cs[15].rgb.green = cs[15].rgb.blue = 0;
	if (!cont)
		cont = system_colors+_system_gray73;
	cs[3].rgb = *cont;
	rgb48add(&cs[1].rgb,cont,0x3333);
	rgb48add(&cs[2].rgb,cont,0x1919);
	i = 0;
	do
		rgb48amp(&cs[i+4].rgb,cont,amplist[i]);
	while(++i<9);
	if (disabled)
	{
		cs[7].rgb = cs[6].rgb = cs[5].rgb =
			cs[4].rgb = cs[2].rgb = cs[1].rgb = *cont;
		rgb48lerp(&cs[8].rgb,&cs[8].rgb,cont,.8*256);
		if (cont!=system_colors+_system_gray93)
		{
			rgb48amp(&cs[15].rgb,cont,.5*256);
			rgb48lerp(&cs[14].rgb,&cs[14].rgb,cont,.5*256);
			rgb48lerp(&cs[13].rgb,&cs[13].rgb,cont,.5*256);
			rgb48lerp(&cs[12].rgb,&cs[12].rgb,cont,.5*256);
			rgb48lerp(&cs[11].rgb,&cs[11].rgb,cont,.5*256);
			rgb48lerp(&cs[10].rgb,&cs[10].rgb,cont,.5*256);
			rgb48lerp(&cs[9].rgb,&cs[9].rgb,cont,.6*256);
		}
		return mTransparent|mTextGray;
	}
	return mTransparent;
};

static uint8	checkBoxMerge[360] =
{ /* based on (but isn't) the Appearance 1.0 checkBox control */
	0x11,0x11,0x11,0x11,0x15,0x00,0x99,0x99,0x99,0x99,0x96,0x00,
	0x13,0x33,0x33,0x33,0x38,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x44,0x44,0x44,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55,0x00,
	0x11,0x11,0x11,0x11,0x15,0xFF,0x99,0x99,0x99,0x99,0x98,0xFF,
	0x13,0x33,0x33,0x33,0x5F,0xFD,0x96,0x66,0x66,0x66,0x6F,0xFD,
	0x13,0x33,0x33,0x35,0xFC,0x00,0x96,0x66,0x66,0x66,0xFF,0x00,
	0x13,0x33,0x33,0x5F,0xCB,0x00,0x96,0x66,0x66,0x6F,0xF9,0x00,
	0x1F,0xC3,0x35,0xFC,0x99,0x00,0x9F,0xF6,0x66,0xFF,0xB5,0x00,
	0x13,0xFF,0x5F,0xF9,0x58,0x00,0x96,0xFF,0x6F,0xFB,0x65,0x00,
	0x13,0x5F,0xFF,0x95,0x48,0x00,0x96,0x6F,0xFF,0xB6,0x65,0x00,
	0x13,0x36,0xF9,0x53,0x48,0x00,0x96,0x68,0xFB,0x66,0x65,0x00,
	0x13,0x44,0x66,0x44,0x48,0x00,0x96,0x66,0x86,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55,0x00,
	0x11,0x11,0x11,0x11,0x15,0x00,0x99,0x99,0x99,0x99,0x96,0x00,
	0x13,0x33,0x33,0x33,0x38,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0xBB,0xBB,0xBB,0x68,0x00,0x96,0xCC,0xCC,0xCC,0x85,0x00,
	0x13,0xFF,0xFF,0xFF,0x88,0x00,0x96,0xFF,0xFF,0xFF,0x95,0x00,
	0x13,0x56,0x66,0x66,0x68,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x44,0x44,0x44,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55
};

static DEF_CTL_RTN(checkBoxProc)
{
	grafPort	*p;
	penState	pen;
	int		u;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{24,30,{0,0},12,4,0,4,4,0,0,(int8*)checkBoxMerge};
		rect2dq		b;
		rect16		oldcl;
		rgb48		oldfg, oldbk;
		const rgb48	*color;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		uint8		oldtm;

		if (!ctl->visible || !p)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor;
		oldcl = p->clip, oldtm = p->txMode;
		rPenNormal(p,&pen);
		SectRect16(&p->clip,&ctl->bounds);
		color = ctl->data ? ctl->data:_defMCtlColors;
		RGBBack(p,&p->bkColor);
		rEraseRect(p,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:color+wCContent,&p->bkColor,
			ctl->hilite==255);
		m.pTable = ct;
		u = (ctl->bounds.top+ctl->bounds.bottom-12)>>1;
		p->bkColor = color[wCFrame];
		if (ctl->hilite==255)
			rgb48lerp(&p->bkColor,color,color+wCContent,.5*256);
		RGBFore(p,&p->bkColor);
		FixRect(&b,ctl->bounds.left+2,u,ctl->bounds.left+14,u+12);
		p->penw = p->penh = 0x10000;
		rRect(p,&b);
		rBltPix(p,&m,(ctl->hilite-1<254u)*12,PIN(ctl->value,0,2)*10,12,10,
			ctl->bounds.left+3,u+1,12,10,mTransparent);
		ctlBoxText(ctl,color,18);
		p->clip = oldcl;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	} else if (msg==wcMsgType)
		return ctlTypeCheckBox;
	return ctlGeneral(ctl,msg,data);
}

static uint8	radioBtnMerge[408] =
{ /* based on (but isn't) the Appearance 1.0 radioBtn control */
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xB9,0x43,0x35,0x9B,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0x53,0x21,0x11,0x38,0xB0,0xBC,0xBA,0x99,0x88,0xBB,0x00,
	0xD9,0x31,0x11,0x22,0x35,0xBD,0xCB,0xA9,0x88,0x76,0x6B,0xD0,
	0xC4,0x21,0x12,0x23,0x45,0x8C,0xBA,0x98,0x87,0x65,0x58,0xC0,
	0xF3,0x11,0x22,0x34,0x56,0x8F,0xB9,0x88,0x76,0x55,0x46,0xF0,
	0xF3,0x12,0x23,0x45,0x56,0x8F,0xB9,0x87,0x65,0x54,0x46,0xF0,
	0xC5,0x12,0x34,0x55,0x67,0x9C,0xB8,0x76,0x55,0x44,0x57,0xC0,
	0xD9,0x33,0x45,0x56,0x78,0xBD,0xC8,0x65,0x54,0x44,0x6B,0xD0,
	0x0B,0x85,0x56,0x67,0x8A,0xB0,0xCB,0x65,0x44,0x56,0x8C,0x00,
	0x0E,0xBB,0x88,0x89,0xBB,0xE0,0xEC,0xB8,0x66,0x7B,0xCE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xBB,0x98,0x88,0xBB,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0xA8,0x66,0x65,0x58,0xB0,0xBB,0xB9,0x99,0x88,0xBB,0x00,
	0xDB,0x87,0xBF,0xFA,0x43,0x9D,0xBB,0xAC,0xFF,0xC7,0x6B,0xD0,
	0xC9,0x6B,0xFF,0xFF,0xA1,0x5C,0xB9,0xCF,0xFF,0xFC,0x48,0xC0,
	0xF8,0x6F,0xFF,0xFF,0xF2,0x3F,0xB9,0xFF,0xFF,0xFF,0x56,0xF0,
	0xF8,0x6F,0xFF,0xFF,0xF2,0x3F,0xB9,0xFF,0xFF,0xFF,0x56,0xF0,
	0xC8,0x5A,0xFF,0xFF,0xA1,0x3C,0xB8,0xCF,0xFF,0xFC,0x46,0xC0,
	0xDB,0x54,0xAF,0xFA,0x33,0x9D,0xC8,0x7C,0xFF,0xC6,0x6B,0xD0,
	0x0B,0x83,0x12,0x21,0x35,0xB0,0xCB,0x64,0x55,0x46,0x8C,0x00,
	0x0E,0xB9,0x53,0x33,0x9B,0xE0,0xEC,0xB8,0x66,0x6B,0xCE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xB9,0x43,0x35,0x9B,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0x53,0x21,0x11,0x38,0xB0,0xBC,0xBA,0x98,0x88,0xBB,0x00,
	0xD9,0x31,0x11,0x22,0x35,0xBD,0xCB,0xA9,0x88,0x76,0x6B,0xD0,
	0xC4,0x21,0x12,0x23,0x45,0x8C,0xBA,0x99,0x98,0x75,0x48,0xC0,
	0xF3,0x19,0xBB,0xBB,0xB6,0x8F,0xB9,0xCC,0xCC,0xCB,0x46,0xF0,
	0xF3,0x1C,0xFF,0xFF,0xC6,0x8F,0xB9,0xFF,0xFF,0xFF,0x46,0xF0,
	0xC5,0x12,0x45,0x66,0x67,0x9C,0xB8,0x76,0x55,0x44,0x57,0xC0,
	0xD9,0x33,0x45,0x56,0x78,0xB6,0xB8,0x65,0x54,0x44,0x6B,0xD0,
	0x0B,0x85,0x55,0x67,0x8A,0xB0,0xBB,0x64,0x44,0x56,0x8B,0x00,
	0x0E,0xBB,0x88,0x89,0xBB,0xE0,0xEB,0xB8,0x66,0x7B,0xBE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD
};

static DEF_CTL_RTN(radioBtnProc)
{
	grafPort	*p;
	penState	pen;
	int		u;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{24,33,{0,0},12,4,0,4,4,0,0,(int8*)radioBtnMerge};
		rect16		oldcl;
		rgb48		oldfg, oldbk;
		const rgb48	*color;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		uint8		oldtm;

		if (!ctl->visible || !p)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor;
		oldcl = p->clip, oldtm = p->txMode;
		rPenNormal(p,&pen);
		SectRect16(&p->clip,&ctl->bounds);
		color = ctl->data ? ctl->data:_defMCtlColors;
		RGBFore(p,&p->bkColor);
		rFillRect(p,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:color+wCContent,&p->bkColor,
			ctl->hilite==255);
		m.pTable = ct;
		u = (ctl->bounds.top+ctl->bounds.bottom-12)>>1;
		//p->bkColor = color[wCFrame];
		if (ctl->hilite==255)
			rgb48lerp(&p->bkColor,color,color+wCContent,.5*256);
		rBltPix(p,&m,(ctl->hilite-1<254u)*11,PIN(ctl->value,0,2)*11,12,12,
			ctl->bounds.left+2,u,12,12,mSrcCopy);
		ctlBoxText(ctl,color,18);
		p->clip = oldcl, p->txMode = oldtm;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	} else if (msg==wcMsgType)
		return ctlTypeRadioBtn;
	return ctlGeneral(ctl,msg,data);
}

/* scroll bar */

 #define sb_xup			0
 #define sb_xdown		15
 #define sb_xthumb		30
 #define sb_xhilite		45
 #define sb_yvert		0
 #define sb_yhoriz		15

static uint8	scrollBarMerge[1426] =
{ /* based on (but isn't) the Appearance 1.0 scrollBar control */
	0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xF0,0xC1,0x11,
	0x11,0x11,0x11,0x11,0x11,0x3C,0x11,0x11,0x11,0x11,0x11,0x11,
	0x13,0xC2,0x44,0x44,0x44,0x44,0x44,0x44,0xDF,0xBB,0xBB,0xBB,
	0xBB,0xBB,0xBB,0xB8,0xFB,0xBB,0xBB,0xBB,0xBB,0xBB,0xBB,0x8F,
	0x4D,0xDD,0xDD,0xDD,0xDD,0xDD,0xDA,0xF0,0xC1,0x33,0x33,0x33,
	0x33,0x33,0x33,0x6C,0x13,0x33,0x33,0x33,0x33,0x33,0x36,0xC4,
	0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xB8,0x88,0x88,0x88,0x88,
	0x88,0x85,0xFB,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,0xDA,0xAA,
	0xAA,0xAA,0xAA,0xAA,0xAC,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,
	0x33,0x6C,0x13,0x33,0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x77,
	0x77,0x77,0x77,0x77,0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,
	0xFB,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,0xDA,0xAA,0xAA,0xAA,
	0xAA,0xAA,0xAC,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,
	0x13,0x33,0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x24,0x44,0x44,
	0x46,0x77,0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,
	0x88,0x88,0x88,0x88,0x88,0x5F,0xDA,0xA4,0x77,0x77,0x77,0x9A,
	0xAC,0xF0,0xC1,0x33,0x33,0x36,0x63,0x33,0x33,0x6C,0x13,0x33,
	0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x7C,0xCC,0xCC,0xCC,0x87,
	0xAF,0xB8,0x88,0x88,0xAA,0x88,0x88,0x85,0xFB,0x88,0x88,0x88,
	0x88,0x88,0x88,0x5F,0xDA,0xAA,0xFF,0xFF,0xFF,0xFB,0xAC,0xF0,
	0xC1,0x33,0x33,0x8F,0xF8,0x33,0x33,0x6C,0x13,0x8C,0xFF,0xFF,
	0xFF,0xF8,0x36,0xC4,0x77,0x24,0x44,0x44,0x46,0x77,0xAF,0xB8,
	0x88,0x8B,0xFF,0xB8,0x88,0x85,0xFB,0x8B,0xFF,0xFF,0xFF,0xFF,
	0xB8,0x5F,0xDA,0xA4,0x77,0x77,0x77,0x9A,0xAC,0xF0,0xC1,0x33,
	0x38,0xFF,0xFF,0x83,0x33,0x6C,0x13,0x38,0xCF,0xFF,0xFF,0x83,
	0x36,0xC4,0x77,0x7C,0xCC,0xCC,0xCC,0x87,0xAF,0xB8,0x88,0xBF,
	0xFF,0xFB,0x88,0x85,0xFB,0x88,0xBF,0xFF,0xFF,0xFB,0x88,0x5F,
	0xDA,0xAA,0xFF,0xFF,0xFF,0xFB,0xAC,0xF0,0xC1,0x33,0x8F,0xFF,
	0xFF,0xF8,0x33,0x6C,0x13,0x33,0x8C,0xFF,0xF8,0x33,0x36,0xC4,
	0x77,0x24,0x44,0x44,0x46,0x77,0xAF,0xB8,0x8B,0xFF,0xFF,0xFF,
	0xB8,0x85,0xFB,0x88,0x8B,0xFF,0xFF,0xB8,0x88,0x5F,0xDA,0xA4,
	0x77,0x77,0x77,0x9A,0xAC,0xF0,0xC1,0x38,0xCC,0xCC,0xCC,0xCC,
	0x83,0x6C,0x13,0x33,0x38,0xCF,0x83,0x33,0x36,0xC4,0x77,0x7C,
	0xCC,0xCC,0xCC,0x87,0xAF,0xB8,0xBF,0xFF,0xFF,0xFF,0xFB,0x85,
	0xFB,0x88,0x88,0xBF,0xFB,0x88,0x88,0x5F,0xDA,0xAA,0xFF,0xFF,
	0xFF,0xFB,0xAC,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,
	0x13,0x33,0x33,0x66,0x33,0x33,0x36,0xC4,0x77,0x24,0x44,0x44,
	0x46,0x77,0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,
	0x88,0x8A,0xA8,0x88,0x88,0x5F,0xDA,0xA4,0x77,0x77,0x77,0x9A,
	0xAC,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,0x13,0x33,
	0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x7C,0xCC,0xCC,0xCC,0x87,
	0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,0x88,0x88,
	0x88,0x88,0x88,0x5F,0xDA,0xAA,0xFF,0xFF,0xFF,0xFB,0xAC,0xF0,
	0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,0x13,0x33,0x33,0x33,
	0x33,0x33,0x36,0xC4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xB8,
	0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,0x88,0x88,0x88,0x88,
	0x88,0x5F,0xDA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAC,0xF0,0xC1,0x33,
	0x33,0x33,0x33,0x33,0x33,0x6C,0x13,0x33,0x33,0x33,0x33,0x33,
	0x36,0xC4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xB8,0x88,0x88,
	0x88,0x88,0x88,0x85,0xFB,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,
	0xDA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAC,0xF0,0xC3,0x66,0x66,0x66,
	0x66,0x66,0x66,0x8C,0x36,0x66,0x66,0x66,0x66,0x66,0x68,0xCD,
	0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAF,0x85,0x55,0x55,0x55,0x55,
	0x55,0x55,0xF8,0x55,0x55,0x55,0x55,0x55,0x55,0x5F,0xAC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xCC,0xF0,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xF0,0xC1,0x11,0x11,0x11,0x11,0x11,0x11,0x3C,
	0x11,0x11,0x11,0x11,0x11,0x11,0x13,0xC2,0x44,0x44,0x44,0x44,
	0x44,0x44,0xDF,0xBB,0xBB,0xBB,0xBB,0xBB,0xBB,0xB8,0xFB,0xBB,
	0xBB,0xBB,0xBB,0xBB,0xBB,0x8F,0x4D,0xDD,0xDD,0xDD,0xDD,0xDD,
	0xDA,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,0x13,0x33,
	0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x77,0x77,0x77,0x77,0x77,
	0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,0x88,0x88,
	0x88,0x88,0x88,0x5F,0xDA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAC,0xF0,
	0xC1,0x33,0x33,0x33,0x38,0x33,0x33,0x6C,0x13,0x33,0x38,0x33,
	0x33,0x33,0x36,0xC4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xB8,
	0x88,0x88,0x88,0xB8,0x88,0x85,0xFB,0x88,0x88,0xB8,0x88,0x88,
	0x88,0x5F,0xDA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAC,0xF0,0xC1,0x33,
	0x33,0x33,0x8F,0x33,0x33,0x6C,0x13,0x33,0x3C,0x83,0x33,0x33,
	0x36,0xC4,0x77,0x27,0x27,0x27,0x27,0x77,0xAF,0xB8,0x88,0x88,
	0x8B,0xF8,0x88,0x85,0xFB,0x88,0x88,0xFB,0x88,0x88,0x88,0x5F,
	0xDA,0xA4,0xA4,0xA4,0xA4,0xAA,0xAC,0xF0,0xC1,0x33,0x33,0x38,
	0xFF,0x33,0x33,0x6C,0x13,0x33,0x3C,0xF8,0x33,0x33,0x36,0xC4,
	0x77,0x4C,0x4C,0x4C,0x4C,0x77,0xAF,0xB8,0x88,0x88,0xBF,0xF8,
	0x88,0x85,0xFB,0x88,0x88,0xFF,0xB8,0x88,0x88,0x5F,0xDA,0xA7,
	0xF7,0xF7,0xF7,0xFA,0xAC,0xF0,0xC1,0x33,0x33,0x8F,0xFF,0x33,
	0x33,0x6C,0x13,0x33,0x3C,0xFF,0x83,0x33,0x36,0xC4,0x77,0x4C,
	0x4C,0x4C,0x4C,0x77,0xAF,0xB8,0x88,0x8B,0xFF,0xF8,0x88,0x85,
	0xFB,0x88,0x88,0xFF,0xFB,0x88,0x88,0x5F,0xDA,0xA7,0xF7,0xF7,
	0xF7,0xFA,0xAC,0xF0,0xC1,0x33,0x36,0xFF,0xFF,0x33,0x33,0x6C,
	0x13,0x33,0x3C,0xFF,0xF6,0x33,0x36,0xC4,0x77,0x4C,0x4C,0x4C,
	0x4C,0x77,0xAF,0xB8,0x88,0xAF,0xFF,0xF8,0x88,0x85,0xFB,0x88,
	0x88,0xFF,0xFF,0xA8,0x88,0x5F,0xDA,0xA7,0xF7,0xF7,0xF7,0xFA,
	0xAC,0xF0,0xC1,0x33,0x36,0xCF,0xFF,0x33,0x33,0x6C,0x13,0x33,
	0x3C,0xFF,0xF6,0x33,0x36,0xC4,0x77,0x4C,0x4C,0x4C,0x4C,0x77,
	0xAF,0xB8,0x88,0xAF,0xFF,0xF8,0x88,0x85,0xFB,0x88,0x88,0xFF,
	0xFF,0xA8,0x88,0x5F,0xDA,0xA7,0xF7,0xF7,0xF7,0xFA,0xAC,0xF0,
	0xC1,0x33,0x33,0x8C,0xFF,0x33,0x33,0x6C,0x13,0x33,0x3C,0xFF,
	0x83,0x33,0x36,0xC4,0x77,0x4C,0x4C,0x4C,0x4C,0x77,0xAF,0xB8,
	0x88,0x8B,0xFF,0xF8,0x88,0x85,0xFB,0x88,0x88,0xFF,0xFB,0x88,
	0x88,0x5F,0xDA,0xA7,0xF7,0xF7,0xF7,0xFA,0xAC,0xF0,0xC1,0x33,
	0x33,0x38,0xCF,0x33,0x33,0x6C,0x13,0x33,0x3C,0xF8,0x33,0x33,
	0x36,0xC4,0x77,0x4C,0x4C,0x4C,0x4C,0x77,0xAF,0xB8,0x88,0x88,
	0xBF,0xF8,0x88,0x85,0xFB,0x88,0x88,0xFF,0xB8,0x88,0x88,0x5F,
	0xDA,0xA7,0xF7,0xF7,0xF7,0xFA,0xAC,0xF0,0xC1,0x33,0x33,0x33,
	0x8C,0x33,0x33,0x6C,0x13,0x33,0x3C,0x83,0x33,0x33,0x36,0xC4,
	0x77,0x6C,0x6C,0x6C,0x6C,0x77,0xAF,0xB8,0x88,0x88,0x8B,0xF8,
	0x88,0x85,0xFB,0x88,0x88,0xFB,0x88,0x88,0x88,0x5F,0xDA,0xA9,
	0xF9,0xF9,0xF9,0xFA,0xAC,0xF0,0xC1,0x33,0x33,0x33,0x38,0x33,
	0x33,0x6C,0x13,0x33,0x38,0x33,0x33,0x33,0x36,0xC4,0x77,0x78,
	0x78,0x78,0x78,0x77,0xAF,0xB8,0x88,0x88,0x88,0xB8,0x88,0x85,
	0xFB,0x88,0x88,0xB8,0x88,0x88,0x88,0x5F,0xDA,0xAA,0xBA,0xBA,
	0xBA,0xBA,0xAC,0xF0,0xC1,0x33,0x33,0x33,0x33,0x33,0x33,0x6C,
	0x13,0x33,0x33,0x33,0x33,0x33,0x36,0xC4,0x77,0x77,0x77,0x77,
	0x77,0x77,0xAF,0xB8,0x88,0x88,0x88,0x88,0x88,0x85,0xFB,0x88,
	0x88,0x88,0x88,0x88,0x88,0x5F,0xDA,0xAA,0xAA,0xAA,0xAA,0xAA,
	0xAC,0xF0,0xC3,0x66,0x66,0x66,0x66,0x66,0x66,0x8C,0x36,0x66,
	0x66,0x66,0x66,0x66,0x68,0xCD,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,
	0xAF,0x85,0x55,0x55,0x55,0x55,0x55,0x55,0xF8,0x55,0x55,0x55,
	0x55,0x55,0x55,0x5F,0xAC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xF0,
	0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

/* works both vertically and horizontally */
static uint8	wellPat[16] =
	{12,10,8,6,6,6,6,6,6,6,6,6,6,5,4,15};
static uint8	wellVShadow[32] =
{
	12,10,10,10,10,10,10,10,10,10,10,10,10,10,9,15,
	12,10,8,8,8,8,8,8,8,8,8,8,8,8,7,15
};
static uint8	wellHShadow[32] =
{
	12,12,
	10,10,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	9,7,
	15,15
};

typedef struct sBarLocs
{
	int		tx, ty, dx, dy;
	boolean		vert;
} sBarLocs;

static int calcScrollBarLocs(const ctlObj *ctl,
	sBarLocs *locs,
	int32 value)
{
	int32		u, h;
	uint8		hilite = ctl->hilite;
	boolean		vert;

	if ((u=RECTANGLE_WIDTH(&ctl->bounds))<=0
			|| (h=RECTANGLE_HEIGHT(&ctl->bounds))<=0)
		return locs ? -1:0;
	if ((vert=h>u)!=0)
		u = h;
	if (locs)
		locs->vert = vert;
	if ((u-=46)<0)
		u = 0;
	if (ctl->min>=ctl->max)
	{
		if (!locs)
			goto end;
		hilite = 255, u = 16;
		goto calc;
	}
	if (locs)
	{
		h = MAX(value,ctl->min);
		if (h>ctl->max)
			h = ctl->max;
		u = ufixmuldiv(u,h-ctl->min,ctl->max-ctl->min)+15;
	calc:	if (vert)
		{
			locs->tx = ctl->bounds.left, locs->ty = u+ctl->bounds.top;
			locs->dx = ctl->bounds.left, locs->dy = ctl->bounds.bottom-16;
		} else
		{
			locs->tx = u+ctl->bounds.left, locs->ty = ctl->bounds.top;
			locs->dx = ctl->bounds.right-16, locs->dy = ctl->bounds.top;
		}
	} else
	{
		if (value<0 || (value-=(vert ? ctl->bounds.top:ctl->bounds.left)+15)<0)
			value = 0;
		return ufixmuldiv(value,ctl->max-ctl->min,u);
	}
end:	return hilite;
}

static DEF_CTL_RTN(scrollBarProc)
{
	grafPort	*p;
	penState	pen;
	sBarLocs	locs;
	int		hilite;

	if (!ctl)
		return -1;
	if (msg==wcMsgType)
		return ctlTypeScrollBar;
	p = &ctl->wind->port;
	if ((hilite=calcScrollBarLocs(ctl,&locs,ctl->value))<0)
		return 0;
	if (msg==wcMsgDragMeas)
		return calcScrollBarLocs(ctl,NULL,locs.vert
			? ((const pt2d16*)data)->y:((const pt2d16*)data)->x);
	if (msg==wcMsgMarkMeas)
	{
		pt2d16		pt;

		pt.x = locs.tx!=(int16)locs.tx ? locs.tx>>31^IMAX(int16):locs.tx;
		pt.y = locs.ty!=(int16)locs.ty ? locs.ty>>31^IMAX(int16):locs.ty;
		return *(int32*)&pt;
	}
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{91,31,{0,0},46,4,0,4,4,0,0,(int8*)scrollBarMerge};
		rect16		oldcl;
		rgb48		oldfg;
		const rgb48	*colors;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		int		dw, dh, sy;

		if (!ctl->visible || !p)
			return 0;
		oldcl = p->clip, oldfg = p->fgColor;
		rPenNormal(p,&pen);
		SectRect16(&p->clip,&ctl->bounds);
		colors = ctl->data ? ctl->data:_defMCtlColors;
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:colors+wCContent,&rgb_white,
			hilite==255);
		if (hilite==255)
		{
			rgb48add(&cstor[2].rgb,&cstor[2].rgb,0x1111);
			cstor[11] = cstor[10] = cstor[9] = cstor[8] = cstor[7] =
				cstor[6] = cstor[5] = cstor[4] = cstor[3] = cstor[2];
		}
		m.pTable = ct;
		sy = locs.vert ? sb_yvert:sb_yhoriz;
		if (locs.vert || (dw=locs.tx-ctl->bounds.left)>16)
			dw = 16;
		if (!locs.vert || (dh=locs.ty-ctl->bounds.top)>16)
			dh = 16;
		rBltPix(p,&m,(hilite==kControlUpButtonPart)*sb_xhilite+sb_xup,sy,dw,dh,
			ctl->bounds.left,ctl->bounds.top,dw,dh,mSrcCopy);
		if (hilite!=255)
			rBltPix(p,&m,(hilite==kControlLabelPart)*sb_xhilite+sb_xthumb,sy,16,16,
				locs.tx,locs.ty,16,16,mSrcCopy);
		rBltPix(p,&m,(hilite==kControlDownButtonPart)*sb_xhilite+sb_xdown,sy,16,16,
			locs.dx,locs.dy,16,16,mSrcCopy);
		m.pSize = m.bpp = 8;
		if (locs.vert)
		{
			int		dy;

			m.h = 2, m.w = m.pitch = 16;
			dy = ctl->bounds.top+16;
			if (hilite!=255)
			{
				dh = locs.ty-dy;
				m.baseAddr = (int8*)wellVShadow;
				rBltPix(p,&m,0,0,16,2,ctl->bounds.left,dy,16,MIN(dh,2),0);
				m.baseAddr = (int8*)wellPat;
				rBltPix(p,&m,0,0,16,1,ctl->bounds.left,dy+2,16,dh-2,0);
				dy = locs.ty+16, dh = locs.dy-dy;
				m.baseAddr = (int8*)wellVShadow;
				rBltPix(p,&m,0,0,16,2,ctl->bounds.left,dy,16,MIN(dh,2),0);
				dy+= 2, dh-= 2;
			} else
				dh = locs.dy-dy;
			m.baseAddr = (int8*)wellPat;
			rBltPix(p,&m,0,0,16,1,ctl->bounds.left,dy,16,dh,0);
		} else
		{
			int		dx;

			m.w = m.pitch = 2, m.h = 16;
			dx = ctl->bounds.left+16;
			if (hilite!=255)
			{
				dw = locs.tx-dx;
				m.baseAddr = (int8*)wellHShadow;
				rBltPix(p,&m,0,0,2,16,dx,ctl->bounds.top,MIN(dw,2),16,0);
				m.baseAddr = (int8*)wellPat;
				m.w = m.pitch = 1, m.h = 16;
				rBltPix(p,&m,0,0,1,16,dx+2,ctl->bounds.top,dw-2,16,0);
				dx = locs.tx+16, dw = locs.dx-dx;
				m.baseAddr = (int8*)wellHShadow;
				m.w = m.pitch = 2, m.h = 16;
				rBltPix(p,&m,0,0,2,16,dx,ctl->bounds.top,MIN(dw,2),16,0);
				dx+= 2, dw-= 2;
			} else
				dw = locs.dx-dx;
			m.baseAddr = (int8*)wellPat;
			m.w = m.pitch = 1, m.h = 16;
			rBltPix(p,&m,0,0,1,16,dx,ctl->bounds.top,dw,16,0);
		}
		p->clip = oldcl;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		return 1;
	} else if (msg==wcMsgType)
		return ctlTypeScrollBar;
	if (ctl->hilite==255) /* disabled */
		return 0;
	if ((msg&~1)==wcMsgTest)
	{
		rect16		wbounds = ctl->bounds;
		pt2d16		pt;
		int		r;

		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		if (locs.vert)
		{
			if (msg&1)
				OutsetRect16(&wbounds,64,8);
			if (pt.y<wbounds.top || pt.y>=wbounds.bottom)
				return 0;
			if (pt.y<ctl->bounds.top+16)
				{r = kControlUpButtonPart; goto rchk;}
			if (pt.y<locs.ty)
				{r = kPageUpButtonPart; goto rchk;}
			if (pt.y<locs.ty+16)
				{r = kControlLabelPart; goto rchk;}
			if (pt.y<locs.dy)
				{r = kPageDownButtonPart; goto rchk;}
			r = kControlDownButtonPart;
			goto rchk;
		}
		if (msg&1)
			OutsetRect16(&wbounds,8,64);
		if (pt.x<wbounds.left || pt.x>=wbounds.right)
			return 0;
		if (pt.x<ctl->bounds.left+16)
			{r = kControlUpButtonPart; goto rchk;}
		if (pt.x<locs.tx)
			{r = kPageUpButtonPart; goto rchk;}
		if (pt.x<locs.tx+16)
			{r = kControlLabelPart; goto rchk;}
		if (pt.x<locs.dx)
			{r = kPageDownButtonPart; goto rchk;}
		r = kControlDownButtonPart;
	rchk:	return ptInRect16(&wbounds,pt) ? r:0;
	}
	return ctlGeneral(ctl,msg,data);
}

static DEF_TRK_RTN(sBarTrack)
{
	if (part==kControlLabelPart)
		wSetControlValue(ctl,value);
};

/* popup */

static uint8	popUpMerge[608] =
{
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0xCC,0xCC,0xCC,0xCC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,
	0xC0,0x00,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,
	0x33,0x33,0x33,0x33,0x33,0x45,0x7B,0xCC,0x00,0x33,0x33,0x32,
	0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,
	0x33,0x45,0xAC,0xC0,0x65,0x33,0x11,0x11,0x11,0x11,0x11,0x11,
	0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x12,0x35,0x6B,0xCF,0x67,
	0x32,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,
	0x11,0x11,0x22,0x35,0x78,0xCF,0x67,0x32,0x11,0x23,0x33,0x33,
	0x33,0x33,0x33,0x43,0x33,0x33,0x33,0x33,0x33,0x33,0x56,0x88,
	0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x33,0x33,0x35,0xC4,0x33,
	0x33,0x33,0x33,0x33,0x34,0x57,0x89,0xFF,0x67,0x32,0x11,0x33,
	0x33,0x33,0x33,0x33,0x5C,0xCC,0x53,0x33,0x33,0x33,0x33,0x34,
	0x57,0x99,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x33,0x35,0xCC,
	0xCF,0xC5,0x33,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,
	0x11,0x33,0x33,0x33,0x33,0x5C,0xCC,0xFF,0xFC,0x53,0x33,0x33,
	0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x35,
	0xCC,0xCF,0xFF,0xFF,0xC5,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,
	0x67,0x32,0x11,0x33,0x33,0x33,0x5C,0xCC,0xFF,0xFF,0xFF,0xFC,
	0x53,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,
	0x34,0xCC,0xCF,0xFF,0xFF,0xFF,0xFF,0xC4,0x33,0x33,0x34,0x67,
	0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x33,0x33,0x33,0x33,
	0x33,0x33,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,
	0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,
	0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x33,0x33,
	0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,
	0x32,0x11,0x33,0x33,0x34,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xC4,
	0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,
	0x5C,0xCC,0xCC,0xCC,0xCC,0xCC,0x53,0x33,0x33,0x34,0x67,0x9A,
	0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x35,0xCC,0xCC,0xCC,0xCC,
	0xC5,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,
	0x33,0x33,0x33,0x5C,0xCF,0xFF,0xFF,0x53,0x33,0x33,0x33,0x34,
	0x67,0x9A,0xFF,0x67,0x32,0x11,0x33,0x33,0x33,0x33,0x35,0xCC,
	0xFF,0xF5,0x33,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,
	0x11,0x33,0x33,0x33,0x33,0x33,0x5C,0xCF,0x53,0x33,0x33,0x33,
	0x33,0x34,0x67,0x9A,0xFF,0x67,0x32,0x12,0x33,0x33,0x33,0x33,
	0x33,0x35,0xC5,0x33,0x33,0x33,0x33,0x33,0x34,0x67,0x9A,0xFF,
	0x67,0x32,0x12,0x33,0x33,0x33,0x33,0x33,0x33,0x43,0x33,0x33,
	0x33,0x33,0x33,0x45,0x67,0x9A,0xFF,0x67,0x32,0x23,0x44,0x44,
	0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x56,0x68,
	0x9A,0xFF,0x67,0x33,0x35,0x66,0x66,0x66,0x66,0x66,0x66,0x66,
	0x66,0x66,0x66,0x66,0x66,0x66,0x78,0xAB,0xFF,0x67,0x44,0x57,
	0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,
	0x78,0x89,0xAF,0xFF,0x67,0x56,0x89,0x99,0x99,0x99,0x99,0x99,
	0x99,0x99,0x99,0x99,0x99,0x99,0x99,0x99,0xAA,0xFF,0xF0,0x67,
	0x78,0x99,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,
	0xAA,0xAA,0xAA,0xBF,0xFF,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xF0,
	0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00,0x00
};

static DEF_CTL_RTN(popUpProc)
{
	grafPort	*p;
	penState	pen;
	int		u, h;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{38,32,{0,0},19,4,0,4,4,0,0,(int8*)popUpMerge};
		rect16		oldcl;
		rgb48		oldfg, oldbk;
		const rgb48	*colors;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		uint8		oldtm;

		if (!ctl->visible || !p)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor;
		oldcl = p->clip, oldtm = p->txMode;
		rPenNormal(p,&pen);
		SectRect16(&p->clip,&ctl->bounds);
		colors = _defMCtlColors;
		RGBFore(p,&p->bkColor);
		rFillRect(p,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:colors+wCContent,&p->bkColor,
			ctl->hilite==255);
		m.pTable = ct;
		if ((h=ctl->bounds.bottom-ctl->bounds.top)>32)
			h = 32;
		u = (ctl->bounds.top+ctl->bounds.bottom-h)>>1;
		p->bkColor = colors[wCFrame];
		if (ctl->hilite==255)
			rgb48lerp(&p->bkColor,colors,colors+wCContent,.5*256);
		rBltPix(p,&m,0,0,38,32,
			ctl->bounds.left+4+ctl->max,u,(38*h+16)>>5,h,mTransparent);
		/* 15, 13, 10, 2, 0 */
		ctlBoxText(ctl,colors,4);
		p->clip = oldcl, p->txMode = oldtm;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	}
	if (msg==wcMsgType)
		return ctlTypePopUpMenu;
	if ((msg&~1)==wcMsgClick)
	{
		POINT		lpt;
		pt2d16		pt = mEventState.where;

		if (data)
		{
			grafPort	*p;

			pt = *(const pt2d16*)data;
			lpt.x = pt.x, lpt.y = pt.y;
			p = &ctl->wind->port;
			lpt.x-= p->origin.x+p->map->pOrigin.x;
			lpt.y-= p->origin.y+p->map->pOrigin.y;
		} else
			lpt.x = pt.x, lpt.y = pt.y;
		ClientToScreen(world_screen_window,&lpt);
		wHiliteControl(ctl,1);
		SetMenuDefaultItem(ctl->data,ctl->value,0);
		//for(i=0,v=ctl->value-1,c=GetMenuItemCount(ctl->data); i<c; ++i)
		//	CheckMenuItem(ctl->data,ctl->value,i==v ? MF_CHECKED:MF_UNCHECKED);
		if ((u=TrackPopupMenu(ctl->data,TPM_LEFTALIGN,lpt.x,lpt.y,0,
				world_screen_window,NULL))!=0)
		{
			MSG		msg;

			if (PeekMessage(&msg,world_screen_window,WM_COMMAND,WM_COMMAND,PM_REMOVE))
			{
				TranslateMessage(&msg);
				if (!HIWORD(msg.wParam))
					ctl->value = LOWORD(msg.wParam);
				DispatchMessage(&msg);
			}
		}
		wHiliteControl(ctl,0);
		return u;
	}
	if (msg==wcMsgDispose)
	{
		DestroyMenu(ctl->data); ctl->data = NULL;
		return 0;
	}
	return ctlGeneral(ctl,msg,data);
}

/* text box */

static DEF_CTL_RTN(textBoxProc)
{
	if (msg==wcMsgDraw)
	{
		rect2dq		b;
		grafPort	*port = &ctl->wind->port;
		penState	oldstate;
		textBoxObj	*tbox;

		if ((tbox=ctl->data)!=NULL
				&& tbox->flags&tBoxFrame)
		{
			rPenNormal(port,&oldstate);
			FixRect(&b,ctl->bounds.left-1,ctl->bounds.top-1,
				ctl->bounds.right+1,ctl->bounds.bottom+1);
			rRect(port,&b);
			rRestorePen(port,&oldstate,1);
		}
		return tBoxDraw(tbox);
	}
	if (msg==wcMsgType)
		return ((textBoxObj*)ctl->data)->flags&tBoxFrame
			? ctlTypeEditText:ctlTypeStaticText;
	if (msg==wcMsgTest)
		return ctl->data && ctl->hilite!=255
			&& ((textBoxObj*)ctl->data)->flags&(tBoxModifiable|tBoxListSelect)
			&& ptInRect16(&((textBoxObj*)ctl->data)->bounds,*(const pt2d16*)data)
			? kControlEditTextPart:0;
	if ((msg&~1)==wcMsgClick)
	{
		pt2d16		pt;

		if (ctl->hilite==255)
			return 0;
		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		return tBoxClick(ctl->data,pt,mEventState.modifiers,
			msg&1 ? dblClickFlag:0)>0 ? kControlEditTextPart:0;
	}
	if (msg==wcMsgDispose)
		{tBoxDispose(ctl->data); ctl->data = NULL;}
	return -8;
}

static DEF_CTL_RTN(pictProc)
{
	if (msg==wcMsgType)
		return ctlTypePicture;
	if (msg==wcMsgDraw)
	{
		pixMap		*m;
		int		r;

		if (!ctl->visible || set_mhandle_lock(ctl->data,1)<0)
			return 0;
		m = renderQDPict(*(const uint8**)ctl->data,get_mhandle_size(ctl->data));
		set_mhandle_lock(ctl->data,0);
		if (!m)
			return memory_error();
		r = rBltPix(&ctl->wind->port,m,0,0,m->w,m->h,
			ctl->bounds.left,ctl->bounds.top,
			RECTANGLE_WIDTH(&ctl->bounds),RECTANGLE_HEIGHT(&ctl->bounds),
			m->bpp>ctl->wind->port.map->bpp ? mDither:mSrcCopy);
		disposePixMap(m);
		return r;
	}
	if (msg==wcMsgDispose)
		{dispose_mhandle(ctl->data); ctl->data = NULL;}
	return ctlGeneral(ctl,msg,data);
}

static DEF_CTL_RTN(iconProc)
{
	if (msg==wcMsgType)
		return ctlTypeIcon;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{32,32,{0,0},32>>3,1,0,1,3};

		if (!ctl->visible || !ctl->data)
			return 0;
		m.baseAddr = *(mhandle)ctl->data;
		return rBltPix(&ctl->wind->port,&m,0,0,m.w,m.h,
			ctl->bounds.left,ctl->bounds.top,
			RECTANGLE_WIDTH(&ctl->bounds),RECTANGLE_HEIGHT(&ctl->bounds),mSrcCopy);
	}
	if (msg==wcMsgDispose)
		{dispose_mhandle(ctl->data); ctl->data = NULL;}
	return ctlGeneral(ctl,msg,data);
}

/* ---- high-level */

int wFindControl(const windowRef wind,
	pt2d16 loc,
	ctlObj **out)
{
	ctlObj		*ctl;
	int		r = 0;

	if (!wind || !out)
		return -1;
	for(ctl=wind->ctlList; ctl; ctl=ctl->next)
		if (ctl->visible && ptInRect16(&ctl->bounds,loc)
				&& (r=ctl->rtn(ctl,wcMsgTest,&loc))>0)
			{*out = ctl; break;}
	return r;
}

int32 wCountControls(windowRef wind)
{
	ctlObj		*ctl;
	long		c;

	if (!wind)
		return -1;
	for(c=0,ctl=wind->ctlList; ctl; ctl=ctl->next)
		++c;
	return c;
}

int32 wGetControlIndex(const windowRef wind,
	const ctlObj *s)
{
	ctlObj		*ctl;
	long		c;

	if (!wind)
		return -1;
	for(c=0,ctl=wind->ctlList; ctl; ctl=ctl->next,++c)
		if (ctl==s)
			return c;
	return -1;
}

int wDrawControls(windowRef wind)
{
	ctlObj		*ctl;
	long		r = 0, c;

	if ((c=wCountControls(wind))<=0)
		return c;
	/* render in reverse */
	if (--c)
	{
		ctlObj		**list = _alloca(sizeof(ctlObj*)*c);
		long		i;

		for(i=0,ctl=wind->ctlList; ++i<=c;)
		{
			ctl = ctl->next;
			list[i-1] = ctl;
		}
		for(i=c; --i>=0;)
		{
			ctl = list[i];
			if (ctl->visible)
				r+= ctl->rtn(ctl,wcMsgDraw,0);
		}
	}
	ctl = wind->ctlList;
	if (ctl->visible)
		r+= ctl->rtn(ctl,wcMsgDraw,0);
	if (wind->ctlDefault && wind->ctlDefault->visible)
	{
		rect2dq		outset;
		rect16		bounds = wind->ctlDefault->bounds;
		penState	oldstate;
		grafPort	*port = &wind->port;
		rgb48		oldfg = port->fgColor;
		rgb48		oldop = port->opColor;

		rPenNormal(port,&oldstate);
		port->penw = port->penh = 0x30000;
		RGBFore(port,&rgb_black);
		port->pnMode = mBlend;
		outset.top = subu(bounds.top<<16,0x40000-0x8000);
		outset.left = subu(bounds.left<<16,0x40000-0x8000);
		outset.bottom = addu(bounds.bottom<<16,0x40000+0x8000);
		outset.right = addu(bounds.right<<16,0x40000+0x8000);
		port->opColor = system_colors[_system_gray53];
		rRoundRect(port,&outset,0x60000,0x60000);
		outset.top = subu(bounds.top<<16,0x30000-0x8000);
		outset.left = subu(bounds.left<<16,0x30000-0x8000);
		outset.bottom = addu(bounds.bottom<<16,0x30000+0x8000);
		outset.right = addu(bounds.right<<16,0x30000+0x8000);
		port->opColor = system_colors[_system_gray67];
		rRoundRect(port,&outset,0x40000,0x40000);
		rRestorePen(port,&oldstate,1);
		RGBFore(port,&oldfg);
		port->opColor = oldop;
	}
	return r;
}

int wDraw1Control(ctlObj *ctl)
{
	if (!ctl)
		return -1;
	if (!ctl->visible || ctl->rtn(ctl,wcMsgDraw,0)<=0)
		return 0;
	update_screen_window_rect(ctl->wind,ctl->bounds.left-2,ctl->bounds.top-2,
		ctl->bounds.right+2,ctl->bounds.bottom+2);
	return 1;
}

static DEF_CTL_RTN(nullProc)
{
	if (msg==wcMsgType)
		return -1;
	if (msg+0u>wcMsgWTest)
		return -8;
	return ctlGeneral(ctl,msg,data);
}

ctlRoutine	f_ctlr[] =
{
	nullProc,
	pushBtnProc, checkBoxProc, radioBtnProc, scrollBarProc,
	popUpProc, pushBtnProc, textBoxProc, textBoxProc,
	pictProc, iconProc
};

ctlObj *wNewControl(windowRef wind,
	const rect16 *bounds,
	const uint8 *title,
	boolean visible,
	int32 value,
	int32 min,
	int32 max,
	int type,
	long int refCon)
{
	ctlObj		**sctl, *ctl;

	if (!wind || type+1u>GET_ARRAY_COUNT(f_ctlr))
		return NULL;
	for(sctl=&wind->ctlList; *sctl; sctl=&(**sctl).next)
		;
	if ((ctl=calloc(sizeof(ctlObj),1))==NULL)
		return NULL;
	*sctl = ctl;
	ctl->wind = wind;
	if (title)
		memcpy(ctl->title,title,*title+1);
	ctl->bounds = *bounds;
	ctl->value = value;
	ctl->min = min;
	ctl->max = max;
	ctl->refCon = refCon;
	ctl->txFont = ffDefault;
	ctl->rtn = f_ctlr[type+1];
	if (type==ctlTypeScrollBar)
		ctl->trk = sBarTrack;
	else if ((type&~1)==ctlTypeStaticText)
	{
		rect16		mbounds = *bounds;

		if (type==ctlTypeEditText)
			OutsetRect16(&mbounds,2,2);
		ctl->data = tBoxNew(wind,&mbounds,0,0,(tBoxAutoScroll|tBoxActive|tBoxModifiable
			|(type&1 ? tBoxFrame:0))^value>>16);
		if (type!=ctlTypeEditText)
		{
			mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 0;
			tBoxSetMargins(ctl->data,&mbounds);
		}
	}
	if ((ctl->visible=visible)!=0)
		ctl->rtn(ctl,wcMsgDraw,0);
	return ctl;
}

int wDisposeControl(ctlObj *ctl)
{
	rect16		mbounds;
	windowRef	wind;
	int		r;

	if (!ctl || !ctl->rtn)
		return -1;
	r = 0;
	if ((wind=ctl->wind)!=NULL)
	{
		ctlObj		**sctl;

		for(sctl=&wind->ctlList; *sctl; sctl=&(**sctl).next)
			if (*sctl==ctl)
				{*sctl = ctl->next, ++r; break;}
	}
	if (ctl==wind->ctlFocus)
		wind->ctlFocus = NULL;
	if (ctl==wind->ctlDefault)
		wind->ctlDefault = NULL;
	if (ctl==wind->ctlCancel)
		wind->ctlCancel = NULL;
	if (ctl->data)
		ctl->rtn(ctl,wcMsgDispose,0);
	mbounds = ctl->bounds;
	OutsetRect16(&mbounds,2,2);
	InvalWindowRect(ctl->wind,&mbounds);
	ctl->rtn = NULL;
	ctl->next = NULL;
	ctl->wind = NULL;
	free(ctl);
	return r;
}

int wKillControls(windowRef wind)
{
	int		r;

	if (!wind)
		return -1;
	r = 0;
	while(wind->ctlList)
		r+= wDisposeControl(wind->ctlList);
	return r;
}

int wIdleControls(windowRef wind)
{
	pt2d16		loc;
	ctlObj		*ctl;

	if (!wind)
		return -1;
	get_mouse_loc(&loc.x,&loc.y);
	if ((ctl=wind->ctlFocus)!=NULL && ctl->hilite!=255
			&& (ctl->rtn(ctl,wcMsgType,NULL)&~1)==ctlTypeStaticText)
		tBoxIdle((textBoxObj*)ctl->data);
	if (wFindControl(wind,loc,&ctl)>0
			&& (ctl->rtn(ctl,wcMsgType,NULL)&~1)==ctlTypeStaticText
			&& ((textBoxObj*)ctl->data)->flags&tBoxModifiable
			&& ptInRect16(&((textBoxObj*)ctl->data)->bounds,loc))
	{
#ifdef TEXT_DRAGDROP
		SLRgn		*hRgn = tBoxSelRgn((textBoxObj*)ctl->data);

		set_mouse_cursor(ptInRgn(hRgn,loc.x,loc.y)<=0,0);
		free(hRgn);
#else
		set_mouse_cursor(1,0);
#endif
		return 1;
	}
	set_mouse_cursor(0,0);
	return 0;
}

ctlObj *wControlEvent(windowRef wind,
	const eventRecord *event)
{
	ctlObj		*ctl = NULL;

	if (!wind || !event)
		return NULL;
	switch(event->what)
	{
	case nullEvent:
		wIdleControls(wind);
		return NULL;
	case mouseUp:
		if ((ctl=wind->ctlFocus)!=NULL)
			if ((ctl->rtn(ctl,wcMsgType,NULL)&~1)
					==ctlTypeStaticText && event->message&wheelFlag)
				tBoxScroll((textBoxObj*)ctl->data,event->message&wheelDownFlag
					? ctlWheelScrollCount:-ctlWheelScrollCount,0);
			else
				ctl = NULL;
		break;
	case mouseDown:
	{
		pt2d16		loc = event->where;
		int		type;

		global_to_local_pt(&loc);
		if (wFindControl(wind,loc,&ctl)<=0)
			{ctl = NULL; break;}
		type = ctl->rtn(ctl,wcMsgType,NULL)&~1;
		if (ctl!=wind->ctlFocus && type==ctlTypeStaticText)
		{
			if (wind->ctlFocus && (wind->ctlFocus->rtn(wind->ctlFocus,wcMsgType,NULL)&~1)
					==ctlTypeStaticText)
				tBoxSelect(wind->ctlFocus->data,0,0);
			wind->ctlFocus = ctl;
		}
		if (!ctl->rtn(ctl,event->message&dblClickFlag
				? wcMsgDblClick:wcMsgClick,&loc))
			{ctl = NULL; break;}
		if (wind->ctlFocus && wind->ctlDefault && wind->ctlFocus==ctl
				&& event->message&dblClickFlag
				&& type==ctlTypeStaticText
				&& ((textBoxObj*)ctl->data)->flags&tBoxListSelect)
			goto defo;
		break;
	}
	case keyDown:
	case autoKey:
		switch((uint8)event->message)
		{
		case kEnterCharCode:
			goto defo;
		case kEscapeCharCode:
			ctl = wind->ctlCancel;
			goto anim;

		default:
			if ((ctl=wind->ctlFocus)!=NULL
					&& (ctl->rtn(ctl,wcMsgType,NULL)&~1)!=ctlTypeStaticText)
				ctl = NULL;
			else
			{
				if (ctl && ((textBoxObj*)ctl->data)->flags&tBoxListSelect
						&& (uint8)event->message==kReturnCharCode)
				{
				defo:	ctl = wind->ctlDefault;
				anim:	if (!ctl)
						break;
					wHiliteControl(ctl,1);
					wDelay(NATIVE_TICKS_PER_SECOND/12);
					wHiliteControl(ctl,0);
					wDelay(NATIVE_TICKS_PER_SECOND/12);
				}
				if (ctl)
					tBoxKey((textBoxObj*)ctl->data,(uint8)event->message,
						event->modifiers);
			}
		}
		break;
	case updateEvt:
		wBeginUpdate((windowRef)event->message);
		wDrawControls((windowRef)event->message);
		wEndUpdate((windowRef)event->message);
	}
	return ctl;
}

/* ---- basic setting routines */

int wSetControlBounds(ctlObj *ctl,
	const rect16 *bounds)
{
	textBoxObj	*tbox;
	int		type;

	if (!ctl || !bounds)
		return -1;
	if (!memcmp(&ctl->bounds,bounds,sizeof(bounds)))
		return 0;
	ctl->bounds = *bounds;
	type = ctl->rtn(ctl,wcMsgType,0);
	if ((type&~1)==ctlTypeStaticText && (tbox=ctl->data)!=NULL)
	{
		rect16		mbounds;

		tbox->bounds = *bounds;
		if (type==ctlTypeEditText)
		{
			OutsetRect16(&tbox->bounds,2,2);
			mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 2;
		} else
			mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 0;
		tBoxSetMargins(tbox,&mbounds);
	}
	return 1+wDraw1Control(ctl);
}

int wSetControlValue(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	value = PIN(value,ctl->min,ctl->max);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wSetControlMinimum(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	ctl->min = value;
	value = PIN(ctl->value,value,ctl->max);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wSetControlMaximum(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	ctl->max = value;
	value = PIN(ctl->value,ctl->min,value);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wHiliteControl(ctlObj *ctl,
	uint8 hilite)
{
	if (!ctl)
		return -1;
	if (hilite==ctl->hilite)
		return 0;
	ctl->hilite = hilite;
	return 1+wDraw1Control(ctl);
}

int wSetControlTitle(ctlObj *ctl,
	const uint8 *title)
{
	if (!ctl)
		return -1;
	if (!memcmp(ctl->title,title,*title+1))
		return 0;
	memcpy(ctl->title,title,*title+1);
	return 1+wDraw1Control(ctl);
}
