/*
 mevent.c

   Revised: 2015.01.07
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.22
  Formerly mtimer, moved event code implemented in spanner.c & cseries here.
2014.08.21
  The Microseconds equivalent version is producing results 1.00247 times
faster than GetTickCount()*1000.
2015.01.07
  Turns out GetTickCount is the one returning an inaccurate value
(see perftime).
*/

#include <process.h>

#define USE_HIGH_RES_TIMER	1
#define USE_WINEVENT_TIMER	0

#if defined(WIN32) || defined(WIN64)
 #include <windows.h>
 #define ThdMutexLock(t)	EnterCriticalSection(&(t)->criticalSection)
 #define ThdMutexUnlock(t)	LeaveCriticalSection(&(t)->criticalSection)
 #define EvtMutexLock()		EnterCriticalSection(&eventSection)
 #define EvtMutexUnlock()	LeaveCriticalSection(&eventSection)
#endif

#include "mevent.h"
#include "mwind.h"

#if 0 /* same precision as GetTickCount, slower */
 #include <time.h>
 #define milliseconds()		clock()
#elif USE_WINEVENT_TIMER
 #define milliseconds()		timeGetTime()
#else
 #define milliseconds()		GetTickCount()
#endif

/* ---- constants */

/* standard EventQueue only carries 20 */
#define EVENTQMASK		31

/* ---- typedefs */

typedef struct timerThd
{
#if defined(WINUSERAPI) && !USE_WINEVENT_TIMER
	CRITICAL_SECTION	criticalSection;
#endif
#if USE_HIGH_RES_TIMER
	int64		ltick;
#else
	int32		ltick;
 #if USE_WINEVENT_TIMER
	HANDLE		ticArr;
	UINT		ticEID;
 #endif
#endif
	uint16		count, maximum;
	timerEntry	list[1];
} timerThd;

/* ---- globals */

WNDCLASS	mEventWClass;

HPALETTE	world_screen_pal;
HBITMAP		world_screen_dib;
HWND		world_screen_window;

eventRecord	mEventState = {0,btnState|rightBtnState};
uint8		mKeyMap[20];
static uint8	deadKey;

static uint8	mVBL, altDown;
static uint8	eventCount, eventWPos, eventRPos;

uint8		mEventSleepToggle, mVBLInitMS = 5; /* 70% */

static eventRecord	eventQueue[EVENTQMASK+1];

#ifdef WINUSERAPI
static CRITICAL_SECTION	eventSection;
#endif

/* ---- timer */

#if !TARGET_OS_MAC
uint32 TickCount(void)
{
/*	MSG		msg;

	if (PeekMessage(&msg,world_screen_window,0,0,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}*/
	return mEventState.when;
}
#endif

int PrimeTimer(timerThread *thd,
	timerEntry *entry,
	int64 mcs)
{
	int64		n;
	timerThd	*tdata = (timerThd*)thd;
	int		r;

	if (!tdata->count || !tdata->maximum)
		return -1;
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	n = entry->mcsremaining;
	if (!int64n(n))
		{r = 0; goto end;} /* already primed */
	r = 1;
	(void)int64add(n,mcs);
	if (int64n(n))
		int64xset(n,0);
	entry->mcsremaining = n;
end:
#if !USE_WINEVENT_TIMER
	ThdMutexUnlock(tdata);
#endif
	return r;
}

timerEntry *InstallTimer(timerThread *thd,
	timerRoutine rtn)
{
	timerThd	*tdata;
	timerEntry	*t;

	if (!rtn || (tdata=(timerThd*)thd)==NULL
			|| tdata->count>=tdata->maximum)
		return NULL; /* bad param or maxed out */
	/* probably no need to lock; the count is incremented last */
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	t = tdata->list+tdata->count;
	int64xset(t->mcsremaining,-1);
	t->routine = rtn;
	++tdata->count;
#if !USE_WINEVENT_TIMER
	ThdMutexUnlock(tdata);
#endif
	return t;
}

int RemoveTimer(timerThread *thd,
	timerEntry *entry)
{
	timerThd	*tdata;
	timerEntry	*t;
	int32		c;

	if (!entry || (tdata=(timerThd*)thd)==NULL
			|| (c=tdata->count)==0 || !tdata->maximum)
		return -1;
	t = tdata->list;
	do
		if (t==entry)
		{
		#if !USE_WINEVENT_TIMER
			ThdMutexLock(tdata);
		#endif
			--tdata->count, ++t;
			memcpy(t-1,t,--c*sizeof(*t));
		#if !USE_WINEVENT_TIMER
			ThdMutexUnlock(tdata);
		#endif
			return 1;
		}
	while(++t,--c);
	return 0;
}

#if defined(__MT__) || USE_WINEVENT_TIMER
static inline
#endif
void timerThreadIdle(timerThread *thd)
{
	int32		c;
	timerThd	*tdata = (timerThd*)thd;
#if USE_HIGH_RES_TIMER
	int64		mcs, mcsd;
#else
	int32		mcsd;
 #if !USE_WINEVENT_TIMER
	int32		tkc;
 #endif
#endif
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	if ((c=tdata->count)==0)
		goto end;
#if USE_HIGH_RES_TIMER
	Microseconds((uint64*)&mcs);
	mcsd = mcs;
	(void)int64sub(mcsd,tdata->ltick);
	tdata->ltick = mcs;
	if (int64p(mcsd))
#elif !USE_WINEVENT_TIMER
	mcsd = tkc = milliseconds();
	mcsd-= tdata->ltick;
	tdata->ltick = tkc;
	if (mcsd>0)
#else
	mcsd = 1;
#endif
	{
		timerEntry	*t = tdata->list;
	#if !USE_HIGH_RES_TIMER
		mcsd*= 1000;
	#endif
		do
			if (!int64n(t->mcsremaining))
			{
				(void)int64sub(t->mcsremaining,mcsd);
				if (int64n(t->mcsremaining))
					t->routine(t);
			}
		while(++t,--c);
	}
end:
#if USE_WINEVENT_TIMER
	;
#else
	ThdMutexUnlock(tdata);
#endif
}

#if USE_WINEVENT_TIMER
static void CALLBACK ticEvtFunc(UINT id,
	UINT msg,
	DWORD_PTR user,
	DWORD_PTR dw1,
	DWORD_PTR dw2)
{
	timerThd	*tdata = (timerThd*)user;

	timerThreadIdle((timerThread*)tdata);
	SetEvent(tdata->ticArr);
}
#elif defined(__MT__)
static void timerThreadFunc(void *param)
{
	timerThd	*tdata = param;
//	int8		toggle = 0;
  #if USE_HIGH_RES_TIMER
	Microseconds((UnsignedWide*)&tdata->ltick);
  #else
	tdata->ltick = milliseconds()+1;
  #endif
	while(tdata->maximum)
	{
		timerThreadIdle(param);
		Sleep(1);
	}
	_endthread();
}
#endif

timerThread *CreateTimerThread(int maxCount)
{
	timerThd	*tdata;
#if defined(__MT__) && !USE_WINEVENT_TIMER
	int		r;
#endif
	if (maxCount-1>65534u || (tdata=calloc(sizeof(timerThd)
			+sizeof(timerEntry)*(maxCount-1),1))==NULL)
		return NULL;
	tdata->maximum = maxCount;
#if USE_WINEVENT_TIMER
	tdata->ltick = milliseconds()+1;
	tdata->ticArr = CreateEvent(NULL,FALSE,FALSE,NULL);
	tdata->ticEID = timeSetEvent(1,0,ticEvtFunc,(DWORD_PTR)tdata,TIME_PERIODIC);
#else
	InitializeCriticalSection(&tdata->criticalSection);
 #if defined(__MT__)
	if ((r=_beginthread(timerThreadFunc,0,tdata))==-1)
	{
		DeleteCriticalSection(&tdata->criticalSection);
		free(tdata);
		return NULL;
	}
	/* Setting to 'highest' from 'above normal' fixes some inaccuracies.
	  Doing things with the Task Manager may affect the timer,
	  but moving generic windows (including our app's) around no longer does.
	*/
	SetThreadPriority((HANDLE)r,THREAD_PRIORITY_HIGHEST);
 #elif USE_HIGH_RES_TIMER
	Microseconds((UnsignedWide*)&tdata->ltick);
 #else
	tdata->ltick = milliseconds()+1;
 #endif
#endif
	return (timerThread*)tdata;
}

int EndTimerThread(timerThread *thd)
{
	timerThd	*tdata;

	if ((tdata=(timerThd*)thd)==NULL || !tdata->maximum)
		return -1;
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	/* this should quickly kill the thread from inside */
	tdata->count = tdata->maximum = 0;
#if USE_WINEVENT_TIMER
	timeKillEvent(tdata->ticEID);
	if (tdata->ticArr)
		CloseHandle(tdata->ticArr);
	timeEndPeriod(0);
#else
	ThdMutexUnlock(tdata);
	Sleep(1);
	DeleteCriticalSection(&tdata->criticalSection);
#endif
	free(tdata);
	return 0;
}

void mOSVBL_timer(timerEntry *this)
{ /* 16625 mcs (133/8 ms), the Classic Mac OS tick rate - see my_cseries.h */
	++mEventState.when;
	(void)int64addu32(this->mcsremaining,SMALL_TICK_TIME);
	if (int64n(this->mcsremaining))
		int64xset(this->mcsremaining,0);
	mVBL = mVBLInitMS;
}

/* ---- events */

boolean mouse_still_down(void)
{
	MSG		msg;

	while(PeekMessage(&msg,world_screen_window,WM_MOUSEMOVE,WM_MBUTTONDBLCLK,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	if (!(mEventSleepToggle+=0x40))
		Sleep(1);
	global_idle_proc();
	return !((mEventState.modifiers&mEventState.modifiers*2)&btnState);
}

#define MODKEYS_6L	(1<<(MKV_MOD&7)|1<<(MKV_LALT&7)|(1<<(MKV_LSHIFT&7) \
	|1<<(MKV_CAPSLOCK&7)|1<<(MKV_LCONTROL&7) \
	|1<<(MKV_RSHIFT&7)|1<<(MKV_RALT&7) \
	|1<<(MKV_RCONTROL&7))<<8|(1<<(MKV_RGUI&7))<<16)

uint8 *MGetKeys(uint8 outmap[20])
{
	uint8		vkmap[256];
	MSG		msg;

	if (!outmap)
		return NULL;
	while(PeekMessage(&msg,world_screen_window,WM_KEYDOWN,WM_SYSKEYUP,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	/* only GetKeyboardState will deliver independent left/right modifier keys */
	/* the problem: keys like Enter and Return are inseparable */
	/* key up/down event reading, done above, is the only way to get true Return/Enter */
	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	if (GetKeyboardState(vkmap))
	{
		if (vkmap[VK_LSHIFT]>>1)
			_MKSET(mKeyMap,MKV_LSHIFT);
		if (vkmap[VK_RSHIFT]>>1)
			_MKSET(mKeyMap,MKV_RSHIFT);
		if (vkmap[VK_CAPITAL]&1) /* sticky capsLock */
			_MKSET(mKeyMap,MKV_CAPSLOCK);
		if (vkmap[VK_LMENU]>>1)
			_MKSET(mKeyMap,MKV_LALT);
		if (vkmap[VK_RMENU]>>1)
			_MKSET(mKeyMap,MKV_RALT);
		if (GetKeyState(VK_LCONTROL)>>1)
			_MKSET(mKeyMap,MKV_LCONTROL);
		if (GetKeyState(VK_RCONTROL)>>1)
			_MKSET(mKeyMap,MKV_RCONTROL);
	}
	memcpy(outmap,mKeyMap,20);
	return outmap;
}

/* Do Not add PeekMessage/GetMessage here; mWndProc calls this */
uint16 updateEventState(void)
{
	uint16		state = _MKDOWN(mKeyMap,MKV_OPT_APPS) ? optionKey:0;

	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	if (GetKeyState(VK_SHIFT)>>1)
	{
		state|= shiftKey;
		if (GetKeyState(VK_LSHIFT)>>1)
			_MKSET(mKeyMap,MKV_LSHIFT);
		if (GetKeyState(VK_RSHIFT)>>1)
			_MKSET(mKeyMap,MKV_RSHIFT);
	}
	if (GetKeyState(VK_CAPITAL)&1) /* sticky capsLock */
		state|= capsLock, _MKSET(mKeyMap,MKV_CAPSLOCK);
	if (GetKeyState(VK_CONTROL)>>1)
	{
		state|= controlKey;
		if (GetKeyState(VK_LCONTROL)>>1)
			_MKSET(mKeyMap,MKV_LCONTROL);
		if (GetKeyState(VK_RCONTROL)>>1)
			_MKSET(mKeyMap,MKV_RCONTROL);
	}
	if (altDown)
		state|= altKey;
	if (altDown || GetKeyState(VK_MENU)>>1)
	{
		if (GetKeyState(VK_LMENU)>>1)
			_MKSET(mKeyMap,MKV_LALT);
		if (GetKeyState(VK_RMENU)>>1)
			_MKSET(mKeyMap,MKV_RALT);
	}
	mEventState.modifiers&= btnState|rightBtnState;
	return mEventState.modifiers|= state;
}

static boolean evtDeQueue(eventRecord *event,
	uint32 mask)
{
	int		r, c, i = eventRPos;

	if ((c=eventCount)==0 || !mask)
		return 0;
	EvtMutexLock();
	r = 0;
	do
	{
		eventRecord	*a = eventQueue+i;

		i = (i+1)&EVENTQMASK;
		if (1u<<a->what&mask)
		{
			*event = *a;
			eventWPos&= EVENTQMASK;
			if (--eventCount)
			{ /* remove from list */
				do
				{
					eventRecord	*n = eventQueue+i;

					i = (i+1)&EVENTQMASK;
					*a = *n, a = n;
				} while(--c);
			}
			--eventWPos;
			a->what = 0;
			r = 1;
			break;
		}
	} while(--c);
	EvtMutexUnlock();
	return r;
}

int PostMEvent(const eventRecord *event)
{
	int		r;

	if (!event)
		return -1;
	EvtMutexLock();
	r = 1;
	if (eventCount<=EVENTQMASK)
		++eventCount;
	else if (event->what==nullEvent) /* full - ignore excess null events */
		{r = 0; goto end;}
	else
		++eventRPos, eventRPos&= EVENTQMASK; /* otherwise overwrite old */
	eventQueue[eventWPos&EVENTQMASK] = *event, ++eventWPos;
end:	EvtMutexUnlock();
	return r;
}

boolean GetNextEvent(uint32 mask,
	eventRecord *event)
{
	MSG		msg;
	boolean		had = FALSE;

	while(PeekMessage(&msg,world_screen_window,0,0,PM_REMOVE))
	{
		had = TRUE;
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	if (!event)
		goto end;
	if (evtDeQueue(event,mask))
		return 1;
	*event = mEventState;
	event->what = nullEvent;
end:	if (mVBL && !(mEventSleepToggle+=0x20))
		{--mVBL; Sleep(1);}
	return 0;
}

/* 2014.07.30 */
void FlushEvents(uint32 mask,
	uint32 stopMask)
{
	eventRecord	event;
	/* appears to be a negation mask (idk) */
	for(mask&=~stopMask; evtDeQueue(&event,mask); )
		;
}

static LRESULT CALLBACK mWndProc(HWND window,
	UINT what,
	WPARAM wParam,
	LPARAM lParam)
{
	static boolean	hasleft;
	int		r;

	switch(what)
	{
	case WM_CREATE:
		break;
	case WM_CLOSE:
	case WM_QUIT:
		mEventState.what = closeEvent;
		PostMEvent(&mEventState);
		break;
	//case WM_PALETTECHANGED: -- system colors changed
	//	break;
	case WM_PAINT:
		{
			PAINTSTRUCT	ps;

			BeginPaint(window,&ps);
			if (screen_desktop_window.port.map)
				update_screen_rect(0,0,
					screen_desktop_window.port.map->w,
					screen_desktop_window.port.map->h);
			EndPaint(window,&ps);
		}
		break;
	case WM_NCHITTEST:
		if ((r=DefWindowProc(window,what,wParam,lParam))!=HTCLIENT)
			hasleft = 1;
		return r;
	case WM_NCMOUSEMOVE:
		hasleft = 1;
		goto norm;
	case WM_MOUSEMOVE:
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		if ((~mEventState.modifiers&(rightBtnState|btnState))
				&& !(wParam&(MK_LBUTTON|MK_RBUTTON)))
		{
			mEventState.message = 0;
			mEventState.modifiers|= rightBtnState|btnState;
			mEventState.what = mouseUp;
			PostMEvent(&mEventState);
		}
		hasleft = 0;
		break;		/* We need to set the cursor */
	case WM_SETCURSOR:
		if (LOWORD(lParam)==HTCLIENT)
		{
			SetCursor(mEventWClass.hCursor);
			return TRUE;
		}
		goto norm;
	//case WM_CHAR:
	//case WM_DEADCHAR: -- dead keys are accent modifiers (e.g., umlaut)
	//case WM_SYSDEADCHAR:
	case WM_ERASEBKGND: /* Basilisk II: to prevent flashing */
		return 1;
	/* MSDN: be sure to return TRUE if processing XBUTTON events */
	case WM_RBUTTONDBLCLK:
	case WM_RBUTTONDOWN:
		mEventState.message = what==WM_RBUTTONDBLCLK ? dblClickFlag:0;
		mEventState.modifiers&=~rightBtnState;
		goto mdn;
	case WM_LBUTTONDBLCLK:
	case WM_LBUTTONDOWN:
		mEventState.message = what==WM_LBUTTONDBLCLK ? dblClickFlag:0;
		mEventState.modifiers&=~btnState;
	mdn:	mEventState.what = mouseDown;
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		PostMEvent(&mEventState);
		break;
	case WM_NCRBUTTONUP:
	case WM_RBUTTONUP:
		mEventState.message = 0;
		mEventState.modifiers|= rightBtnState;
		goto mup;
	case WM_NCLBUTTONUP:
	case WM_LBUTTONUP:
		mEventState.message = 0;
		mEventState.modifiers|= btnState;
		goto mup;
	case WM_MOUSEWHEEL:
		mEventState.message = (wParam>>23&wheelDownFlag)|wheelFlag;
	mup:	mEventState.what = mouseUp;
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		PostMEvent(&mEventState);
		break;
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
	case WM_SYSKEYUP:
	case WM_KEYUP:
	{ /* need to trap the ALT and WIN keys! .. no success on WIN keys */
		uint8		mvk, down = (uint8)(1&~what), kchar, dchar;

		if (wParam+0u>=sizeof(mkvtab_win) || (mvk=mkvtab_win[wParam])>=MKC)
			break;
		if (down && lParam&0x40000000)
		{ /* auto-repeat */
			mEventState.what = autoKey;
			goto kchr;
		}
		if (!(lParam&0x1000000))
			switch(wParam)
			{ /* numlock-off numpad translation */
			case VK_CLEAR: mvk = MKV_KP5; break;
			case VK_PRIOR: mvk = MKV_KP9; break;
			case VK_NEXT: mvk = MKV_KP3; break;
			case VK_END: mvk = MKV_KP1; break;
			case VK_HOME: mvk = MKV_KP7; break;
			case VK_LEFT: mvk = MKV_KP4; break;
			case VK_RIGHT: mvk = MKV_KP6; break;
			case VK_UP: mvk = MKV_KP8; break;
			case VK_DOWN: mvk = MKV_KP2; break;
			case VK_INSERT: mvk = MKV_KP0; break;
			case VK_DELETE: mvk = MKV_KPDECIMAL;
			}
		else if (wParam==VK_RETURN)
			mvk = MKV_ENTER;
		altDown = (what&~1)==WM_SYSKEYDOWN && lParam&0x20000000;
		if (down)
			_MKSET(mKeyMap,mvk);
		else
			_MKCLR(mKeyMap,mvk);
		updateEventState();
		/* as an EventRecord post: state&=~eventModifiedSwitch(mvk); */

		mEventState.what = keyUp-down;
	kchr:	dchar = 0;
		if ((kchar=mKCHRproc(mvk,mEventState.modifiers,
				mEventState.what==keyDown ? &deadKey:NULL,
				mEventState.what==keyDown ? &dchar:NULL))==0)
			mEventState.what = nullEvent;
		else
		{
			if (dchar)
			{
				mEventState.message = dchar|KBL<<8;
				PostMEvent(&mEventState);
			}
			mEventState.message = mvk<<8|kchar;
			PostMEvent(&mEventState);
		}
	case WM_HOTKEY:
	//case WM_CONTEXTMENU: /* if you want to disable contextual menu on titlebar */
		break;
	}
	case WM_ACTIVATEAPP:
		mEventState.message = (wParam&resumeFlag)|suspendResumeMessage<<24;
		mEventState.what = osEvt;
		memset(mKeyMap,0,sizeof(mKeyMap));
		PostMEvent(&mEventState);
	case WM_SYSCOMMAND:
		if ((wParam&0xFFF0)==SC_KEYMENU) /* alt-space et al. */
			break;
	case WM_NULL:
	default:
	norm:	return DefWindowProc(window,what,wParam,lParam);
	}
	return 0;
}

int wRegisterDev(HINSTANCE devInstance,
	HICON icon,
	HCURSOR curs,
	HBRUSH bgBrush,
	TCHAR *name)
{
	static boolean	init;

	if (!name)
		return -1;
	mEventWClass.style = CS_OWNDC|CS_DBLCLKS;
	mEventWClass.lpfnWndProc = mWndProc;
	mEventWClass.cbWndExtra = mEventWClass.cbClsExtra = 0;
	mEventWClass.hInstance = devInstance;
	mEventWClass.hIcon = icon;
	mEventWClass.hCursor = curs;
	mEventWClass.hbrBackground = bgBrush;
	mEventWClass.lpszMenuName = NULL;
	mEventWClass.lpszClassName = name;

	if (!init)
		{InitializeCriticalSection(&eventSection); init = 1;}
	return !!RegisterClass(&mEventWClass);
}
