/*
 mfont.c
  (minimal) font loading, bitmapped depths 1, 2, 4 and 8

  Created: 2014.02.28 12:43 EST
   Author: Adam P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.29
  Bitmapped font loading now works.  Update: calloc size was incorrect.
2014.04.11
  Loading by ID done by cast of name param as a int16.
2014.04.17
  Added spExtra, made x fixed-point.
2014.04.21
  Fixed underlining bug that wrote outside destination buffer.
2014.05.20
  Zero width for -1 (should widTable be int8?).
*/

#include <enviro64.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_cseries.h"
#include "../resex/resex.h"
#include "mtext.h"

/* ---- constants */

/* system font characters */
#define MacSysCharCommand	17
#define MacSysCharCheck		18
#define MacSysCharDiamond	19
#define MacSysCharApple		20

/* a bit of specs from Fonts.r & Fonts.h */
/* these constants & structures summarize, but aren't used at this level of rendering */

/* map flags */
#define famFixedWidth		0x80 /* otherwise, proportional */
#define famIgnoreFractWidths	0x40
#define famIntegralExtra	0x20 /* otherwise, fixed-point */
#define famIgnoreFractEnable	0x10
#define famNoSpacingAdjust	0x8 /* otherwise, adjust char. spacing */
/* reserved flags */
/* table flags */
#define tblHasCharWidths	0x2
#define tblHasImageHeights	0x1

/* ---- typedefs */

typedef struct fontAssc/*(6)*/
{
	int16		size;
	int16		style;
	int16		ID;
} fontAssc;

#include <pshpack2.h>

typedef struct fontFamilyDesc
{
	uint8		famFlags, tblFlags;
	int16		familyID;
	uint16		firstChar, lastChar;
	int16		ascent, descent;
	int16		leading, maxWidth;
	/* offsets */
/*x10*/	int32		widthsTableOffset;
	int32		kerningTableOffset;
	int32		styleTableOffset;
	int16 rsv1;
	int16		exWidBold;
/*x20*/	int16		exWidItalic;
	int16		exWidUnderline;
	int16		exWidOutline;
	int16		exWidShadow;
	int16		exWidCondensed;
	int16		exWidExtended;
	int16 rsv2;
	int32 rsvIntl;
	int16		version;
	/* font association table */
/*x34*/	uint16		asscCount;
	fontAssc	assc[1];
	/* 32-bit offset table if version 2 */
	/* bounding box table: zcount.w style.w, Rect */
} fontFamilyDesc;

#include <poppack.h>

#define bmtNoFontExpansion	0x40
#define bmtFixedWidth		0x20 /* otherwise, proportional */
#define bmtColor		0x2 /* otherwise, black on white */
#define bmtSynthetic		0x1

#define bmfHasfctb		0x80 /* has font color table */
#define bmfDepthMask		0xC /* bpp = 1<<x */
#define bmfHasCharWidths	0x2
#define bmfHasImageHeights	0x1

#define bmfValidType(v)		((v&0x9C)==0x90)
#define bmfValidFlag(v)		!(v&0x70)

typedef struct bmFont
{
	uint8		type; /* required signature: 1xx100xx */
	uint8		flag; /* required signature: x000xxxx */
	uint16		firstChar, lastChar;
	int16		maxWidth, maxKerning;
	int16		negDescent;
	int16		bmWidth, bmHeight;
	/* docs say negDescent is added as high-word if != -1, but really if >0 */
	uint16		OWTLoc;

	int16		ascent, descent, leading;
	int16		wordPitch;
	int16		bm[1];
	/* directly after is the loc. table (lastChar-firstChar+3 ct) */
	/* OWTLoc to offset/width table, then optional char widths/img heights */
} bmFont;

/* ---- globals */

ffMemory	*fmFonts[256];
uint8		ffDefault;

/* ---- code */

static void b2to8(uint8 *d,
	const uint8 *s,
	uint32 c)
{
	do
	{
		uint8		v = *s++;

		*d = v>>6;
		d[1] = v>>4&3;
		d[2] = v>>2&3;
		d[3] = v&=3;
	} while(d+=4,--c);
}

static void b4to8(uint8 *d,
	const uint8 *s,
	uint32 c)
{
	uint8		v;

	//if (c&1)
	//	v = *s, *d = !!(v&0xF0), d[1] = !!(v&=15), d+= 2, ++s;
	if (c>>=1)
		do
		{
			v = *s, *d = v&0xF0, d[1] = v&=15;
			v = s[1], d[2] = v&0xF0, d[3] = v&=15;
		} while(d+=4,s+=2,--c);
}

/* loading */

/* some code here initially based on code found in AlephOne/RenderOther/sdl_fonts.cpp */

static int loadFont(bmfMemory *f,
	const uint8 *_s,
	long z)
{
	const uint8	*s;
	uint32		o, v, c, allocsize;
	uint8		u;
	/* pack flags */
	if ((s=_s)==NULL)
		return -1;
	u = s[1];
	if (!bmfValidType(*s) || !bmfValidFlag(u)
			|| z<=OFFSET(bmFont,bm))
		return 0;
	f->flag = (*s&0x60)|(u&bmfHasfctb)|(u<<2&12);
	s+= 2;
	f->firstChar = va_be16(s);
	f->lastChar = va_be16(s);
	if (f->lastChar<f->firstChar)
		return 0;
	f->maxWidth = va_be16(s);
	f->maxKerning = va_be16(s);
	v = (int16)va_be16(s);
	f->bmWidth = va_be16(s);
	f->bmHeight = va_be16(s);
	o = va_be16(s);
	if ((signed)v>=0) //!=-1u)
		o|= v<<16;
	o = o*2+16; /* actual offset is word count from OWT field */
	if (o>=z)
		goto err;
	f->ascent = va_be16(s);
	f->descent = va_be16(s);
	f->leading = va_be16(s);
	v = va_be16(s)*2;
	u = u>>2&3;
	f->pitch = v*8u>>u;
	f->bmSize = (f->bmHeight*f->pitch+3)&~3;
	c = f->lastChar-f->firstChar+3;
	/* 2B ea. for the location table (a int16 list) */
	/* plus 2x1B ea. for the width-pair table */
	allocsize = f->bmSize+c*(sizeof(int16)*3);
	if (f->bmSize<=0 || (f->data=calloc(1,allocsize))==NULL)
	{
	err:	f->bmHeight = f->bmWidth = 0;
		return f->bmSize>0 ? -2:0;
	}
	v*= f->bmHeight;
	switch(u)
	{
	case 0:	b1to8(f->data,s,v);
		break;
	case 1:	b2to8(f->data,s,v);
		break;
	case 2:	b4to8(f->data,s,v);
		break;
	case 3: memcpy(f->data,s,v);
	}
	s+= v;
	{ /* loc table, wid table; */
		int16		*d = (int16*)(f->bmSize+f->data);

		do
			*d++ = (int16)va_be16(s);
		while(--c);
		f->widTable = d;
		if ((v=(f->lastChar-f->firstChar+3)*2)>0)
		{
			s = _s+o;
			do
				*d++ = *(const int8*)s;
			while(++s,--v);
		}
	}
	return 1;
}

ffMemory *ffLoad(ResFile_Memory *rMap,
	const uint8 *name,
	int ID)
{
	mhandle		ffh, h;
	ffMemory	*ff;
	int8		*s;
	uint16		c, i;

	if ((ffh=ResFile_GetResource(rMap,MK4CC('F','O','N','D'),name,ID,NULL))==NULL)
		return NULL;
	set_mhandle_lock(ffh,1);
	s = OFFSET(fontFamilyDesc,asscCount)+*ffh;
	if ((c=va_be16(s)+1)==0 || (ff=calloc(OFFSET(ffMemory,font)
			+sizeof(bmfMemory)*c,1))==NULL)
		return NULL;
	ff->asscCount = c, i = 0;
	do
	{
		bmfMemory	*f = ff->font+i;
		int		v;

		f->size = va_be16(s);
		f->style = s[1], s+= 2;
		v = va_be16(s);
		if ((h=ResFile_GetResource(rMap,MK4CC('N','F','N','T'),NULL,v,NULL))==NULL
				&& (h=ResFile_GetResource(rMap,MK4CC('F','O','N','T'),
				NULL,v,NULL))==NULL)
			continue;
		set_mhandle_lock(h,1);
		loadFont(f,(const uint8*)*h,get_mhandle_size(h));
		dispose_mhandle(h);
	} while(++i<c);
	dispose_mhandle(ffh);
	return ff;
}

/*
 to emulate QD Text, we'd have to find size divisibility for "ideal" scaling
*/
const bmfMemory *ffMatch(ffMemory *ff,
	int size,
	int8 styl)
{
	const bmfMemory	*fm, *fc, *fz;
	int		c, v;

	if (!ff || (c=ff->asscCount)<=0)
		return NULL;
	if (size<=0)
		size = 12;
	fm = ff->font, fz = fc = NULL, styl&= 3;
	do
	{
		v = fm->style&3;
		if (!v && (!fz || ABS(size-fm->size)<=ABS(size-fz->size)))
			fz = fm;
		if (fm->size>=size)
		{
			if (v==styl && fm->size==size)
				return fm;
			/* closest style without going over */
			if (fm->size==size && !(v&~styl))
				fc = fm;
		}
	} while(++fm,--c);
	/* worst case, closest size without style */
	return fc ? fc:fz ? fz:ff->font;
}

int ffFree(ffMemory *ff)
{
	bmfMemory	*fm;
	int		c;

	if (!ff || (c=ff->asscCount)<0)
		return -1;
	if (c)
	{
		fm = ff->font;
		do
			if (fm->data)
				{free(fm->data); fm->data = NULL;}
		while(++fm,--c);
	}
	c = ff->asscCount;
	ff->asscCount = 0;
	free(ff);
	return c;
}

/* global loader */

int fmLoadAll(ResFile_Memory *rmap)
{
	ffMemory	**ff = fmFonts;
	int		i = 0, r = 0;

	do
		r+= !*ff && (*ff=ffLoad(rmap,NULL,i))!=NULL;
	while(++ff,++i<GET_ARRAY_COUNT(fmFonts));
	return r;
}

int fmLoadMaster(const char *name)
{
	ResFile_Memory	*fontRMap;
	FILE		*fontFile;
	int		r = 0;

	if (!name)
		return -1;
	if ((fontFile=fopen(name,"rb"))==NULL)
		return 0;
	if (ResFile_LoadMap(fontFile,0,&fontRMap,0)>=0)
		r = fmLoadAll(fontRMap);
	free(fontRMap);
	fclose(fontFile);
	return r;
}

void fmUnloadMaster(void)
{
	ffMemory	**ff = fmFonts;
	int		c = GET_ARRAY_COUNT(fmFonts);

	do
		if (*ff)
			{ffFree(*ff); *ff = NULL;}
	while(++ff,--c);
}

/* ---- port settings */

int rTextFont(grafPort *p,
	uint16 font)
{
	if (!p)
		return -1;
	p->txFont = font;
	p->txFM = ffMatch(fmFonts[font>=GET_ARRAY_COUNT(fmFonts) ? 0:font],p->txSize,p->txStyle);
	return !!p->txFM;
}

int rTextSize(grafPort *p,
	uint16 size)
{
	if (!p)
		return -1;
	p->txFM = ffMatch(fmFonts[p->txFont>=GET_ARRAY_COUNT(fmFonts) ? 0:p->txFont],
		p->txSize=size,p->txStyle);
	return !!p->txFM;
}

int rTextStyle(grafPort *p,
	uint8 style)
{
	if (!p)
		return -1;
	p->txFM = ffMatch(fmFonts[p->txFont>=GET_ARRAY_COUNT(fmFonts) ? 0:p->txFont],
		p->txSize,p->txStyle=style);
	return !!p->txFM;
}
