/*
 mgraf.h
  limited graphics library, with some initial basis on QuickDraw

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __MGRAF_H
#pragma once

#include "../my_cseries.h"

/* ---- constants */

enum
{ /* pType bits */
	pTypeIndexed,
	pTypeFixed	= 0x1,	/* pTable lock */
	pTypeDirect	= 0x2,	/* no table - gamma ramp applies to device */
	pTypeAlpha	= 0x4,	/* has alpha channel (what's left at top) */
	pTypeLGreen	= 0x40,	/* green channel is 1 bit larger (i.e., 5:6:5) */
	pTypeLSB	= 0x80,	/* values in little endian ordering */

	pTypeFixedBit	= 0,
	pTypeDirectBit,
	pTypeAlphaBit,
	pTypeLGreenBit	= 6,
	pTypeLSBBit
};

enum
{ /* dpi */
	hvResDefault	= 72<<16
};

enum
{ /* pen modes */
	mSrcCopy,		/* no transparency by default */
	mSrcOr,			/* pixel-merge, OR indexed bits or AND direct */
	mSrcXor,		/* inversion */
	mSrcBic,		/* pixel-merge, AND indexed/OR direct */
	mSrcNotCopy,		/* not: invert src; reverse colors for solids/patterns */
	mSrcNotOr,
	mSrcNotXor,
	mSrcNotBic,
	mTransparent,		/* mask out index-zero or bkColor src pixels */
	mBlend,			/* interpolate, given opColor */
	mAdMax,			/* highest of src/dst components */
	mAdMin,			/* lowest of src/dst components */
	mAddPin,		/* add, no higher than channels in opColor */
	mAddOver,		/* add, no clipping */
	mSubPin,		/* subtract, no low than channels in opColor */
	mSubOver,		/* subtract, no clipping */
	mSrcPat		= 0x10,	/* img flag: pattern with mx/mh position repeats */
	mTextGray	= 0x20,	/* text flag: render w. 50% fg/bg blend or gray pattern */
	mDither		= 0x40	/* img flag: dither if dst. can't accurately represent */
};

enum
{ /* color conversion routines */
	ccGray,
	ccDirect,
	ccDirectBS32,
	ccDirectBS16,
	ccNative	= 0x2,
	ccIndexed	= 0x4,	/* passed to remapper, supresses indexed output */

	remapReverse	= 0x8,	/* builds in reverse for logical cts */
	remapForce30	= 0x40	/* forces 10:10:10 output */
};

/* ---- typedefs */

typedef struct rect2dq
{
	ifixed		top, left, bottom, right;
} rect2dq;

/* explicit scanline region (smallest size=empty; rects require redundancies) */
typedef struct SLRgn
{
	int32		size; /* total size, including header */
	rectangle	bBox;
	int16		x[1]; /* lines begin and end with negative space */
} SLRgn;

typedef struct pixMap/*(40+)*/
{ /* aligned like crazy */
	uint16		w, h;
	pt2d16		pOrigin;
	int32		pitch; /* bytes per line */
	uint8		pSize; /* 1, 2, 4, 8, 16 or 32; otherwise, codec type */
	uint8		pType; /* see above */
	uint8		bpp; /* 1-32 */
	uint8		minCmp; /* smallest bit count for a direct component */
/*x10*/	ufixed		hRes, vRes; /* horiz. & vert. ppi resolution */
	int8		*baseAddr; /* access to the pixel data (may be NULL) */
	void		*pHand; /* the applicable mhandle, if used for baseAddr */
	/* start of the applicable tables */
	cTab64		*pTable;
	/* invalidated every time the ct seed changes */
	uint8		*iTable; /* inverse table; minCmp=resolution */
} pixMap;

typedef struct pixPat
{ /* like QD, only the largest power of two contained is valid for the width */
  /* however, here, the height can be any non-zero value */
	uint8		*px; /* 8-bit only */
	void		*t; /* native table (16 or 32 bits, generated on-the-fly) */
	uint16		w, h, pitch;
	cTab64		ct; /* bit #31 of seed = direct/gray format for 't' */
} pixPat;

typedef struct penState
{
	int32		penw, penh;
	pt2dq		pnLoc;
	uint8		pnMode;
	pixPat		*pPat;
} penState;

typedef struct grafPort
{
	pixMap		*map;
	rect16		clip;
	pt2d16		origin;
	uint32		fgSolid, bkSolid;
	rgb48		fgColor, bkColor, opColor;
	pixPat		*pPat;
	ifixed		penw, penh;
	pt2dq		pnLoc;
	uint8		pnMode, txMode, txStyle;
	uint16		txFont, txSize;

	const struct bmfMemory	*txFM;

	ufixed		chExtra, spExtra;
	ifixed		tabAlign, tabOrigin;
} grafPort;

/* for indexed-native, gt is the inverse table cast as a 'long int' */
/* otherwise, the green channel bit count */
#define DEF_RGB_RTN(n)		uint32 n(int i, long int gt, const rgb48 *s)
#define TYP_RGB_RTN(n)		uint32(*n)(int, long int, const rgb48*)

typedef TYP_RGB_RTN(rgbRoutine);

/* ---- macros, inlines */

static __forceinline ifixed addu(ifixed a,
	ufixed b)
{
	ifixed		v = a+b;

	return v<a ? IMAX(v):v;
}

static __forceinline ifixed subu(ifixed a,
	ufixed b)
{
	ifixed		v = a-b;

	return v>a ? IMIN(v):v;
}

static __forceinline ufixed idim(ifixed a,
	ifixed b)
{
	return a>b ? a-b:b-a;
}

static __forceinline inline uint32 formDirect(const rgb48 *c,
	int g,
	int rb)
{
	int		h = 16-rb;

	return (c->red>>h<<g|c->green<<g>>16)<<rb|c->blue>>h;
}

#define formGray(r,g,b,u)	(u=r,u+=(((u+g)>>1)+b)>>1,(u>>1)+g)

static inline int formPxFmt(const pixMap *m)
	{return m->pSize<<8|(m->minCmp ? m->minCmp:m->bpp)|(m->pType&(pTypeLSB|pTypeLGreen));}

/* unaligned read macros */
#if env68k|envppc|envx86
 #define urb16(a)		(0+*(const uint16*)(a))
 #define urb32(a)		(0+*(const uint32*)(a))
 #define urw32(a)		(0+*(const uint32*)(a))
#else
 #define urb16(a)		((uint32)((const uint8*)(a))[!BIG_ENDIAN]<<8 \
	|((const uint8*)(a))[BIG_ENDIAN])
 #define urb32(a)		((uint32)((const uint8*)(a))[3*!BIG_ENDIAN]<<24 \
	|((const uint8*)(a))[2-BIG_ENDIAN]<<16 \
	|((const uint8*)(a))[1+BIG_ENDIAN]<<8 \
	|((const uint8*)(a))[3*BIG_ENDIAN])
 #define urw32(a)		((uint32)((const uint16*)(a))[!BIG_ENDIAN]<<16 \
	|((const uint16*)(a))[BIG_ENDIAN])
#endif

#define q2dsort2(a,b)		do{ pt2dq _u; \
	if(a.y>b.y) _u=a,a=b,b=_u; \
}while(0)

/* ---- globals */

extern const uint32	pSizeMasks[32], pMCmpMasks[16];
extern const uint8	pDepthX[32];
extern const rgbRoutine	rgbCCRoutine[8];

/* ---- prototypes */

/*
 newPixMap
	Allocates a pixMap, using calloc and the fields given.
	Pass zero for bpp to attempt to match the screen format.
	Call disposePixMap when done.
  Notes:
	It assumes the pType, based on bpp.
	The initial NULL pTable implies grayscale for indexed modes.
	The pitch is aligned to four bytes.
	The user may modify the structure to see fit.
  Result:
	The map, or NULL on error.
*/
pixMap *newPixMap(int bpp,
	int width,
	int height);
/*
 disposePixMap
	Disposes of a newPixMap-created pixMap.
	Pixels are not disposed if the pHand field is NULL.
*/
int disposePixMap(pixMap*);

/*
 pmAllocate
	Creates a movable native handle for a buffer.
	The contents of the buffer is undefined.
  Result:
	Returns 1 if a buffer was created.
*/
int pmAllocate(pixMap*);
/*
 pmFree
	Deallocates the movable pixel buffer, nothing else.
*/
int pmFree(pixMap*);
/*
 pmErase
	Fills a single-pixel pen for the entire bounds.
	The pixMap must be locked prior to the call.
  Result:
	Returns 1 on success; zero if unlocked or empty.
*/
int pmErase(pixMap*,
	uint32 penc);
/*
 pmFilter
	Performs a direct filter for a number of scanlines.
	The ctable must be >= the size indicated by the green channel depth
	- or, for grayscale images, m->bpp.
	This is faster and more capable than using the mBlend mode.
  Notes:
	The pixMap must be locked prior to the call.
	Grayscale tables call for 1024 entries.
  Result:
	Returns 1 on success; zero if unlocked or either ptr are empty.
*/
int pmFilter(pixMap *m,
	const cTab64 *t,
	int c); /* clipped */

/*
 lockPixels
	Locks the pixmap baseAddr, updating the map if necessary.
  Result:
	The pixels base address, or NULL on error.
*/
int8 *lockPixels(pixMap *p);
/*
 unlockPixels
	Unlocks the pixmap baseAddr, updating the map if necessary.
  Result:
	Positive if unlocked, zero if nothing changed, negative on error.
*/
int unlockPixels(pixMap *p);

/*
 openGrafPort
	Initializes a grafPort, with default values and a pixMap.
	Close with closeGrafPort.
  Note:
	This expects the port to not have a valid pixMap (will not dispose).
	This calls newPixMap; see the comment on that function above.
  Result:
	Positive on success; otherwise, an error.
*/
int openGrafPort(grafPort *p,
	int bpp,
	int width,
	int height,
	boolean allocate);
/*
 closeGrafPort
	Disposes of the contents within a grafPort.
*/
int closeGrafPort(grafPort*);

/*
 pixMapToNativeDesc
	Allocates and defines a native platform bitmap/pixmap from a pixMap.
*/
#ifdef WINUSERAPI
BITMAPINFO *pixMapToNativeDesc(pixMap*);
#elif TARGET_OS_MAC
PixMapHandle pixMapToNativeDesc(pixMap*);
#endif

/* color selection & inverse tables */

/*
 RGBFore/RGBBack
	Sets the fore/back color, RGB and pen.
	If a pattern is loaded, the following ct entry will be set:
	 fore: last; back: first.
*/
void RGBFore(grafPort *p,
	const rgb48*);
void RGBBack(grafPort *p,
	const rgb48*);

/*
 RGBToPen
	Forms a multi-pixel for the passed pixmap format, given a color.
	It may build an inverse table for the pixMap.
  Result:
	Can only return zero on error.
*/
uint32 RGBToPen(pixMap *m,
	const rgb48*); /* ptr must be valid */
/*
 PenToRGB
	Converts a pen color to a 48-bit color.
  Note:
	This only takes the lower bits pixel of the passed value.
  Result:
	Returns 1 on success; negative on error.
*/
int PenToRGB(pixMap *m,
	uint32 penc,
	rgb48 *rgb);
/*
 getPixelRGB
	Converts a pixel, given a point, to a 48-bit color.
  Note:
	This may call lockPixels/unlockPixels.
  Result:
	Returns 1 on success; negative on error.
	Returns zero if out of bounds, where 'v' is left undefined.
*/
int getPixelRGB(pixMap *m,
	int x,
	int y,
	rgb48 *v);
/*
 pmGetRGBCC
	Calculates the color conversion code for a pixMap.
  Result:
	The index in the rgbCCRoutine table; negative on error (gt is undefined).
*/
int pmGetRGBCC(pixMap *m,
	long int *gt); /* must be valid */

int copyCTable(cTab64 **out,
	const cTab64 *in);

/*
 makeITables
	Builds inverse & direct tables, given a color table.
	minCmp is where the channel resolution resides.
	The res. for a gray table is always 8, though minCmp=0.
  Result:
	The effective resolution (0=grays); otherwise a basic error code.
*/
int makeITables(const cTab64 *ct,
	uint8 **itab, /* must be valid */
	int bpp,
	int res,
	boolean forceColor); /* don't make gray table */
int pmMakeITables(pixMap*,
	int res);

/*
 pmDefaultMacTable
	Depending on the bpp depth, loads the appropriate Mac OS table.
  Result:
	Returns 1 on success; zero if the pm has a direct or fixed clut.
*/
int pmDefaultMacTable(pixMap *m,
	boolean gray);

void *rGenRemappingTable(pixMap *m, /* must be valid */
	const cTab64 *ct,
	uint8 bpp,
	uint8 flag,
	boolean is32b);

/* patterns */

/*
 rMakePat
	Creates a pixPat out of an 8x8 monochrome map.
	Replaces pattern in 'pat' if !=NULL.
  Result:
	1 on success; negative on error.
*/
int rMakePat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *fore,
	const rgb48 *back);
#define rDisposePat(p)		rMakePat(&p,NULL,NULL,NULL);
void rPenPat(grafPort *p,
	const uint8 inpat[8]);
/*
 rCopyPat
	Allocates a copy of a pattern.  Use rDisposePat when done.
*/
int rCopyPat(pixPat **out, /* must be valid */
	const pixPat *in);
/*
 rSavePen
	Makes a save-state for a port, copying the pattern if it exists.
	Use rFreePen (or really rRestorePen w. 'freeState') when done with it.
*/
int rSavePen(grafPort *p,
	penState *state);
int rRestorePen(grafPort *p,
	penState *state,
	boolean freeState);
#define rFreePen(s) rRestorePen(NULL,s,1)
/*
 rPenNormal
	Sets the pen state to the default state, the way it would be at creation.
	The optional 'oldstate' allows you to save it, no additional allocations.
	Like rSavePen, if you won't restore the pen, use rFreePen.
*/
int rPenNormal(grafPort *p,
	penState *oldstate);

/* graphics */

/*
 OutsetRect16
	Outsets a 16-bit rectangle, performing maximum clipping if necessary.
  Note:
	This does not perform minimum clipping (e.g., if either x or y are negative).
*/
void OutsetRect16(rect16 *b, /* must be valid */
	int x,
	int y);
void OffsetRect16(rect16 *b, /* must be valid */
	int x,
	int y);
/*
 SectRect16
	Calculates the smaller boundary within two rectangles.
	Returns 0 if empty.
*/
int SectRect16(rect16 *d, /* must be valid */
	const rect16 *s);
int UnionRect16(rect16 *d, /* must be valid */
	const rect16 *s);
/*
 FixRect
	Expands a four 16-bit coordinates to a rounded, fixed-point rectangle.
*/
void FixRect(rect2dq *b, /* must be valid */
	int left,
	int top,
	int right,
	int bottom);
/*
 FixPt
	Expands a coordinate to a rounded, fixed-point rectangle.
	The values are pinned to the maximum range.
*/
void FixPt(pt2dq*, /* must be valid */
	int x,
	int y);

int ptInRect16(const rect16*,
	pt2d16);

/* the results for the renderers:
 -2 = memory error (std utility error)
 -1 = parameter error or bad/uninitialized data
  0 = clipped-out or empty
  1 = within bounds (doesn't necessarily mean any pixels were changed)
*/

/* regions */

/*
 newRgn
	Creates an empty region; bBox top/bottom are zero.
	Returns NULL on error.
*/
SLRgn *newRgn(void);

/*
 rgnXorLine
	Inserts the integral x-coords of a line into a region.
	Automatically adjusts the size and (re)allocates.
	The ptr in rrgn should not become invalid (unless already corrupted).
  Result:
	1 on success; 0 if nothing added; negative on error.
*/
int rgnXorLine(SLRgn **rrgn,
	int16 cliptop,
	int16 clipbot,
	pt2dq p0,
	pt2dq p1);

/*
 rectInRgn
	Scans a region to see if the passed rectangle intersects.
  Result:
	1 if true; 0 if false; negative on error.
*/
int rectInRgn(SLRgn*,
	int left,
	int top,
	int right,
	int bottom);
int ptInRgn(SLRgn*,
	int x,
	int y);
/*
 rgnCompact
	Removes any excess in a region, adjusting the bBox height if necessary.
	bitmapToRgn calls this.
*/
SLRgn *rgnCompact(SLRgn*,
	boolean linesOnly); /* remove only excess hollow lines above/below */
/*
 rgnOffsetXY
	Offsets (moves) a region, given an x/y delta.
	The resulting region is clipped and compacted.
  Result:
	The new region, or NULL (and disposed) if empty.
*/
SLRgn *rgnOffsetXY(SLRgn *rgn,
	int dx,
	int dy);

/*
 bitmapToRgn
	Generates a region from a monochrome bitmap.
  Result:
	The region; NULL on error.
*/
SLRgn *bitmapToRgn(const uint8*,
	int pitch,
	const rect16 *bounds);

/*
 rRegion
	Given the current pen mode, paints the area within a region.
  NOTE:
	Framing is not currently implemented, but a hollow pen is recognized.
*/
int rRegion(grafPort *p,
	const SLRgn *rgn);

/*
 rTriangle
	Given the current drawing mode, paints the area within three points.
	This routine ignores the pen width & height.
*/
int rTriangle(grafPort *p,
	pt2dq p0,
	pt2dq p1,
	pt2dq p2);

/*
 rPolygon
	Given the current pen mode, paints a polygon.
	If the penw is valid, a line frame is drawn.
	(A "single-pixel" frame is drawn if only the penh is invalid.)
*/
int rPolygon(grafPort *p,
	const pt2dq *ep,
	int count);
/*
 rLine
	Like QuickDraw, paints a rectangular pen between two pens.
	If the pen is invalid, a "single-pixel" line is drawn.
*/
int rLine(grafPort *p,
	pt2dq p0,
	pt2dq p1);
/*
 rCLineList
	Paints rectangular pen between a connected list of points.
	If the pen is invalid, a "single-pixel" line is drawn.
*/
int rCLineList(grafPort *p,
	const pt2dq *ep,
	int ptc);
/*
 rOval
	Paints the area within an oval, given rectangular bounds.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rOval(grafPort *p,
	const rect2dq *ov);
/*
 rRect
	Paints the area within a rectangle.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rRect(grafPort *p,
	const rect2dq *rc);
/*
 rRoundRect
	Paints the area within a rectangle with round edges (rx/ry).
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
int rRoundRect(grafPort *p,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry);
/*
 rFillRect
	Paints an integral rectangle, ignoring the current mode & pen size.
	The current pen color/pattern is used in a solid transfer.
	This is faster than rRect for solid painting.
*/
int rFillRectMode(grafPort *p,
	int left,
	int top,
	int right,
	int bottom,
	uint8 mode);
#define rFillRect(p,l,t,r,b) rFillRectMode(p,l,t,r,b,mSrcCopy)
#define rEraseRect(p,l,t,r,b) rFillRectMode(p,l,t,r,b,mSrcNotCopy)
/*
 rInvertRect
	Like rFillRectMode(...,mSrcXor) but ignores the color/pPat.
*/
int rInvertRect(grafPort *p,
	int left,
	int top,
	int right,
	int bottom);

/*
 rBltPix
	Transfers a pixMap to a port, within clipping bounds.
	Scaling is performed if any widths or heights are inequal.
*/
int rBltPix(grafPort *p,
	pixMap *srcpm,
	int srcx,
	int srcy,
	int srcw,
	int srch,
	int dstx,
	int dsty,
	int dstw,
	int dsth,
	uint8 mode);

/*
 decompressPictFile
	This reads a QD PICT file and calls renderQDPict.
*/
pixMap *decompressPictFile(const char *path,
	int *err);

/* ---- basic conversion, copy */

void b1to8(uint8 *d, /* must be valid */
	const uint8 *s, /* must be valid */
	int byteCount);

/*
 rCopyBits
	Emphasis on 'bits,' this routine copies a line, aligned or unaligned.
*/
void rCopyBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dstBitOffset,
	int srcBitOffset,
	int bitCount);
void rNotCopyBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm); /* source complement mask */
void rOrBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);
void rXorBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);
void rAndBits(int8 *a, /* must be valid */
	const int8 *b, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);

#define __MGRAF_H
#endif
