/*
 mgraphic.c

   Revised: 2014.09.17
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.02.24
  Took advantage of 32-bit div. result to maximize slopdiv ratio range.
(For horizontal lines, up half would not've shown up; now complete.)
2014.02.25
  Framed ovals!  Finally figured out the clipping mech.
rInitClip2 now a separate - but inlined - function.
2014.02.26
  Fixed framing outset, implemented round-rectangles (framed- broken).
2014.02.27
  Fixed framed round-rectangles, rewriting the initialization separately.
2014.03.05
  pixMap origin/bounds like QD, except x/y/w/h instead of 'rect16'.
  Fixed Asm slopdiv - the int64qset was treating a signed value unsigned.
  Got custom port clipping & origin working...
2014.03.06
  Separated the crux of line drawing into stdLine for an efficient rCLineList.
  rInitRoundRect: fixed incorrect use of clip-top when oy should be used.
2014.03.08
  Universal support for all depths with fillx/invrx, using the same method
used in the M68K Asm version, but written from scratch in C.
Additional disadvantage of LSB: bswap required for string masks.
(The M68K version, however, used bitfield r/w - smaller.)
2014.03.09
  iTable 8bpp RGB translucency working, but slow.
- now sped up by building a direct table instead of passing the cSpec table;
at most 512 bytes, worth it.  Besides, od28-6 already builds one.
  Finally implemented region code... now, to get it to work...

2014.03.10
  Split from mgraf.c, into mblt and mgraphic.
2014.03.11
  Got region code working, finally.  Rendering, too.
2014.04.05
  More rgn compacting; define MINIMAL_RGNCOMPACT for a smaller version 
(but probably slower all around, as excess empty lines may be processed).
2014.04.09
  Fixed two rgn bugs: reading a boundary at exactly the region end,
and another inadvertently regarding compaction (now done always).
  Added ptInRgn.
2014.05.17/8
  Added ptInRect, corrected names of some routines:
rUnionRect16 to SectRect16, rgnSectLine to rgnXorLine.
2014.08.10
  Revamped single-pixel line drawing - inclusive; looks perfect now.
2014.08.16
  No, single-pixel stdLine wasn't perfect - not lining up with polygons.
The slope sign now affects the initial point in rounding.
(By itself, better results than QD's?)
*/

#include "mgraf.h"
#include "mblt.h"
#include "../ftrig.h"

/* ---- code */

int ptInRect16(const rect16 *b,
	pt2d16 pt)
{
	if (!b)
		return 0;
	return pt.y>=b->top && pt.x>=b->left
		&& pt.y<b->bottom && pt.x<b->right;
}

void FixPt(pt2dq *pt,
	int x,
	int y)
{
	pt->x = x!=(int16)x ? x>>31^IMAX(pt->x):x<<16|0x8000;
	pt->y = y!=(int16)y ? y>>31^IMAX(pt->y):y<<16|0x8000;
}

void FixRect(rect2dq *b,
	int left,
	int top,
	int right,
	int bottom)
{
	b->top = top!=(int16)top ? top>>31^IMAX(ifixed):top<<16|0x8000;
	b->left = left!=(int16)left ? left>>31^IMAX(ifixed):left<<16|0x8000;
	b->bottom = bottom!=(int16)bottom ? bottom>>31^IMAX(ifixed):bottom<<16|0x8000;
	b->right = right!=(int16)right ? right>>31^IMAX(ifixed):right<<16|0x8000;
}

void OutsetRect16(rect16 *b,
	int x,
	int y)
{
	int32		v;

	if ((v=b->top-y)<IMIN(b->top))
		v = IMIN(b->top);
	b->top = v;
	if ((v=b->left-x)<IMIN(b->left))
		v = IMIN(b->left);
	b->left = v;
	if ((v=b->bottom+y)>IMAX(b->bottom))
		v = IMAX(b->bottom);
	b->bottom = v;
	if ((v=b->right+x)>IMAX(b->left))
		v = IMAX(b->left);
	b->right = v;
}

void OffsetRect16(rect16 *b,
	int x,
	int y)
{
	int32		v;

	v = b->top+y, b->top = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = b->left+x, b->left = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = b->bottom+y, b->bottom = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = b->right+x, b->right = v!=(int16)v ? v>>31^IMAX(int16):v;
}

int SectRect16(rect16 *d,
	const rect16 *s)
{
	if (s)
		if (s->top>=s->bottom || s->left>=s->right)
			*d = *s;
		else
		{
			if (s->top>d->top)
				d->top = s->top;
			if (s->left>d->left)
				d->left = s->left;
			if (s->bottom<d->bottom)
				d->bottom = s->bottom;
			if (s->right<d->right)
				d->right = s->right;
		}
	return d->bottom>d->top && d->right>d->left;
}

int UnionRect16(rect16 *d,
	const rect16 *s)
{
	if (s)
		if (d->bottom<=d->top || d->right<=d->left)
			*d = *s;
		else if (s->top<s->bottom && s->left<s->right)
		{
			if (s->top<d->top)
				d->top = s->top;
			if (s->left<d->left)
				d->left = s->left;
			if (s->bottom>d->bottom)
				d->bottom = s->bottom;
			if (s->right>d->right)
				d->right = s->right;
		}
	return d->bottom>d->top && d->right>d->left;
}

static int rHalfOutRect(bltParam *f,
	rect2dq *b)
{
	ufixed		o;

	if (f->vx<0 || f->vy<0)
		f->vx =-1;
	else
	{
		if (!(f->vx|f->vy))
			return 0; /* hollow pen */
		/* adding any part of the fraction make ovals look bad */
		o = f->vx>>1&~0xFFFF;
		b->left = subu(b->left,o), b->right = addu(b->right,o);
		o = f->vy>>1&~0xFFFF;
		b->top = subu(b->top,o), b->bottom = addu(b->bottom,o);
	}
	if (b->bottom<=b->top)
		return 0;
	return 1;
}

/* renderers, starting with the trapezoid renderers
 rTrapezoid, rTriangle, stdLine:/rCLineList/rLineList
*/

#if 0
#define initPenLine(f,p0,p1)	do{ \
	q2dsort2(p0,p1); /* sorting must come first, for p1y-p0y diff below */ \
	if ((f.vx|f.vy)<0) { /* non-pen -> single-pixel (hack) */ \
		if (idim(p0.x,p1.x)>=(ufixed)(p1.y-p0.y)) \
			f.vx=0, f.vy=0x10000; \
		else	f.vy=0, f.vx=0x10000; \
	} \
}while(0)
#endif

#define adjustHLine1(f,o,c)	do{ \
	if(c>f.cl.right) c=f.cl.right; \
	if(o<f.cl.left) o=f.cl.left; \
	c-=o; \
	if((o-=f.ox)<0) c+=o, o=0; \
}while(0)
/* two counts, adjusted with offset */
#define adjustHLine2(f,o,c,c2)	do{ \
	if(c>f.cl.right) c=f.cl.right; \
	if(c2>f.cl.right) c2=f.cl.right; \
	if(o<f.cl.left) o=f.cl.left; \
	c-=o,c2-=o; \
	if((o-=f.ox)<0) c+=o,c2+=o,o=0; \
}while(0)

void rTrapezoid(trapezoid *f,
	pt2dq p0,
	pt2dq p1)
{
	int8		*a = f->prm.p, *e;
	int64		x0, x1, dx0, dx1;
	int32		c;
	int8		xr;
	/* clip bottom */
	if ((c=(p1.y>>16)-f->prm.cl.top)<0)
		c = 0;
	a+= f->prm.oz*(c+f->prm.cl.top-f->prm.oy);
	if (a<(e=f->prm.e))
		e = a;
	/* clip top */
	if (f->prm.a>=e)
		goto end;
	if ((c=f->prm.cl.top-(p0.y>>16))<0)
		c = 0;
	dx0 = rslopdiv(&x0,p0.x,p0.y,p1.x,p1.y,c);
	x1 = f->x1, dx1 = f->dx1;
	/* whether to reverse x0:x1 */
	xr = int64eq(x0,x1) ? int64ge(dx0,dx1):int64ge(x0,x1);
	rBltOrigin(&f->prm,f->prm.ox,p0.y>>16);
	do
	{ /* clip and blt line */
		int32		o;

		if (xr)
			o = int64hi(x1), c = int64hi(x0);
		else
			o = int64hi(x0), c = int64hi(x1);
		adjustHLine1(f->prm,o,c);
		if (c>0)
			f->prm.rtn(o,c,&f->prm);
		(void)int64add(x0,dx0);
		(void)int64add(x1,dx1);
	} while(rEndLine(&f->prm,e));
	f->x1 = x1;
end:	f->prm.a = e;
}

int rTriangle(grafPort *p,
	pt2dq p0,
	pt2dq p1,
	pt2dq p2)
{
	trapezoid	f; /* 2 trapezoids */
	int		c, bl, br;

	if (!p)
		return -1;
	/* sort */
	q2dsort2(p0,p1);
	q2dsort2(p0,p2);
	q2dsort2(p1,p2);
	if ((p0.x==p1.x && p0.y==p1.y) || (p1.x==p2.x && p1.y==p2.y))
		return 0;
	if ((c=rInitBlt(&f.prm,p,NULL,NULL,p0.y>>16,p2.y>>16,p->pnMode))<0)
		{c = 0; goto end;}
	/* clip horizontal bounds */
	if ((bl=br=p0.x)>p1.x)
		bl = p1.x;
	else
		br = p1.x;
	if (bl>p2.x)
		bl = p2.x;
	if ((bl>>=16)>f.prm.cl.left)
		f.prm.cl.left = bl;
	if (br<p2.x)
		br = p2.x;
	if ((br=br>>16)<f.prm.cl.right)
		f.prm.cl.right = br;
	if (f.prm.cl.right<=f.prm.cl.left)
		{c = 0; goto end;}
	/* the only persistent slope */
	f.dx1 = rslopdiv(&f.x1,p0.x,p0.y,p2.x,p2.y,c);
	rTrapezoid(&f,p0,p1);
	rTrapezoid(&f,p1,p2);
	c = 1;
end:	rEndBlt(&f.prm);
	return c;
}

static int stdLine(trapezoid *f,
	pt2dq p0,
	pt2dq p1)
{
	pt2dq		p2;
	ifixed		u;
	int32		olvx = f->prm.vx, olvy = f->prm.vy;
	int		o, c, ky, v;
	boolean		opl = (olvx|olvy)<0;

	q2dsort2(p0,p1);
	if ((c=addu(p1.y,opl ? 0x10000:f->prm.vy)>>16)>f->prm.cl.bottom)
		c = f->prm.cl.bottom;
	if (c<=f->prm.cl.top)
		return 0; /* clipped out-top */
	if ((o=p0.y>>16)>=f->prm.cl.bottom)
		return 0; /* clipped out-bottom */
	ky = rInitClip2(&f->prm,o,c,&f->prm.a,&f->prm.e);
	if (opl)
	{ /* single-pixel line */
		o = p0.x>>16, c = p1.x>>16;
		if (o>c)
			v = c, c = o, o = v;
		if (o>f->prm.cl.left)
			f->prm.cl.left = o;
		if (++c<f->prm.cl.right)
			f->prm.cl.right = c;
		if (f->prm.cl.right<=f->prm.cl.left)
			return 0;
		rBltOrigin(&f->prm,f->prm.ox,p0.y>>16);
		f->dx1 = rslopdiv(&f->x1,p0.x,p0.y,p1.x,p1.y,ky);
		o = (p0.x>>16)+(p0.x>p1.x);
		if (p1.y-p0.y>=idim(p1.x,p0.x))
		{
			int64		hdx = f->dx1;

			if (p1.x<p0.x)
				(void)int64neg(hdx), int64subu32(hdx,1); //was add 0x1
			int64shr(hdx,1);
			(void)int64add(f->x1,hdx);
		} else //was simply +.5
			(void)int64addhl(f->x1,0,int64hi(f->dx1)>>31^0x80000000);
		goto sk;
		do
		{
			o = int64hi(f->x1);
			(void)int64add(f->x1,f->dx1);
		sk:	c = int64hi(f->x1);
			if (o==c)
				c = o+1;
			else if (o>c)
				v = c, c = o, o = v;
			adjustHLine1(f->prm,o,c);
			if (c>0)
				f->prm.rtn(o,c,&f->prm);
		} while(rEndLine(&f->prm,f->prm.e));
		return 1;
	}
	p2.y = addu(p1.y,f->prm.vy);
	p2.x = u = addu(p1.x,f->prm.vx);
	if (p0.x<=p1.x)
	{ /* flat left: x1 = top of top-right to bottom-right */
		p2.x = p1.x, p1.x = u;
		f->dx1 = rslopdiv(&f->x1,addu(p0.x,f->prm.vx),p0.y,p1.x,p1.y,ky);
		/* clip for flat ends (error requires both l/r) */
		if ((u=p1.x>>16)<f->prm.cl.right)
			f->prm.cl.right = u;
		if ((u=p0.x>>16)>f->prm.cl.left)
			f->prm.cl.left = u;
	} else
	{ /* reverse: x1 = top of top-left to bottom-left */
		f->dx1 = rslopdiv(&f->x1,p0.x,p0.y,p1.x,p1.y,ky);
		p0.x = addu(p0.x,f->prm.vx);
		/* clip for flat ends (error requires both l/r) */
		if ((u=p0.x>>16)<f->prm.cl.right)
			f->prm.cl.right = u;
		if ((u=p1.x>>16)>f->prm.cl.left)
			f->prm.cl.left = u;
	}
	if (f->prm.cl.right<=f->prm.cl.left)
		return 0;
	p1.x = p0.x, p1.y = addu(p0.y,f->prm.vy);
	rTrapezoid(f,p0,p1);
	rTrapezoid(f,p1,p2);
	f->prm.vx = olvx, f->prm.vy = olvy;
	return 1;
}

int rCLineList(grafPort *p, /* must be valid */
	const pt2dq *ep, /* must be valid */
	int ptc) /* must be >1 */
{
	trapezoid	f;
	rect32		cl;
	pt2dq		p0, p1;
	const pt2dq	*pt;
	ifixed		pw, ph;
	ifixed		dx, dy;
	int32		c, bl, br;
	int		r;

	if (!p || !ep || ptc<2)
		return -1;
	f.prm.vx = p->penw, f.prm.vy = p->penh;
	if (!(f.prm.vx|f.prm.vy))
		return 0; /* hollow pen */
	if (f.prm.vy<0)
		f.prm.vx =-1;
	/* cl temporarily used to build the polygon's fixed-point bounds */
	bl = f.prm.cl.top = IMAX(ifixed);
	br = f.prm.cl.bottom = IMIN(ifixed);
	pt = ep, c = ptc;
	do
	{
		if (bl>pt->x)
			bl = pt->x;
		if (br<pt->x)
			br = pt->x;
		if (f.prm.cl.top>pt->y)
			f.prm.cl.top = pt->y;
		if (f.prm.cl.bottom<pt->y)
			f.prm.cl.bottom = pt->y;
	} while(++pt,--c);
	ph = f.prm.vy;
	if ((pw=f.prm.vx)<0)
		pw = ph = 0x10000, br+= ph, f.prm.cl.bottom+= ph;
	else
	{ /* frame expansion */
		br+= f.prm.vx, f.prm.cl.bottom+= f.prm.vy;
		c = f.prm.vx>>1&~0xFFFF;
		bl-= c, br+= c;
		c = f.prm.vy>>1&~0xFFFF;
		f.prm.cl.top = subu(f.prm.cl.top,c);
		f.prm.cl.bottom = addu(f.prm.cl.bottom,c);
	}
	if ((c=rInitBlt(&f.prm,p,NULL,NULL,f.prm.cl.top>>16,f.prm.cl.bottom>>16,p->pnMode))<0)
		{r = 0; goto end;}
	if ((bl>>=16)>f.prm.cl.left)
		f.prm.cl.left = bl;
	if ((br>>=16)<f.prm.cl.right)
		f.prm.cl.right = br;
	if (f.prm.cl.right<=f.prm.cl.left)
		{r = 0; goto end;}
	cl = f.prm.cl;
	dx = pw>>1&~0xFFFF, dy = ph>>1&~0xFFFF;
	pw = f.prm.vx, ph = f.prm.vy; /* reset */
	r = 0, --ptc;
	do
	{
		p0.x = subu(ep->x,dx), p0.y = subu(ep->y,dy);
		++ep;
		p1.x = subu(ep->x,dx), p1.y = subu(ep->y,dy);
		/* reinit clipping */
		f.prm.cl = cl;
		r|= stdLine(&f,p0,p1);
	} while(--ptc);
	r = 1;
end:	rEndBlt(&f.prm);
	return r;
}

int rLine(grafPort *p,
	pt2dq p0,
	pt2dq p1)
{ /* this takes advantage of l/r clipping - 2 trapezoids */
	trapezoid	f;
	int32		c;

	if (!p)
		return -1;
	f.prm.vx = p->penw, f.prm.vy = p->penh;
	if (!(f.prm.vx|f.prm.vy))
		return 0;
	p->pnLoc = p1;
	q2dsort2(p0,p1);
	if ((c=rInitBlt(&f.prm,p,NULL,NULL,p0.y>>16,addu(p1.y,(f.prm.vx|f.prm.vy)<0 ? 0x10000
			:f.prm.vy)>>16,p->pnMode))<0)
		{c = 0; goto end;}
	stdLine(&f,p0,p1);
	c = 1;
end:	rEndBlt(&f.prm);
	return c;
}

int rRegion(grafPort *p,
	const SLRgn *rgn)
{
	bltParam	f;
	rect16		b;
	const int16	*s, *e;
	int32		o, c;

	if (!p || !rgn || (c=rgn->size)<OFFSET(SLRgn,x) || c&1)
		return -1;
	s = rgn->x, e = (const int16*)(rgn->size+(const int8*)rgn);
	if (e<=s || (o=p->penw|p->penh)==0)
		return 0;
	b = rgn->bBox;
	if (o>=0)
		OutsetRect16(&b,p->penw>>17,p->penh>>17);
	if ((c=rInitBlt(&f,p,NULL,NULL,b.top,b.bottom,p->pnMode))<0
			|| b.left>=f.cl.right
			|| b.right<=f.cl.left)
		{c = 0; goto end;}
	--e;
	if (c) /* skip clipped-out lines */
		do
		{
			if (s>=e || s[1]<*s)
				return -1;
			for(++s; *s>=s[-1]; ++s)
				;
		} while(--c);
	rBltOrigin(&f,f.ox,b.top);
	do
	{
		if (++s>=e)
			return f.a+f.oz>=f.e ? 1:-1;
		if (*s<s[-1])
			return -1;
		do
		{
			if (s>=e)
				return -1;
			if ((c=s[1])<(o=*s))
				{++s; break;}
			adjustHLine1(f,o,c);
			if (c>0)
				f.rtn(o,c,&f);
		} while((s+=2)<e && *s>=s[-1]);
	} while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

/* For polygons, QD basically builds a sect-region.  This is optimal. */

int rPolygon(grafPort *p,
	const pt2dq *ep,
	int c)
{
	SLRgn		*rgn;
	const pt2dq	*pt;

	if (!p)
		return -1;
	if (p->penw>=0) /* frame - QD does not force closure, and neither should this */
		return rCLineList(p,ep,c);
	if ((pt=ep)==NULL)
		return -1;
	if (c<3 || !(p->penw|p->penh))
		return 0; /* no area */
	--c, rgn = NULL, pt = ep;
	do
		if (rgnXorLine(&rgn,p->clip.top,p->clip.bottom,*pt,pt[1])<0)
			goto bad;
	while(++pt,--c);
	/* area painting requires automatic close */
	if (rgnXorLine(&rgn,p->clip.top,p->clip.bottom,*pt,*ep)<0)
	{
	bad:	if (rgn)
			free(rgn);
		return -1;
	}
	c = rRegion(p,rgn);
	if (rgn)
		free(rgn);
	return c;
}

/* rectangle-parameter functions
  rOval, rRect, rRoundRect
*/

int rOval(grafPort *p,
	const rect2dq *ov)
{
	bltParam	f;
	rect2dq	b; /* outer-bounds */
	int8		*ia, *ie; /* interior markers (outline case) */
	uint64		dy0, dy1;
	int64		y0, y1;
	ufixed		w;
	ifixed		cx, hx;
	int32		c, o;

	if (!ov || !p)
		return -1;
	b = *ov, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b))
		return 0;
	if ((c=rInitBlt(&f,p,NULL,NULL,b.top>>16,b.bottom>>16,p->pnMode))<0)
		{c = 0; goto end;}
	/* rarely, some narrow fractions/large ratios manage to jutt a lot */
	if ((o=b.left>>16)>f.cl.left)
		f.cl.left = o;
	if ((o=(b.right-1)>>16)<f.cl.right)
		f.cl.right = o; /* to not -1 this will result in full-width edge pixels */
	if (f.cl.right<=f.cl.left)
		{c = 0; goto end;}
	/* cx +/- w~/(1-(y/h)**2) */
	w = b.right-b.left;
	cx = b.left+(w>>1)+(w&1); /* any round not toward hollow left */
	/* set up y increment & position */
	dy0 = rincdiv(&y0,b.top,b.bottom,c);
	int64subhl(y0,0x40000000,0);
	int64xset(dy1,0);
	if (w<=(hx=f.vx) || (hx=w-f.vx*2)+0u>w) /* for non-pen case, >-1u is impossible */
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;

		c = rInitClip2(&f,it>>16,(ib-1)>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
		{
			dy1 = rincdiv(&y1,it,ib,c);
			int64subhl(y1,0x40000000,0);
		}
	}
	rBltOrigin(&f,f.ox,b.top>>16);
	do
	{
		o = rarcmul(int64hi(y0),w);
		c = addu(cx,o)>>16, o = subu(cx,o)>>16;
		if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				ufixed	i = rarcmul(int64hi(y1),hx);
				int32	ic;

				ic = subu(cx,i)>>16;
				adjustHLine1(f,o,ic);
				if (ic>0)
					f.rtn(o,ic,&f);
				o = (addu(cx,i)-1)>>16;
				(void)int64add(y1,dy1);
			}
		adjustHLine1(f,o,c);
		if (c>0)
			f.rtn(o,c,&f);
		(void)int64add(y0,dy0);
	} while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

int rRect(grafPort *p,
	const rect2dq *rc)
{
	bltParam	f;
	rect2dq		b; /* outer-bounds */
	int8		*ia, *ie; /* interior markers (outline case) */
	ufixed		w;
	int32		o, c, ic1, io, ic2;

	if (!rc || !p)
		return -1;
	b = *rc, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b))
		return 0;
	if (rInitBlt(&f,p,NULL,NULL,b.top>>16,b.bottom>>16,p->pnMode)<0)
		{c = 0; goto end;}
	o = b.left>>16, ic2 = c = b.right>>16;
	w = b.right-b.left, ic1 = io = 0;
	if (w<=f.vx || w-f.vx*2u>w)
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;
		/* `(ib-1)>>16` produces thicker line in zero-frac case */
		rInitClip2(&f,it>>16,ib>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
		{
			ic1 = (b.left+f.vx)>>16;
			io = (b.right-f.vx)>>16;
			adjustHLine1(f,io,ic2);
		}
	}
	/* rRect the fastest, with all the clipping done before the loop */
	adjustHLine2(f,o,c,ic1);
	if (c<=0)
		{c = 0; goto end;}
	rBltOrigin(&f,f.ox,b.top>>16);
	do
	{
		if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				if (ic1>0)
					f.rtn(o,ic1,&f);
				if (ic2>0)
					f.rtn(io,ic2,&f);
				continue;
			}
		f.rtn(o,c,&f);
	} while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

int rFillRectMode(grafPort *p,
	int o,
	int top,
	int c,
	int bottom,
	uint8 mode)
{
	bltParam	f;

	if (!p)
		return -1;
	if (rInitBlt(&f,p,NULL,NULL,top,bottom,mode)<0)
		{c = 0; goto end;}
	adjustHLine1(f,o,c);
	if (c<=0)
		{c = 0; goto end;}
	rBltOrigin(&f,f.ox,o);
	do
		f.rtn(o,c,&f);
	while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

int rInvertRect(grafPort *p,
	int o,
	int top,
	int right,
	int bottom)
{
	bltParam	f;
	uint32		oldsl;
	int		c;

	if (!p || !p->map)
		return -1;
	oldsl = p->fgSolid, p->fgSolid = p->map->bpp>8 ? 0:~0;
	c = rInitBlt(&f,p,NULL,NULL,top,bottom,mSrcXor);
	p->fgSolid = oldsl;
	if (c<0)
		{c = 0; goto end;}
	c = right;
	adjustHLine1(f,o,c);
	if (c<=0)
		{c = 0; goto end;}
	rBltOrigin(&f,f.ox,o);
	do
		f.rtn(o,c,&f);
	while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

/* Fast clipping is the hard part with fixed-point round rectangles;
 must get the full-width vertical range in integral coordinates.
   Also, no matter how you slice it (no pun intended), the lower part
 fraction & skip must be calculated separately.
   All of this is complex enough to separate it all into its own function.
*/
typedef struct RRect
{
	const int8	*ua, *ue; /* full-width markers */
	int64		dy, y0, y1; /* y-inc, top of rect-rounding positions */
	ufixed		rw, mw; /* round & mid-width; rarcmul(y,rw)+mw */
} RRect;

static void rInitRoundRect(RRect *r,
	const bltParam *f,
	int32 ky,
	ufixed w,
	ifixed bt,
	ifixed bb,
	ufixed rx,
	ufixed ry,
	ufixed oy)
{
	int32		uy0, uy1;
	uint16		uy1f;

	if ((rx+=rx)>w)
		rx = w; /* x-collapsed */
	r->rw = rx, r->mw = (w-rx)>>1;
	uy0 = addu(bt,ry)>>16, uy1 = subu(bb,ry);
	uy1f = (uint16)uy1, uy1>>= 16;
	ry+= ry;
	if (uy1<=uy0)
		r->ua = r->ue = (int8*)-1, ry = bb-bt; /* y-collapsed */
	else
	{
		r->ua = f->p+f->oz*(uy0-f->oy);
		r->ue = f->p+f->oz*(uy1-f->oy);
	}
	/* set up y increment & position */
	if (!ry)
	{
		int64xset(r->dy,0);
		r->y0 = r->y1 = r->dy;
	} else
	{
		r->dy = rincdiv(&r->y0,bt,bt+ry,ky);
		int64subhl(r->y0,0x40000000,0);
		r->y1 = r->dy; int64shr(r->y1,16);
		if ((ky=f->cl.top-uy1)<0)
			ky = 0;
		rLineFrac(ky,ky,uy1f);
		int64mulu32(r->y1,ky);
	}
}

int rRoundRect(grafPort *p,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry)
{
	bltParam	f;
	rect2dq	b; /* outer-bounds */
	RRect		rr, rri; /* exterior & interior (outline case) */
	int8		*ia, *ie; /* interior markers */
	ufixed		w;
	ifixed		cx;
	int32		c, o;

	if (rx<0)
		rx = 0;
	if (ry<0)
		ry = 0;
	if (!(rx|ry))
		return rRect(p,rc);
	if (!rc || !p)
		return -1;
	b = *rc, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b))
		return 0;
	if ((c=rInitBlt(&f,p,NULL,NULL,b.top>>16,b.bottom>>16,p->pnMode))<0)
		{c = 0; goto end;}
	/* rarely, some narrow fractions/large ratios manage to jutt a lot */
	if ((o=b.left>>16)>f.cl.left)
		f.cl.left = o;
	if ((o=(b.right-1)>>16)<f.cl.right)
		f.cl.right = o; /* to not -1 this will result in full-width edge pixels */
	if (f.cl.right<=f.cl.left)
		{c = 0; goto end;}
	w = b.right-b.left;
	cx = b.left+(w>>1)+(w&1); /* any frac round not toward hollow left */

	rInitRoundRect(&rr,&f,c,w,b.top,b.bottom,rx,ry,p->origin.y);
	if (w<=f.vx || (f.vx=w-f.vx*2)+0u>w)
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;

		c = rInitClip2(&f,it>>16,ib>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
			rInitRoundRect(&rri,&f,c,f.vx,it,ib,rx,ry,p->origin.y);
	}
	rBltOrigin(&f,f.ox,b.top>>16);
	do
	{
		if (f.a>=rr.ua)
			if (f.a>=rr.ue)
				rr.ua = (int8*)-1, rr.y0 = rr.y1;
			else
			{
				o = b.left>>16, c = b.right>>16;
				goto uo;
			}
		o = rarcmul(int64hi(rr.y0),rr.rw)+rr.mw;
		(void)int64add(rr.y0,rr.dy);
		c = addu(cx,o)>>16, o = subu(cx,o)>>16;
	uo:	if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				ufixed	i;
				int32	ic;

				if (f.a>=rri.ua)
					if (f.a>=rri.ue)
						rri.ua = (int8*)-1, rri.y0 = rri.y1;
					else
					{
						i = rri.rw>>1;
						goto ui;
					}
				i = rarcmul(int64hi(rri.y0),rri.rw);
				(void)int64add(rri.y0,rri.dy);
			ui:	i+= rri.mw, ic = subu(cx,i)>>16;
				adjustHLine1(f,o,ic);
				if (ic>0)
					f.rtn(o,ic,&f);
				o = (addu(cx,i)-1)>>16;
			}
		adjustHLine1(f,o,c);
		if (c>0)
			f.rtn(o,c,&f);
	} while(rEndLine(&f,f.e));
	c = 1;
end:	rEndBlt(&f);
	return c;
}

/* ---- high-level image transfer */

int rBlt(bltParam *f,
	int sp,
	int w,
	ifixed fh,
	int x,
	int sx,
	int sy,
	int ky)
{
	if (!f)
		return -1;
	if ((sx=-sx)>=0) /* mblt imaging works the source in terms of origin */
		x+= sx, w-= sx, sx = 0;
	sx+= x;
	if ((w+=x)>f->cl.right)
		w = f->cl.right;
	if (x<f->cl.left)
		x = f->cl.left;
	w-= x;
	if ((x-=f->ox)<0)
		w+= x, x = 0;
	if (w<=0)
		return 0;
	f->vx = sx-f->ox; /* origin fix; 2014.08.10 */
	if (fh<0)
	{
		f->cnv.s = (const int8*)f->cnv.s+(sy+fh)*sp, f->cnv.mp =-sp;
		f->a = f->e, f->oz =-f->oz;
		fh =-fh>>16;
		do
		{
			rEndLine(f,NULL);
			f->rtn(x,w,f);
		} while(--fh);
	} else
	{
		f->cnv.mp = sp, sy<<= 16, sy+= ky*fh;
		if (fh==0x10000)
		{
			f->cnv.s = (const int8*)f->cnv.s+(sy>>16)*sp;
			do
				f->rtn(x,w,f);
			while(rEndLine(f,f->e));
		} else
		{
			const int8	*s = f->cnv.s;

			sy+= fh>>1;
			do
			{
				f->cnv.s = s+(sy>>16)*sp;
				f->rtn(x,w,f);
			} while(sy+=fh,rEndLine(f,f->e));
		}
	}
	return 1;
}

int rBltPix(grafPort *p,
	pixMap *sm,
	int sx,
	int sy,
	int sw,
	int sh,
	int dx,
	int dy,
	int dw,
	int dh,
	uint8 mode)
{
	bltParam	f;
	const void	*s;
	int		fh, c;
	boolean		hl;

	if (!p || !sm || dw<=0 || dh<=0)
		return -1;
	hl = 0;
	//if (sy<0)
	//	dy-= sy, dh+= sy, sy = 0;
	if ((s=sm->baseAddr)==NULL)
		if ((s=lockPixels(sm))==NULL)
			return -1;
		else
			hl = 1;
	fh = (mode&~mSrcPat)|formPxFmt(sm)<<8;
	if (sw==dw)
		f.vx = 0x10000;
	else
		f.vx = sw<=1 ? 0:vfix16div(sw,dw), fh|= 1u<<31;
	if ((c=rInitBlt(&f,p,s,sm->pTable,dy,dy+dh,fh))<0)
		{c = 0; goto end;}
	/* detect overlap in case of move (i.e., scrolling) */
	fh = 0x10000;
	if (sh!=dh)
		fh = sh<=1 ? 0:vfix16div(sh,dh);
	else if (p->map==sm && dy>sy)
	{
		if ((f.vy=dy+dh)>f.cl.bottom)
			f.vy = f.cl.bottom;
		fh = (MAX(dy,f.cl.top)-f.vy)<<16;
	} else if ((f.vy=dy)<f.cl.top)
		f.vy = f.cl.top;
	rBlt(&f,sm->pitch,dw,fh,dx,sx,sy,c);
	c = 1;
end:	rEndBlt(&f);
	if (hl)
		unlockPixels(sm);
	return c;
}
