/*
 mimg.c
  mgraf imaging

  Revised: 2014.08.31
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.07
  Back & forth, realized colors must be 48-bit;
otherwise, QD-loaded colors won't match in gray conversion.
  Implemented rgbToPenColor and a full color & gray makeITable.
2014.04.13
  Fixed gray inverse table building.
2014.04.21
  Added a simple fix to get the closest cyan for inverse tables.
2014.04.23
  Gray inverse tables are now 10-bit, meeting 30-bit precision standards.
This allows for gray tables that cover smaller 16-bit ranges.
*/

#include <string.h>

#if TARGET_OS_MAC
 #define hNew(z)	NewHandle(z)
 #define hLock(h)	(HLock(h),*(int8**)(h))
 #define hUnlock(h)	HUnlock(h)
 #define hFree(h)	DisposeHandle(h)
 #define FORCE_4_ON_2	0
#else
 #include <windows.h>
 #define hNew(z)	LocalAlloc(LMEM_MOVEABLE,z)
 #define hLock(h)	LocalLock(h)
 #define hUnlock(h)	LocalUnlock(h)
 #define hFree(h)	LocalFree(h)
 #define FORCE_4_ON_2	1
#endif

#include "mgraf.h"
#include "../my_cseries.h"
#include "../ftrig.h"

/* ---- globals */

const uint32	pSizeMasks[32] =
{
	1,~0,0x55555555,0x11111111,0x11111111,0x1010101,0x1010101,0x1010101,
	0x1010101,0x10001,0x10001,0x10001,0x10001,0x10001,0x10001,0x10001,
	0x10001,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1
};
const uint32	pMCmpMasks[16] =
{ /* reciprocal table: raised 16 bits */
	0x1010000,0xFFFF0000u,0x55550000,0x24922493,0x11110000,0x8420843,0x4103CF4,0x204060D,
	0x1010000,0x803FA0,0x400FC4,0x2003E1,0x1000F1,0x80039,0x4000D,0x20003
};

const uint8	pDepthX[32] =
	{5,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5};

const uint8	Mac1bppRGB12[2*2] =
	{0xF,0xFF};

/* ---- code */

#if TARGET_OS_MAC
 /* TODO! */
#else
 #define MAX_DIB_SIZE		(sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD))

int getMainScreenDepth(void);
int getMainScreenDepth(void)
{
	int8		b[MAX_DIB_SIZE];
	HDC		hdc;
	HBITMAP		hbm;
	BITMAPINFO	*info;
	int		bpp;

	info = memset(b,0,sizeof(b));
	info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

	hdc = GetDC(NULL);
	hbm = CreateCompatibleBitmap(hdc,1,1); /* 0,0 would produce a mono dib */

	/* first call fills the header (because it's empty) */
	/* the second polls the palette */
	GetDIBits(hdc,hbm,0,1,NULL,info,DIB_RGB_COLORS);
	GetDIBits(hdc,hbm,0,1,NULL,info,DIB_RGB_COLORS);

	DeleteObject(hbm);
	bpp = GetDeviceCaps(hdc,BITSPIXEL);
	ReleaseDC(NULL,hdc);

	if (info->bmiHeader.biBitCount==16
			&& info->bmiHeader.biCompression==BI_BITFIELDS
			&& *(int32*)&info->bmiColors==0x7C00)
		return 15;
	return info->bmiHeader.biBitCount ? info->bmiHeader.biBitCount
		:bpp; /* last resort */
}

static void setQuadGrays(RGBQUAD *d,
	int c)
{
	int		x, dx;

	if (c<=0)
		return;
	dx = (255u<<16)/(c-1), x =~0;
	do
		d->rgbRed = d->rgbGreen = d->rgbBlue = x>>16;
	while(x-=dx,++d,--c);
}

BITMAPINFO *pixMapToNativeDesc(pixMap *m)
{
	BITMAPINFO	*info;
	uint32		pitch;

	if (!m || !m->pSize || (info=calloc(sizeof(BITMAPINFO)
			+(m->pSize>8 ? 3:256)*sizeof(RGBQUAD),1))==NULL)
		return NULL;
	pitch = m->pitch&0x7FFFFFFF;
	info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	info->bmiHeader.biWidth = (pitch<<3)/m->pSize;
	info->bmiHeader.biHeight =-m->h; /* negative for top-down bitmap */
	info->bmiHeader.biPlanes = 1;
	info->bmiHeader.biSizeImage = m->h*pitch;
	info->bmiHeader.biBitCount = m->pSize;
	if (m->pSize<9)
	{
		if (m->pTable)
			setQuadColors(info->bmiColors,m->pTable->colors,m->pTable->count);
		else
			setQuadGrays(info->bmiColors,1u<<m->bpp);
		return info;
	}
	/* can really only fix endian mismatch here for 24/32-bit */
	if (!((m->pType^!BIG_ENDIAN<<pTypeLSBBit)&(pTypeLSB|pTypeLGreen)))
		if ((m->pSize==16 && m->minCmp==5)
				|| (m->pSize==32 && m->minCmp==8))
			return info;
	info->bmiHeader.biCompression = BI_BITFIELDS;
	*(int32*)&info->bmiColors = RGBToPen(m,chromo_colors+_rgb_red);
	*(int32*)&info->bmiColors[1] = RGBToPen(m,chromo_colors+_rgb_green);
	*(int32*)&info->bmiColors[2] = RGBToPen(m,chromo_colors+_rgb_blue);
	return info;
}
#endif

/* pixmap */

pixMap *newPixMap(int bpp,
	int width,
	int height)
{
#ifdef WINUSERAPI
	HDC		hdc;
#endif
	pixMap		*m;

	if ((m=calloc(sizeof(*m),1))==NULL)
		return NULL;
#ifdef WINUSERAPI
	hdc = GetDC(NULL);
	m->hRes = GetDeviceCaps(hdc,LOGPIXELSX)<<16;
	m->vRes = GetDeviceCaps(hdc,LOGPIXELSY)<<16;
	if (width<0)
		width = GetDeviceCaps(hdc,HORZRES);
	if (height<0)
		height = GetDeviceCaps(hdc,VERTRES);
	ReleaseDC(NULL,hdc);
	if (bpp<=0)
		bpp = getMainScreenDepth();
#endif
	if (!m->vRes)
		m->vRes = hvResDefault;
	if (!m->hRes)
		m->hRes = hvResDefault;
	m->h = height, m->w = width;
	if ((m->bpp=bpp)==32)
		m->pSize = 32, bpp = 8, m->pType|= pTypeDirect|pTypeAlpha|!BIG_ENDIAN<<pTypeLSBBit;
	else if (bpp>8)
		m->pSize = bpp>16 ? 32:16, bpp/= 3u, m->pType|= pTypeDirect|!BIG_ENDIAN<<pTypeLSBBit;
	else
		m->pSize = bpp>4 ? 8:bpp>(FORCE_4_ON_2 ? 1:2) ? 4:MAX(bpp,0);
	m->minCmp = bpp;
	if (width>0)
		m->pitch = ((width*m->pSize>>3)+3)&~3;
	return m;
}

int disposePixMap(pixMap *m)
{
	int		r;

	if (!m)
		return -1;
	r = 0;
	if (m->iTable)
		{free(m->iTable); m->iTable = NULL, r+= 4;}
	if (m->pTable)
		{free(m->pTable); m->pTable = NULL, r+= 2;}
	r+= pmFree(m);
	free(m);
	return r;
}

int pmAllocate(pixMap *m)
{
	if (!m)
		return -1;
	if (!m->baseAddr && !m->pHand && m->h>0 && m->pitch>0)
		if ((m->pHand=hNew(m->pitch*m->h))!=NULL)
			return 1;
	return 0;
}

int pmFree(pixMap *m)
{
	if (!m || m->pitch<=0)
		return -1;
	if (m->pHand)
		{hFree(m->pHand); m->pHand = NULL; return 1;}
	m->baseAddr = NULL;
	return 0;
}

int8 *lockPixels(pixMap *m)
{
	void		*a;

	if (!m || m->pitch<=0)
		return NULL; /* invalid parameter or read-only */
	if (m->baseAddr || !m->pHand)
		return m->baseAddr;
	if ((a=hLock(m->pHand))==NULL)
		return NULL;
	return m->baseAddr=a;
}

int unlockPixels(pixMap *m)
{
	void		*a;

	if (!m)
		return -1; /* invalid parameter */
	if ((a=m->pHand)==NULL)
		return 0;
	m->pHand = a, m->baseAddr = NULL;
	hUnlock(a);
	return 1;
}

/* ---- image filtering */

int pmErase(pixMap *m,
	uint32 penc)
{
	uint32		*d;
	int32		z, c;

	if (!m || m->pitch<=0)
		return -1;
	if ((d=(uint32*)m->baseAddr)==NULL || (z=m->h)<=0 || (z*=m->pitch)<=0)
		return 0;
	penc&= ((2u<<(m->bpp-1))-1)*pSizeMasks[m->pSize&31];
	if (m->pSize<2)
		{memset(d,penc,z); return 1;}
	if ((c=z>>6)!=0)
		do
			d[15] = d[14] = d[13] = d[12] =
				d[11] = d[10] = d[9] = d[8] =
				d[7] = d[6] = d[5] = d[4] =
				d[3] = d[2] = d[1] = *d = penc;
		while(d+=16,--c);
	if ((c=z>>3&7)!=0)
		do
			d[1] = *d = penc;
		while(d+=2,--c);
	if (z&4)
		*d++ = penc;
	if (z&2)
		*(int16*)d = (int16)penc;
	return 1;
}

int pmFilter(pixMap *m,
	const cTab64 *ct,
	int c)
{
	uint32		a, kb, kg, kr;
	int		sl, w, h, i, j;

	if (!m || m->pitch<=0)
		return -1;
	if (!ct || !m->baseAddr || (h=m->h)<=0 || (w=m->w)<=0 || (sl=m->pitch)<w)
		return 0;
	if (c+0u<h)
		h = c;
	if (m->pSize<9)
	{
		if (m->pSize<8)
			return -8;
		if (m->pSize<9 && !m->iTable && pmMakeITables(m,m->bpp)<=0)
			return -2;
	}
	i = m->minCmp ? m->minCmp:10, j = i+(m->pType>>pTypeLGreenBit&1);
	kb = (1u<<i)-1, kg = (1u<<j)-1;
	if (ct->count<=kg)
		return -1;
	kg<<= i, j+= i, kr = kb<<j;
	/* the table builder ends up taking all the time when the area is small */
	if (m->pSize<9)
	{
		uint8		st[1024];
		const uint8	*it = m->iTable;
		uint16		*t;
		uint8		*d;

		if ((t=(uint16*)rGenRemappingTable(m,ct,m->bpp,ccIndexed,0))==NULL)
			return -2;
		sl-= w;
		if (m->minCmp)
		{
			const uint16	*dt = (const uint16*)(it+(1u<<m->minCmp*3));
			uint32		r;

			c = 256, d = st;
			do
			{
				a = *dt;
				r = t[a&kb]&kb;
				r|= t[(a&kg)>>i]&kg;
				r|= t[(a&kr)>>j]&kr;
				*d = (uint8)it[r];
			} while(++d,++dt,--c);
		} else
		{
			const uint16	*dt = (const uint16*)(it+1024);

			c = 256, d = st;
			do
				*d = (uint8)it[t[*dt>>6]>>6];
			while(++d,++dt,--c);
		}
		free(t); t = NULL;
		d = (uint8*)m->baseAddr;
		do
		{
			if ((c=w)&1)
				*d = st[*d], ++d;
			if (c>>=1)
				do
					*d = st[*d], d[1] = st[d[1]];
				while(d+=2,--c);
		} while(d+=sl,--h);
	} else if (m->pSize<17)
	{
		uint16		*t, *d = (uint16*)m->baseAddr, r;

		if ((t=(uint16*)rGenRemappingTable(m,ct,16,0,0))==NULL)
			return -2;
		sl-= w*2;
		do
		{
			c = w;
			do
			{
				a = *d;
				r = t[a&kb]&kb;
				r|= t[(a&kg)>>i]&kg;
				r|= t[(a&kr)>>j]&kr;
				*d = (uint16)r;
			} while(++d,--c);
		} while(((int8*)d)+=sl,--h);
		free(t);
	} else
	{
		uint32		*t, *d = (uint32*)m->baseAddr, r;

		if ((t=rGenRemappingTable(m,ct,32,0,1))==NULL)
			return -2;
		sl-= w*4;
		do
		{
			c = w;
			do
			{
				a = *d;
				r = t[a&kb]&kb;
				r|= t[(a&kg)>>i]&kg;
				r|= t[(a&kr)>>j]&kr;
				*d = r;
			} while(++d,--c);
		} while(((int8*)d)+=sl,--h);
		free(t);
	}
	return 1;
}

/* ---- graphics ports */

int openGrafPort(grafPort *p,
	int bpp,
	int width,
	int height,
	boolean allocate)
{
	if (!p)
		return -1;
	memset(p,0,sizeof(*p));
	if ((p->map=newPixMap(bpp,width,height))==NULL)
		return 0;
	p->clip.bottom = height, p->clip.right = width;
	p->penh = p->penw = 0x10000;
	RGBFore(p,&rgb_black);
	RGBBack(p,&rgb_white);
	if (allocate)
		return pmAllocate(p->map);
	return 1;
}

int closeGrafPort(grafPort *p)
{
	if (!p)
		return -1;
	if (p->map)
		{disposePixMap(p->map); p->map = NULL;}
	rPenPat(p,NULL);
	p->penw = MK4CC('D','E','A','D');
	return 1;
}

/* table generation */

int pmDefaultMacTable(pixMap *m,
	boolean gray)
{
	const uint8	*t;
	int		bpp, count;

	if (!m)
		return -1;
	if (m->pType&(pTypeFixed|pTypeDirect) || (bpp=m->bpp)>8)
		return 0;
	if (m->pTable)
		{free(m->pTable); m->pTable = NULL;}
	if (m->iTable)
		{free(m->iTable); m->iTable = NULL;}
	if (gray)
		bpp|= 32;
	if (bpp==40 || bpp==36 || bpp==34 || bpp==33)
		t = NULL;
	else if (bpp==8)
		t = Mac8bppRGB12;
	else if (bpp==4)
		t = Mac4bppRGB24, bpp+= 16;
	else if ((bpp&15)==2)
		t = Mac2bppRGB24, bpp+= 16;
	else if ((bpp&15)==1)
		t = Mac1bppRGB12;
	else
		return 0;
	count = 1u<<(bpp&15);
	if ((m->pTable=malloc(8+8*count))==NULL)
		return -2;
	if (!t)
		ctMakeGrays(m->pTable->colors,count);
	else if (bpp&16)
		expandRGB24(t,m->pTable->colors,count,0);
	else
		expandRGB12(t,m->pTable->colors,count,0);
	m->pTable->seed = GetCTSeed();
	m->pTable->flags = 0x8000;
	m->pTable->count = count;
	return 1;
}

int pmMakeITables(pixMap *m,
	int r)
{
	if (!m || m->bpp>8)
		return -1;
	m->pType&=~pTypeLGreen;
	if ((r=makeITables(m->pTable,&m->iTable,m->bpp,r,0))<0)
		return r;
	m->minCmp = r;
	return 1;
}

int copyCTable(cTab64 **out,
	const cTab64 *in)
{
	void		*p;
	int32		z;

	if (!out)
		return -1;
	if (!in)
	{
		if (*out)
			{free(*out); *out = NULL;}
		return 0;
	}
	z = sizeof(cTab64)-sizeof(cSpec)+sizeof(cSpec)*PIN(in->count,0,256);
	if ((p=realloc(*out,z))==NULL)
	{
		if (!z)
			{*out = NULL; return 0;}
		return -2;
	}
	*out = p;
	memcpy(p,in,z);
	return 1;
}

/* color selection */

int pmGetRGBCC(pixMap *m,
	long int *gt)
{
	if (!m)
		return -1;
	if (m->pType&pTypeDirect)
	{
		*gt = m->minCmp+(m->pType>>pTypeLGreenBit&1);
		return (m->pType^!BIG_ENDIAN<<pTypeLSBBit)&pTypeLSB
			? ccDirectBS32+(m->pSize>>4&1):ccDirect;
	} else if ((*gt=(long int)m->iTable)==0)
	{
		if (pmMakeITables(m,m->bpp)<=0)
			return -2;
		*gt = (long int)m->iTable;
	}
	return ccIndexed|ccNative|(!!m->minCmp);
}

uint32 RGBToPen(pixMap *m,
	const rgb48 *s)
{
	long int	gt;
	int		x;

	if ((x=pmGetRGBCC(m,&gt))<0)
		return 0;
	return rgbCCRoutine[x](m->minCmp,gt,s)*pSizeMasks[m->pSize&31];
}

int PenToRGB(pixMap *m,
	uint32 penc,
	rgb48 *v)
{
	const cTab64	*t;
	uint32		i, j, k;

	if (!m)
		return -1;
	i = m->minCmp&15;
	if (m->pType&pTypeDirect)
	{
		k = (1u<<i)-1, j = i+(m->pType>>pTypeLGreenBit&1);
		v->blue = (uint16)((penc&k)*pMCmpMasks[i]>>16);
		penc>>= i;
		v->green = (uint16)((penc&(1u<<j)-1)*pMCmpMasks[j]>>16);
		v->red = (uint16)((penc>>j&k)*pMCmpMasks[i]>>16);
	} else if ((t=m->pTable)==NULL)
	{
		if (pmMakeITables(m,m->bpp)<=0)
			return -2;
		v->blue = v->green = v->red = ((uint16*)m->iTable)[(penc&(1u<<m->bpp)-1)+512];
	} else
	{
		i = penc&(1u<<m->bpp)-1;
		if (t->flags&0x8000)
		{
			if (i>=t->count)
				return 0; /* index out of range! */
		} else
		{
			for(j=0; j<t->count; ++j)
				if (t->colors[j].value==i)
					goto ixc;
			return 0; /* not in the table */
		}
	ixc:	*v = t->colors[i].rgb;
	}
	return 1;
}

int getPixelRGB(pixMap *m,
	int x,
	int y,
	rgb48 *v)
{
	const void	*s;
	uint32		u;
	boolean		hl;

	if (!m)
		return -1;
	if ((x-=m->pOrigin.x)+0u>=m->w || (y-=m->pOrigin.y)+0u>=m->h)
		return 0;
	hl = 0;
	if ((s=m->baseAddr)==NULL)
		if ((s=lockPixels(m))==NULL)
			return -1;
		else
			hl = 1;
	s = m->baseAddr+y*m->pitch;
	if (m->pSize==32)
	{
		u = ((const uint32*)s)[x];
		if ((m->pType^!BIG_ENDIAN<<pTypeLSBBit)&pTypeLSB)
			u = _reverse32(u);
	} else if (m->pSize==16)
	{
		u = ((const uint16*)s)[x];
		if ((m->pType^!BIG_ENDIAN<<pTypeLSBBit)&pTypeLSB)
			u = _reverse16(u);
	} else
		u = pDepthX[m->pSize], u = ((const uint8*)s)[x<<u>>3]>>(~x<<u&7);
	if (hl)
		unlockPixels(m);
	return PenToRGB(m,u,v);
}

static void setSolid(grafPort *p,
	const rgb48 *v,
	int k)
{
	pixPat		*pat;
	uint32		c;

	if (!p || !v)
		return;
	(&p->fgColor)[k&1] = *v;
	(&p->fgSolid)[k&1] = RGBToPen(p->map,v);
	if ((pat=p->pPat)==NULL || (c=pat->ct.count-1)>255u || !(pat->ct.flags&0x8000))
		return;
	pat->ct.seed = 0; /* only need to invalidate one color...but I'm tired */
	pat->ct.colors[c&~k].rgb = *v;
}

void RGBFore(grafPort *p,
	const rgb48 *v)
{
	setSolid(p,v,0);
}

void RGBBack(grafPort *p,
	const rgb48 *v)
{
	setSolid(p,v,~0);
}

/* ---- patterns */

int rMakePat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *fore,
	const rgb48 *back)
{
	pixPat		*pat;

	if (!outpat)
		return -1;
	if ((pat=*outpat)!=NULL)
	{
		free(pat->px);  pat->px = NULL;
		free(pat->t); pat->t = NULL;
	}
	if (!inpat)
	{
		if (pat)
			free(pat);
		*outpat = NULL;
		return 0;
	}
	if ((pat=realloc(pat,sizeof(pixPat)+sizeof(cSpec)))==NULL)
		return -2;
	*outpat = pat;
	pat->t = NULL, pat->pitch = pat->h = pat->w = 8;
	pat->ct.seed = 0, pat->ct.count = 2, pat->ct.flags = 0x8000;
	if ((pat->px=malloc(64))==NULL)
		return -2;
#ifdef TEST_BITMAPTORGN
	{
		rect16		bounds;

		*(int32*)&bounds.top = 0x10001;
		*(int32*)&bounds.bottom = 0x70007;
		free(bitmapToRgn(inpat,1,&bounds));
	}
#endif
	b1to8(pat->px,inpat,8);
	if (!fore)
		fore = &rgb_black;
	if (!back)
		back = &rgb_white;
	*(int16*)&pat->ct.colors = 0;
	pat->ct.colors->rgb = *back;
	*(int16*)&pat->ct.colors[1] = be16_native(1);
	pat->ct.colors[1].rgb = *fore;
	return 1;
}

void rPenPat(grafPort *p,
	const uint8 inpat[8])
{
	rMakePat(&p->pPat,inpat,&p->fgColor,&p->bkColor);
}

int rCopyPat(pixPat **out,
	const pixPat *in)
{
	int		z;

	if (!in)
		return 0;
	z = sizeof(pixPat)-sizeof(cSpec)+sizeof(cSpec)*in->ct.count;
	if ((*out=malloc(z))==NULL)
		return -2;
	z-= OFFSET(pixPat,w);
	memcpy((int8*)*out+OFFSET(pixPat,w),in+OFFSET(pixPat,w),z);
	(**out).t = NULL; /* redundant data */
	z = in->h*in->pitch;
	if (((**out).px=malloc(z))==NULL)
		return -2;
	memcpy((**out).px,in->px,z);
	return 1;
}

int rSavePen(grafPort *p,
	penState *state)
{
	if (!p || !state)
		return -1;
	state->penw = p->penw;
	state->penh = p->penh;
	state->pnLoc = p->pnLoc;
	state->pnMode = p->pnMode;
	state->pPat = NULL;
	if (!p->pPat)
		return 0;
	return rCopyPat(&state->pPat,p->pPat);
}

int rPenNormal(grafPort *p,
	penState *oldstate)
{
	int		r;

	if (!p)
		return -1;
	if (oldstate)
	{
		oldstate->penw = p->penw;
		oldstate->penh = p->penh;
		oldstate->pnLoc = p->pnLoc;
		oldstate->pnMode = p->pnMode;
		oldstate->pPat = p->pPat;
		r = !!p->pPat;
	} else
		r = rDisposePat(p->pPat);
	p->penw = p->penh = 0x10000;
	p->pnMode = mSrcCopy;
	p->pPat = NULL;

	return r;
}

int rRestorePen(grafPort *p,
	penState *state,
	boolean freeState)
{
	int		r;

	if (!state)
		return -1;
	if (p)
	{
		p->penw = state->penw;
		p->penh = state->penh;
		p->pnLoc = state->pnLoc;
		p->pnMode = state->pnMode;
		r = !!p->pPat;
		rPenPat(p,NULL);
		if (freeState)
		{
			p->pPat = state->pPat;
			return r;
		}
		return rCopyPat(&p->pPat,state->pPat);
	} else if (freeState)
		return rDisposePat(state->pPat);
	return -1;
}

/* ---- color conversion */

DEF_RGB_RTN(RGBGray)
{
	uint32		u;

	return (formGray(s->red,s->green,s->blue,u)^0x1FFFF)<<i>>17;
}
DEF_RGB_RTN(RGBDirect)
	{return formDirect(s,gt,i);}
DEF_RGB_RTN(RGBBDirect32)
{
	uint32		v = formDirect(s,gt,i);

	return _reverse32(v);
}
DEF_RGB_RTN(RGBBDirect16)
{
	uint32		v = formDirect(s,gt,i);

	return _reverse16(v);
}
DEF_RGB_RTN(RGBGray8)
{
	uint32		u;

	return formGray(s->red,s->green,s->blue,u)>>1;
}
DEF_RGB_RTN(RGBDirect8)
	{return formDirect(s,i,i);}
DEF_RGB_RTN(RGBGray8X)
{
	uint32		u;

	return ((const uint8*)gt)[formGray(s->red,s->green,s->blue,u)>>7];
}
DEF_RGB_RTN(RGBDirect8X)
	{return ((const uint8*)gt)[formDirect(s,i,i)];}

const rgbRoutine	rgbCCRoutine[8] =
{
	RGBGray, RGBDirect, RGBBDirect32, RGBBDirect16,
	RGBGray8, RGBDirect8, RGBGray8X, RGBDirect8X
};

/* ---- image loading */

#include <stdio.h>

pixMap *decompressPictFile(const char *path,
	int *err)
{
	FILE		*f = NULL;
	uint8		*b;
	pixMap		*m = NULL;
	long int	z;
	int		r = 0;

	if (!path)
		{r =-1; goto end;}
	if ((f=fopen(path,"rb"))==0)
		{r =-43; goto end;}
	if (fseek(f,0,SEEK_END))
		{r =-3; goto end;}
	if ((z=ftell(f)-512)<=10)
		goto end; /* too small for a PICT file */
	fseek(f,0,SEEK_SET);
	if ((b=calloc(MAX(z,512),1))==NULL)
		{r =-2; goto end;}
	if (fread(b,512,1,f)<=0)
		{r =-3; goto end;}
	{
		const int32	*s = (const int32*)b;
		int		c = 512/16;

		do
			if (*s|s[1]|s[2]|s[3])
				{r =-8; goto end;} /* not a PICT file */
		while(s+=4,--c);
	}
	if (fread(b,z,1,f)<=0)
		r =-3;
	m = renderQDPict(b,z);
	free(b);
end:	if (f)
		fclose(f);
	if (err)
		*err = r;
	return m;
}
