/*
 mrgn.c

   Copyright (C) 2014 Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.15
  Split from mgraphic.
2014.05.23/4
  Added rectInRgn (tested, working), coordinate removal in rgnCompact (untested).
*/

#include <stdlib.h>
#include <string.h>

#include "mgraf.h"
#include "../ftrig.h"

/* ---- code */

static SLRgn *rgnAppend(SLRgn *rgn,
	const int16 *a,
	int az)
{
	void		*p;
	long		z;

	z = rgn->size+az;
	if ((p=realloc(rgn,z))==NULL)
		{free(rgn); return NULL;}
	rgn = p, rgn->size = z;
	memcpy((int8*)rgn+z-az,a,az);
	return rgn;
}

SLRgn *newRgn(void)
{
	SLRgn		*rgn;

	if ((rgn=malloc(OFFSET(SLRgn,x)))==NULL)
		return NULL;
	rgn->bBox.top = rgn->bBox.left = IMAX(rgn->bBox.left);
	rgn->bBox.bottom = rgn->bBox.right = IMIN(rgn->bBox.right);
	rgn->size = OFFSET(SLRgn,x);
	return rgn;
}

int rgnXorLine(SLRgn **rrgn,
	int16 ctop,
	int16 cbot,
	pt2dq p0,
	pt2dq p1)
{
	SLRgn		*rgn;
	const int16	*f, *s, *e;
	int16		*a;
	int64		x, dx;
	int32		o, c, z, newa, newb, newc;
	int32		y0i, y1i, x0i, x1i;

	if (!rrgn)
		return -1;
	if ((rgn=*rrgn)==NULL)
		if ((rgn=newRgn())==NULL)
			return -2;
		else
			*rrgn = rgn;
	q2dsort2(p0,p1);
	if ((y0i=p0.y>>16)<ctop)
		y0i = ctop;
	if ((z=rgn->size)<OFFSET(SLRgn,x) || z&1
			|| rgn->bBox.bottom<=rgn->bBox.top)
	{
		rgn->size = z = OFFSET(SLRgn,x);
		rgn->bBox.top = rgn->bBox.bottom = y0i;
	}
	if ((y1i=p1.y>>16)>cbot)
		y1i = cbot;
	if ((c=y1i-y0i)<=0)
		return 0; /* no addition */
	dx = rslopdiv(&x,p0.x,p0.y,p1.x,p1.y,y1i-(p0.y>>16)); /* from end to work backward */
	o = z;
	/* from-end insertion requires the new size be adequate */
	z+= c*sizeof(int16); /* inserted positions */
	if ((newa=rgn->bBox.top-y0i)>0) /* new lines above */
		z+= newa*(2*sizeof(int16));
	if ((newb=y1i-rgn->bBox.bottom)>0) /* new lines below */
		z+= newb*(2*sizeof(int16));
	if ((newc=MIN(rgn->bBox.bottom,y1i)-MAX(rgn->bBox.top,y0i))<0) /* middle intersection */
	{ /* gap - extents will be longer than actual */
		c = newb>>31, newb+= (newc^c)-c;
		c = newa>>31, newa+= (newc^c)-c;
	}
	if ((rgn=realloc(rgn,z))==NULL)
		return -2;
	*rrgn = rgn, rgn->size = z;
	/* adding from end to perform insertion on same & only buffer */
	f = rgn->x; /* cannot run before this (would be erroneous) */
	e = a = (int16*)(z+(int8*)rgn); /* new write */
	s = (int16*)(o+(int8*)rgn); /* old read */
	if (p0.x>p1.x)
		o = p1.x, c = p0.x;
	else
		o = p0.x, c = p1.x;
	/* fix bounds */
	if ((x0i=o>>16)<rgn->bBox.left)
		rgn->bBox.left = x0i;
	if ((x1i=c>>16)>rgn->bBox.right)
		rgn->bBox.right = x1i;
	if (y0i<rgn->bBox.top)
		rgn->bBox.top = y0i;
	if (y1i>rgn->bBox.bottom)
		rgn->bBox.bottom = y1i;
	/* below */
	if (newb<0) /* copy only */
		do
		{
			*--a = *--s; /* end taken unconditionally */
			if (a<=f || s<f)
				goto bad;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(++newb);
	else if (newb) /* append */
		do
		{
			if ((a-=3)<f)
				goto bad;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*a = IMIN(*a), a[1] = o, a[2] = IMAX(*a);
		} while(--newb);
	if (newc<0) /* hollow non-intersection */
		do
		{
			if ((a-=2)<f)
				goto bad;
			*a = IMIN(*a), a[1] = IMAX(*a);
		} while(++newc);
	else if (newc) /* insertion */
		do
		{
			if (a<f || s<=f)
				goto bad;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			/* endings require neg.-space, and xs can be IMIN/IMAX */
			/* be sure not to remove the start/end when compacting */
			*--a = *--s; /* end taken unconditionally */
			while(s[-1]>o)
				*--a = *--s;
			/* coming before less or same */
			if (a<=f || s<f)
				goto bad;
			if (*--s!=o || s<=f || s[-1]>o) /* compact */
				*--a = o, *--a = *s;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(--newc);
	if (newa<0) /* copy only */
		do
		{
			*--a = *--s; /* end taken unconditionally */
			if (a<=f || s<f)
				goto bad;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(++newa);
	else if (newa) /* append */
		do
		{
			if ((a-=3)<f)
				goto bad;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*a = IMIN(*a), a[1] = o, a[2] = IMAX(*a);
		} while(--newa);
	if (a<f)
	{ /* corrupt */
	bad:	rgn->bBox.top = rgn->bBox.bottom = 0;
		return -1;
	}
#ifndef MINIMAL_RGNCOMPACT
	while(a<e && urw32(a)==(BIG_ENDIAN ? 0x80007FFFu:0x7FFF8000))
		++rgn->bBox.top, a+= 2; /* top shouldn't ovfl */
	while(a+2<e && urw32(e-2)==(BIG_ENDIAN ? 0x80007FFFu:0x7FFF8000))
		--rgn->bBox.bottom, e-= 2; /* bottom shouldn't ovfl */
#endif
	c = (int8*)e-(int8*)a;
	z = c+OFFSET(SLRgn,x);
	if (a!=f)
	{ /* compacting eliminated leading lines */
		memcpy(rgn->x,a,c);
#ifndef MINIMAL_RGNCOMPACT
		goto cpct;
	} else if (rgn->size!=z)
	{
	cpct:	if ((rgn=realloc(rgn,z))!=NULL || !z)
			*rrgn = rgn;
		rgn->size = z;
#endif
	}
	return 1;
}

int ptInRgn(SLRgn *rgn,
	int x,
	int y)
{
	const int16	*s, *e;
	int32		o, c;

	if (!rgn || (c=rgn->size)<OFFSET(SLRgn,x) || c&1)
		return -1;
	if (x<rgn->bBox.left || x>=rgn->bBox.right)
		return 0;
	s = rgn->x, e = (const int16*)(rgn->size+(const int8*)rgn);
	if (e<=s || MAX(rgn->bBox.top,y)>=rgn->bBox.bottom
			|| (c=y-rgn->bBox.top)<0)
		return 0;
	--e;
	if (c) /* skip to applicable line for point */
		do
		{
			if (s>=e || s[1]<*s)
				return -1;
			for(++s; *s>=s[-1]; ++s)
				;
		} while(--c);
	if (++s>=e || *s<s[-1])
		return -1;
	do
	{
		if (s>=e)
			return -1;
		if ((c=s[1])<(o=*s))
			break;
		if (x>=o && x<c)
			return 1;
	} while((s+=2)<e && *s>=s[-1]);
	return 0;
}

int rectInRgn(SLRgn *rgn,
	int left,
	int top,
	int right,
	int bottom)
{
	const int16	*s, *e;
	int32		o, c, h;

	if (!rgn || (c=rgn->size)<OFFSET(SLRgn,x) || c&1)
		return -1;
	if (MAX(rgn->bBox.left,left)>=MIN(rgn->bBox.right,right))
		return 0;
	if (bottom>rgn->bBox.bottom)
		bottom = rgn->bBox.bottom;
	if (top<rgn->bBox.top)
		top = rgn->bBox.top;
	s = rgn->x, e = (const int16*)(rgn->size+(const int8*)rgn);
	if (e<=s || bottom<=top)
		return 0;
	--e;
	if ((c=top-rgn->bBox.top)!=0) /* skip to .top line */
		do
		{
			if (s>=e || s[1]<*s)
				return -1;
			for(++s; *s>=s[-1]; ++s)
				;
		} while(--c);
	h = bottom-top;
	do
	{
		if (++s>=e)
			return h==1 ? 0:-1;
		if (*s<s[-1])
			return -1;
		do
		{
			if (s>=e)
				return -1;
			if ((c=s[1])<(o=*s))
				{++s; break;}
			if (c>right)
				c = right;
			if (o<left)
				o = left;
			if (c>o)
				return 1;
		} while((s+=2)<e && *s>=s[-1]);
	} while(--h);
	return 0;
}

SLRgn *rgnCompact(SLRgn *rgn,
	boolean linesOnly)
{
	int16		*s, *e;
	int32		z, c;

	if (!rgn)
		goto end;
	z = rgn->size;
	if (rgn->bBox.bottom<=rgn->bBox.top || z&1)
		{z = OFFSET(SLRgn,x); rgn->bBox.bottom = rgn->bBox.top; goto set;}
	s = rgn->x, e = (int16*)(z+(int8*)rgn);
	/* coords shouldn't overflow (unless the data's corrupted) */
	while(s<e && urw32(s)==(BIG_ENDIAN ? 0x80007FFFu:0x7FFF8000))
		++rgn->bBox.top, s+= 2;
	while(s+2<e && urw32(e-2)==(BIG_ENDIAN ? 0x80007FFFu:0x7FFF8000))
		--rgn->bBox.bottom, e-= 2;
	if (!linesOnly)
	{ /* remove positive-range redundant coordinates */

	  /* * * this part UNTESTED * * */
		int16		v, *d, *a = s+1; /* skip line start */

		--e;
		while(a<e)
		{
			if (*a!=(v=a[1]))
			{
				a+= 1+(*a>v); /* skip next start */
				continue;
			}
			/* redundant; skip and start copying */
			d = a+=2;
			while(a<e)
			{
				if (*a!=(v=a[1]))
				{
					if (*a>v) /* line end and next start */
						*d++ = *a, ++a;
					*d++ = *a, ++a;
					continue;
				}
				a+= 2; /* redundant */
			}
			*d = *a, e = d+1; /* new end */
			break;
		}
	}
	z = (c=(int8*)e-(int8*)s)+OFFSET(SLRgn,x);

	if (s!=rgn->x)
	{
		memcpy(rgn->x,s,c);
		goto cpct;
	} else if (rgn->size!=z)
	{
		void		*p;

	cpct:	if (z<=OFFSET(SLRgn,x))
			{free(rgn); return NULL;}
		if ((p=realloc(rgn,z))!=NULL)
			rgn = p;
	}
set:	rgn->size = z;
end:	return rgn;
}

/* * * UNTESTED * * */
SLRgn *rgnOffsetXY(SLRgn *rgn,
	int dx,
	int dy)
{
	int16		*e, *s = rgn->x, *a;
	int32		z, c, v;

	if (!rgn || rgn->size<OFFSET(SLRgn,x) || rgn->size&1 || !(dx|dy))
		goto end;
	if (ABS(dx)>65535 || ABS(dy)>65535)
		goto empty;
	/* adjust bounds, clip */
	v = rgn->bBox.left+dx, rgn->bBox.left = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = rgn->bBox.right+dx, rgn->bBox.right = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = rgn->bBox.bottom+dy, rgn->bBox.bottom = v!=(int16)v ? v>>31^IMAX(int16):v;
	v = rgn->bBox.top+dy, rgn->bBox.top = v!=(int16)v ? v>>31^IMAX(int16):v;
	/* dispose if empty */
	if (rgn->bBox.right<=rgn->bBox.left || rgn->bBox.bottom<=rgn->bBox.top)
	{
	empty:	rgn->size = OFFSET(SLRgn,x);
		free(rgn);
		return NULL;
	}
	e = (int16*)(rgn->size+(const int8*)rgn);
	if (e<=s)
		goto end;
	--e;
	/* get first line in region */
	if ((c=IMIN(int16)-v)>0)
	{ /* skip lines clipped out */
		do
		{
			if (s>=e || s[1]<*s)
				goto bad;
			for(++s; *s>=s[-1]; ++s)
				;
		} while(--c);
	}
	a = s; /* offset the coordinates in the new bounds */
	c = RECTANGLE_HEIGHT(&rgn->bBox);
	do
	{
		t_count		oldx;

		if (s>=e || s[1]<*s)
			goto bad;
		for(oldx=*s; *++s>=oldx; )
		{
			if (s>=e)
			{
				if (v==1)
					break;
			bad:	rgn->size|= 1;
				rgn->bBox.bottom = rgn->bBox.top;
				goto end;
			}
			oldx = *a;
			v = oldx+dx, *a = v!=(int16)v ? v>>31^IMAX(int16):v;
		}
	} while(--c);
	z = (int8*)a-(int8*)s+OFFSET(SLRgn,x);
	if (s!=rgn->x)
		memcpy(rgn->x,s,z);
	if (rgn->size!=z)
	{
		void		*p;

		rgn->size = z;
		if ((p=realloc(rgn,z))!=NULL)
			rgn = p;
	}
	rgn = rgnCompact(rgn,0);
end:	return rgn;
}

SLRgn *bitmapToRgn(const uint8 *bm,
	int sp,
	const rect16 *bounds)
{
	int16		b[32], *a;
	SLRgn		*rgn;
	int		w, h, uh;

	if (!bm || sp<=0 || !bounds || (rgn=newRgn())==NULL)
		return NULL;
	rgn->bBox = *bounds;
	if ((w=RECTANGLE_WIDTH(bounds))<=0
			|| (h=RECTANGLE_HEIGHT(bounds))<=0)
		return rgn;
	bm+= sp*bounds->top+(bounds->left>>3);
	a = b, uh = bounds->left&7;

	/* the process is rather simple: add the x for each change, starting w. 'black' */
	do
	{
		const uint8	*s = bm;
		int		c, j, x;
		int8		v, u;

		if (a>b+GET_ARRAY_COUNT(b)-2)
		{
			if ((rgn=rgnAppend(rgn,b,(int8*)a-(int8*)b))==NULL)
				return NULL;
			a = b;
		}
		*a++ = IMIN(*a);
		c = w, x = bounds->left, u = 0;
		v = *s<<uh, j = 8-uh;
		goto bl;
		do
		{
			v = *s, j = 8;
		bl:	if (c<j)
				j = c;
			c-= j;
			do
				if ((int8)(u^v)<0)
				{
					u = v;
					if (a>b+GET_ARRAY_COUNT(b)-2)
					{
						if ((rgn=rgnAppend(rgn,b,(int8*)a-(int8*)b))==NULL)
							return NULL;
						a = b;
					}
					*a++ = x;
				}
			while(v+=v,++x,--j);
		} while(++s,c>0);
		*a++ = IMAX(*a);
	} while(bm+=sp,--h);
	/* it's impossible for pixels to produce multiple x-coordinates */
	return rgnCompact(rgnAppend(rgn,b,(int8*)a-(int8*)b),1);
}
