/*
 mtext.c
  text measurement, rendering, text boxes

   Copyright (C) 2014 Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.09
  Split from mfont.c.
  Put text font/style in grafPort for drawing case.
  Added chExtra; tabAlign (not in QDText).
2014.05.16
  Universal tVisLine, tBoxCharStyle.
  Adding char-to-pt, selection drawing code....not done.
2014.05.17
  A number of new and revised functions, getting mouse selection done.
  tBoxClick selecting with mouse.  Bugs in other functions stick out.
2014.05.19
  Left/right arrows, drag-select...starting to look real.
*/

#include <string.h>

#include "mgraf.h"
#include "mblt.h"
#include "mtext.h"

/* ---- macros, inlines */

__forceinline static int fixIndex(int i,
	const bmfMemory *fm)
{
	if (i>fm->lastChar || (i-=fm->firstChar)<0 || fm->widTable[i*2+1]<0)
		i = fm->lastChar-fm->firstChar+1;
	return i;
}

/* ---- globals */

/* underlining thickness (in pixels) - excluding neighbors on first line */
uint8		ft_ulThick = 1;

/* ---- measurement */

int rTextWidth(const bmfMemory *fm,
	const char *s,
	int z,
	int *itDf,
	ufixed x,
	ufixed spExtra,
	int8 styl)
{
	int		i, w, cw, sb;

	if (!fm || !fm->data || !fm->bmHeight)
		return -1;
	if (z<=0 && (!z || (z=strlen(s))==0))
		return 0;
	styl&=~fm->style, sb = styl&_style_bold, w = 0;
	do
	{
		if ((i=*(const uint8*)s)==' ')
			x+= spExtra;
		w+= x>>16, x&= 0xFFFF;
		if ((cw=fm->widTable[fixIndex(i,fm)*2+1])>0)
			w+= cw+sb;
	} while(++s,--z);
	if (itDf)
		*itDf = styl&_style_italic ? (-2+fm->leading-fm->ascent)>>1:0;
	return w;
}

int rClipText(const bmfMemory *fm,
	const char *tx,
	int z,
	ufixed *txw,
	ufixed w,
	ifixed tabAlign,
	ifixed tabOrigin,
	ufixed chExtra,
	ufixed spExtra,
	int8 styl,
	boolean visible)
{
	const char	*s;
	ufixed		maxw, lw;
	int		i, cw, sb;

	if (!fm || !fm->data || !fm->bmHeight || !txw)
		return -1;
	s = tx;
	if ((z<=0 && (!z || (z=strlen(s))==0)) || (maxw=*txw)==0)
		goto end;
	styl&=~fm->style, sb = styl&_style_bold;
	spExtra&=~(chExtra&IMIN(ifixed));
	do
	{
		lw = w, i = *(const uint8*)s, ++s;
		if (i==' ')
			w+= spExtra;
		else if (i==9 && tabAlign>0)
		{
			w+= tabAlign;
			if (w<lw)
				goto ovfl;
			w-= tabOrigin;
			w-= w%(ufixed)tabAlign;
			w+= tabOrigin;
			continue;
		}
		w+= chExtra;
		if (w<lw)
			goto ovfl;
		if ((cw=fm->widTable[fixIndex(i,fm)*2+1])>0)
			w+= (cw+sb)<<16;
		if (w<lw)
			{ovfl: w =-1u, --s; goto end;}
		if (w>maxw)
		{
			if (!visible)
				--s, lw = w; 
			break;
		}
	} while(--z);
end:	*txw = w;
	return s-tx;
}

int rTextBreak(const bmfMemory *fm,
	const char *start,
	const char *pos,
	int *length,
	ufixed *txw,
	ufixed w,
	ifixed tabAlign,
	ufixed chExtra,
	ufixed spExtra,
	int8 styl)
{
	const char	*s = pos, *ps;
	ufixed		maxw, lw, pw, nw;
	int		z, i, cw, sb, r = tBreakWord;

	if ((z=*length)>0 && (maxw=*txw)!=0)
	{
		spExtra&=~(chExtra&IMIN(ifixed));
		styl&=~fm->style, sb = styl&_style_bold;
		nw = pw = w, ps = NULL;
		do
		{
			lw = w, i = *(const uint8*)s, ++s;
			if (i-10u<4u)
			{ /* line-ending */
				r = tBreakLine;
				goto end;
			} else if (i==9 && tabAlign>0)
			{
				w+= tabAlign;
				if (w<lw)
					{w =-1u, r = tBreakOverflow; goto end;}
				w-= w%(ufixed)tabAlign;
				continue;
			}
			if ((cw=fm->widTable[fixIndex(i,fm)*2+1])>0)
				w+= (cw+sb)<<16;
			if (w<lw)
				{w =-1u, r = tBreakOverflow; goto end;}
			if (i==' ')
				w+= spExtra, ps = s, pw = nw;
			else
				nw = w;
			w+= chExtra;
			if (w<lw)
				{w =-1u, r = tBreakOverflow; goto end;}
		} while(w<=maxw && --z);
		if (!z)
			r = tNoBreak;
		else if (ps==s) /* walk forward to next non-space */
			for(w=nw; z && *s==' '; ++s,--z)
				;
		else if (ps && ps!=start+1)
			s = ps, w = pw; /* after last space covered */
		else
			--s, w = lw, r = tBreakChar;
	}
end:	*txw = w;
	*length = s-pos;
	return r;
}

/* ---- rendering */

static void bltUl(int8 *a,
	int w, /* must be >0 */
	int h,
	int y) /* must be <h */
{
	int		c = w, p;
	int8		l, v, u;

	if (y<=0)
	{
	//	__debugbreak();
		return;
	}
	++y, a+= w*y, p =-w;
	if (++y>=h)
		w = 0; /* height is too small for next-line reads */
	v = a[p]|*a|a[w], l = 0;
	while(--c)
		++a, u = a[p]|*a|a[w], a[-1]|= !(v|l|u), l = v, v = u;
	*a|= !(v|l);
	if (w && (y=ft_ulThick-1)>0)
		memset(a+1,1,w*y);
}

/* expects buffer to have enough capacity */
static int bltChar(const bmfMemory *fm,
	int8 *a,
	int dpitch,
	int i,
	int x, /* plus maxKerning */
	int sx,
	int sy,
	int8 styl)
{
	const uint16	*t;
	const uint8	*s;
	int		w, ow;

	if ((ow=fm->widTable[i*2+1])<=0)
		return 0;
	ow+= styl&_style_bold;
	s = fm->data, t = (uint16*)(s+fm->bmSize);
	w = t[i+1]-t[i]+(styl&_style_bold)-sx;
	if (w>0)
	{
		int		h = fm->bmHeight;

		x+= fm->widTable[i*2], s+= t[i]+sy*fm->pitch;
		do
		{
			int8		*d = x+a;
			int		c = w;

			if ((i=sx)<0)
			{
				if ((c+=i)<=0)
					goto skip;
				d-= i, i = 0;
			}
			if (c>(styl&_style_bold))
				*d|= s[i];
			if (--c>0)
			{
				if (c-=styl&_style_bold)
					if (styl&_style_bold)
						do
							++i, *++d|= s[i]|s[i-1];
						while(--c);
					else
						do
							*++d|= s[++i];
						while(--c);
				if (styl&_style_bold)
					d[1]|= s[i];
			}
		skip:	if (styl&_style_italic && sy&1)
				++sx, --w;
			s+= fm->pitch;
		} while(a+=dpitch,++sy<h);
	}
	return ow;
}

static void bltGray(int8 *b,
	int dp,
	int h,
	int odd)
{
	int32		pat = _be32_native(0x01000100)^(-odd&0x01010101);
	int		z = dp+3;

	do
	{
		int32		*d = (int32*)b;
		int		c;

		if ((c=z>>3)!=0)
			do
				*d+= *d&pat, d[1]+= d[1]&pat;
			while(d+=2,--c);
		if (z&4)
			*d+= *d&pat;
	} while(pat^=0x01010101,b+=dp,--h);
}

static int setGray(grafPort *p)
{ /* get the average for the fore and back */
  /* if we can't get a close enough color match, gray the text (in an opaque way) */
  /* note that penToRGB can still fail (out of range index) */
	rgb48		back, grayc, graym;
	pixMap		*m;

	if ((m=p->map)==NULL)
		return -1;
	if (PenToRGB(m,p->fgSolid,&grayc) && PenToRGB(m,p->bkSolid,&back))
	{
		uint32		penc;
		/* another POCC bug: stuck in 16-bit ops. simply doing `(a+b+1)>>1` */
		penc = (back.red+grayc.red+1)>>1;
		grayc.red = (uint16)penc;
		penc = (back.green+grayc.green+1)>>1;
		grayc.green = (uint16)penc;
		penc = (back.blue+grayc.blue+1)>>1;
		grayc.blue = (uint16)penc;
		penc = RGBToPen(m,&grayc);
		if (PenToRGB(m,penc,&graym)
				&& idim(grayc.red,graym.red)*5
				+idim(grayc.green,graym.green)*6
				+idim(grayc.blue,graym.blue)*4<0x28000)
		{ /* close enough */
			p->fgSolid = penc;
			return 0;
		}
	}
	return 1;
}

ifixed rDrawText(grafPort *p,
	const char *s,
	int z,
	int y,
	ifixed qx)
{
	bltParam	f;
	const bmfMemory *fm;
	char		*buf;
	uint32		oldpenc;
	ifixed		tabAlign, tabOrigin;
	ufixed		w, spExtra, chExtra;
	int		gray, i, x, h, ik, k, o, ox, dx, sp;
	int		sx, sy; /* source-x/y - may be altered for oblique adjustment */
	uint8		styl;

	if (!p || (fm=p->txFM)==NULL || !fm->data || !fm->bmHeight || ABS(y)>65535)
		return IMIN(ifixed);
	if (z<=0 && (!z || (z=strlen(s))==0))
		return qx;
	ox = qx, tabOrigin = 0;
	if ((tabAlign=p->tabAlign)>0)
		tabOrigin = p->tabOrigin%tabAlign;
	chExtra = p->chExtra, spExtra = p->spExtra&~(chExtra&IMIN(ifixed));

	/* bold/oblique adjustment for when the closest bmf doesn't match */
	styl = p->txStyle&~fm->style;
	sx = styl&_style_italic ? (fm->leading-fm->ascent-2)>>1:0;

	/* seek to first visibly affecting character */
	ik = MIN(sx,0), o = p->clip.left+ik-1-ABS(fm->maxKerning);
	o = o!=(int16)o ? o>>31^IMAX(int16):o<<16;
	if (qx<o)
	{
		ifixed		lw;
		int		cw, sb = styl&_style_bold;

		do
		{
			lw = qx, i = *(const uint8*)s, ++s;
			if (i==9 && tabAlign>0)
			{
				qx+= tabAlign;
				if (qx<lw)
					{--z; break;}
				qx-= tabOrigin;
				if ((i=qx%tabAlign)<0)
					i+= tabAlign;
				qx-= i;
				qx+= tabOrigin;
				continue;
			}
			if (qx<lw)
				{--z; break;}
			if ((cw=fm->widTable[fixIndex(i,fm)*2+1])>0)
				qx+= (cw+sb)<<16;
			if (i==' ')
				qx+= spExtra;
			qx+= chExtra;
		} while(--z && qx<o && qx>=lw);
		--s, ++z, qx = lw;
	}
	x = qx>>16;
	if (x+MIN(fm->maxKerning,0)>p->clip.right)
		return qx;
	y-= fm->ascent, h = 0;
	if (!ft_ulThick)
		styl&=~_style_underlined;
	else if (styl&_style_underlined)
		h = 1+ft_ulThick, h = fm->leading+fm->ascent+MAX(fm->descent,h);
	if (h<fm->bmHeight)
		h = fm->bmHeight;
	oldpenc = p->fgSolid;
	gray = p->txMode&mTextGray ? setGray(p):0;

	/* making use of pattern bit for fast remapping */
	if ((k=rInitBlt(&f,p,fm,NULL,y,y+h,p->txMode|mSrcPat))<0)
		{k = qx; goto end;}
	w = (f.cl.right-x)<<16; /* shouldn't overflow */

	/* adjusting the tab origin is tricky; could overflow */
	if (tabAlign>0)
		tabOrigin = (tabOrigin+(qx&~0xFFFF))%tabAlign;
	z = rClipText(fm,s,z,&w,qx&=0xFFFF,tabAlign,tabOrigin,chExtra,spExtra,styl,1);
	if (z<=0 || (w=(w>>16)-ik)<=0)
		{k = qx; goto end;} /* hollow */
	if ((sy=(k-3)&~1)<0)
		sy = 0;
	/* negative displacements allowed, used in some fonts */
	dx = x;
	if ((o=fm->maxKerning)<0)
		dx-= o, w-= o, o = 0;
	sp = (w+4+((chExtra+spExtra)>>16))&~3;
	if ((buf=calloc(sp,1+(h-=sy)))==NULL)
		{k =-2; goto end;}
	if (styl&_style_italic)
		sx+= sy>>1;
	if (tabAlign>0)
		do
		{
			i = *(const uint8*)s, ++s;
			if (i==9)
			{
				qx+= tabAlign;
				qx-= tabOrigin;
				qx-= (ufixed)qx%(ufixed)tabAlign;
				qx+= tabOrigin;
				continue;
			}
			qx+= bltChar(fm,buf,sp,fixIndex(i,fm),((ufixed)qx>>16)+o,sx,sy,styl);
			if (i==' ')
				qx+= spExtra;
			qx+= chExtra;
		} while(--z);
	else
		do
		{
			i = *(const uint8*)s, ++s;
			qx+= bltChar(fm,buf,sp,fixIndex(i,fm),((ufixed)qx>>16)+o,sx,sy,styl)<<16;
			if (i==' ')
				qx+= spExtra;
			qx+= chExtra;
		} while(--z);
	if (styl&_style_underlined)
		bltUl(buf,sp,h,fm->ascent-sy);
	f.cnv.s = buf, f.vx = f.vy = 0;

	if (gray)
		bltGray(buf,sp,h,(x^ox>>16)&1);
	rBlt(&f,sp,w-((styl&_style_italic)>>1),0x10000,dx,0,k-sy,0);
	if (x+(qx>>16)>IMAX(int16))
		k = IMAX(ifixed);
	else
		k = (x<<16)+qx;
	free(buf);
end:	p->fgSolid = oldpenc;
	rEndBlt(&f); /* not actually necessary, given the pattern mode is used */
	return k;
}
