/*
 mtext.h
  include for all built-in text-related code

   Created: 2014.03.30
   Revised: 2014.11.27
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __MTEXT__
 #define __MTEXT__
 #include "mgraf.h"

/* ---- constants */

enum
{
	tNoBreak,
	tBreakLine,
	tBreakWord,
	tBreakChar,
	tBreakOverflow
};

enum
{ /* tBox flags */
	tBoxActive		= 0x1,	/* user can select */
	tBoxModifiable		= 0x2,	/* otherwise read-only, no caret */
	tBoxAutoScroll		= 0x4,	/* view adjusts so for full caret visiblility */
	tBoxFrame		= 0x8,	/* draw edit-text frame */
	tBoxAutoVScroll		= 0x10, /* vert. scroll bar installed on right - tBoxNew */
	tBoxAutoHScroll		= 0x20, /* horiz. scroll bar installed on bottom - tBoxNew */
	tBoxDragDrop		= 0x40, /* can use mouse to drag & drop text */
	tHiliteLayer		= 0x80,	/* hilite before text; otherwise, after, inverting */
	tBoxListSelect		= 0x100, /* when read-only, select whole lines */
	tBoxInsertOnly		= 0x200, /* selection clipped to caret/click point */
	tBoxUseCustBgColor	= 0x400,
	tBoxUDirection		= 0x800, /* to avoid oscillation when selecting w. mouse */
	tBoxBDirection		= 0x1000,/* to avoid oscillation when selecting w. mouse */
	tBoxFullUpdate		= 0x2000,
	tBoxDragState		= 0x4000,
	tBoxCaretState		= 0x8000
};

/* ---- typedefs */

typedef struct bmfMemory
{
	uint8		flag, style;
	int16		size;
	uint16		firstChar, lastChar;
	int16		maxWidth, maxKerning;
	uint16		bmWidth, bmHeight;
	int32		bmSize, pitch;
	int16		ascent, descent, leading;
	uint8		*data;
	const int16	*widTable;
	/* data: location table (uint16*) follows (data+bmSize),
	  width table (int8*) directly follows
	  - 'undefined' char @ (lastChar-firstCharacter+1)*2
	*/
} bmfMemory;

typedef struct ffMemory
{
	int32		asscCount;
	bmfMemory	font[1];
} ffMemory;

typedef struct tStyleRun/*(16)*/
{
	int32		offs;
	uint16		font, size;
	uint8		style, special;
	rgb48		color;
} tStyleRun;

typedef struct tLineRun/*(16)*/
{
	int32		offs;
	int		length; /* bit #31 on if end of line */
	uint16		width, descent;
	int		vpos;
} tLineRun;

#define DEF_WDB_RTN(n)		int32 n(const textBoxObj *tbox, int32 i, boolean reverse)
#define TYP_WDB_RTN(n)		int32(*n)(const struct textBoxObj*, int32, boolean)

typedef struct textBoxObj
{
	rect16		bounds, txbounds;
	int32		length;
	int32		selStart, selEnd; /* may be in reverse order */
	int		scrollV, scrollH, tHeight;
	int		lineCount, styleCount;
	ufixed		maxWidth, selEndX; /* working range really: 0..IMAX(ifixed) */
	ifixed		tabAlign;
	rgb48		hiliteColor, custBgColor;
	uint16		flags;
	int32		caretTime, dragTime;
	char		*text;
	tLineRun	*lines;
	tStyleRun	*styles;
	TYP_WDB_RTN(wordBreak);

	struct windowRecord	*wind;
	struct ctlObj		*ctlScrollV, *ctlScrollH;
} textBoxObj;

/* ---- globals */

extern ffMemory	*fmFonts[256];
extern uint8	ffDefault;

/* ---- prototypes */

/*
 ffLoad
	Loads an entire font family.
	Returns NULL on error.
*/
ffMemory *ffLoad(struct ResFile_Memory *rMap,
	const uint8 *name,
	int ID);
/*
 ffMatch
	Finds or generates the closest bitmapped font.
*/
const bmfMemory *ffMatch(ffMemory *ff,
	int size,
	int8 styl);
/*
 ffFree
	Frees all memory allocated for a font family.
	All bmfs obtained from this family will be invalid.
*/
int ffFree(ffMemory *ff);

/*
 fmLoadMaster
	Loads the global table from a resource file.
*/
int fmLoadMaster(const char *name);
int fmLoadAll(struct ResFile_Memory *rmap);
void fmUnloadMaster(void);

/* settings */

int rTextFont(grafPort *p,
	uint16 font);
int rTextSize(grafPort *p,
	uint16 size);
int rTextStyle(grafPort *p,
	uint8 style);

/* measurement, rendering */

/*
 rTextWidth
	Calculates the integral width of a string.
*/
int rTextWidth(const bmfMemory *fm,
	const char *s,
	int length,
	int *italDeficit,
	ufixed x,
	ufixed spExtra,
	int8 styl);
/*
 rClipText
	Finds the maximum width-clipped length for a string.
*/
int rClipText(const bmfMemory *fm,
	const char *text,
	int length,
	ufixed *maxWidth,
	ufixed x,
	ifixed tabAlign,
	ifixed tabOrigin,
	ufixed chExtra,
	ufixed spExtra,
	int8 styl,
	boolean visible); /* include clipped char */
/*
 rTextBreak
	Finds the first non-space or line-ending break.
	The resulting maxWidth does not include breaking whitespace.
  Result:
	The next non-space character or next line; negative on error.
*/
int rTextBreak(const bmfMemory *fm, /* must be valid */
	const char *lineStart, /* can be NULL */
	const char *pos, /* must be valid */
	int *count, /* must be valid */
	ufixed *maxWidth, /* must be valid */
	ufixed w,
	ifixed tabAlign, /* origin is 0.0 */
	ufixed chExtra,
	ufixed spExtra,
	int8 styl);

/*
 rDrawText
	Renders text to a port, given the port's text settings.
	Offscreen imaging is used.
	Returns 0 if nothing drawn; next position if positve; IMIN(ifixed) on error.
*/
ifixed rDrawText(grafPort *p,
	const char*,
	int length,
	int y,
	ifixed x);

/* text boxes */

/*
 tBoxNew
	Creates a new textbox object, copying the string ptr and length.
*/
textBoxObj *tBoxNew(struct windowRecord*,
	const rect16 *bounds,
	const char*,
	int32 length,
	uint16 flags);
int tBoxDispose(textBoxObj*);

/*
 tBoxRecalc
	Calculates the line-runs in a textbox.
	An optional starting line can speed up insertion.
  Result:
	1 on success; zero if empty.
*/
int tBoxRecalc(textBoxObj*,
	int startLine); /* zero-based, pinned to existing range */
/*
 tBoxDraw
	Renders the textbox, erasing the back first (using bkColor).
	It will use tBoxRecalc if necessary.
  Result:
	1 on success; zero if empty.
*/
int tBoxDraw(textBoxObj*);
/*
 tBoxUpdate
	Draws and updates the visible portion of the native window.
	Optionally, the scroll ranges/positions can be updated.
*/
int tBoxUpdate(textBoxObj *tbox,
	const rect16 *area, /* can be NULL */
	int firstLine); /* pass negative for no recalc; IMAX for scrolls only */

/*
 tBoxScroll
	Adjusts the textbox scrolls, and updates if changed.
	The new values are pinned to the practical range (tBoxMaxScroll).
*/
int tBoxScroll(textBoxObj*,
	int amtv,
	int amth);
int tBoxMaxScroll(textBoxObj*,
	int32 *maxv,
	int32 *maxh);
/*
 tBoxSelect
	Sets the textbox selection, and updates if changed.
	The new values are pinned to the functional range.
*/
int tBoxSelect(textBoxObj*,
	int start,
	int end);
/*
 tBoxClick
	Handles cursor selection, dragging.
*/
int tBoxClick(textBoxObj*,
	pt2d16,
	uint16 eventModifiers,
	boolean dblClick);
/*
 tBoxIdle
	If active and writable, the caret is updated.
*/
int tBoxIdle(textBoxObj*);
/*
 tBoxKey
	If active and writable, handles a key function (may insert a char).
*/
int tBoxKey(textBoxObj*,
	int key,
	uint16 eventModifiers);
/*
 tBoxDelete
	If active and writable, removes text and styles within a range.
	The range can run in reverse (start after end).
*/
int tBoxDelete(textBoxObj*,
	int32 start,
	int32 end,
	boolean update);
int tBoxInsert(textBoxObj*,
	int32 start,
	const char *s,
	int32 length,
	const tStyleRun *style,
	boolean update);
/*
 tBoxPinView
	Adjusts the scrolls so selEnd is in view.
  Note:
	This does not render, but returns a positive mask when drawing is called for.
*/
int tBoxPinView(textBoxObj*);
/*
 tBoxSetMargin
	Sets the text boundary based on the outer bounds and passed margins.
	These values can be inverted to draw outside the erased bounds.
  Result:
	1 on success; 0 if the new text bounds are empty.
*/
int tBoxSetMargins(textBoxObj *tbox,
	const rect16 *margins); /* can be NULL */
/*
 tVisLine
	Finds the first visible line-run in a list.
  Notes:
	NULL will be returned if 'y' is past the visible range.
*/
const tLineRun *tVisLine(const tLineRun *s,
	int lineCount,
	int y);
/*
 tBoxCharLine
	Calculates the measured caret line for a character.
*/
const tLineRun *tBoxCharLine(textBoxObj*,
	int32 offs,
	ufixed *x, /* unscrolled box coordinate; can be NULL */
	int *top, /* can be NULL */
	int *bottom); /* can be NULL */
/*
 tBoxPtToChar
	Finds the character for an unscrolled box coordinate.
	This routine rounds to the nearest character edge.
*/
int32 tBoxPtToChar(textBoxObj*,
	ufixed x,
	int y);
/*
 tboxSelRgn
	Generates a region for the selection, clipped vertically.
	Returns NULL if there is no applicable or visible selection beyond a caret.
*/
SLRgn *tBoxSelRgn(textBoxObj*);
/*
 tBoxCharStyle
	Returns the style for a character.
	If there are no styles, the default will be generated.
*/
const tStyleRun *tBoxCharStyle(textBoxObj*,
	int32 i,
	int *remCount); /* remaining run count; can be NULL */
#endif
