/*
 mwind.h
  virtual window/dialog and control structures

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __MWIND__
 #define __MWIND__
 #include "mgraf.h"

/* ---- constants */

#define dialogKind		2
#define userKind		8

enum
{ /* window/control colors */
	wCFrame,
	wCTitle,
	wCContent
};

enum
{			/* data input	result */
	wcMsgDraw,	/* none		visibility */
	wcMsgType,	/* none		type index, below */
	wcMsgTest,	/* pt2d16*	label part for coord */
	wcMsgWTest,	/* pt2d16*	label part for wide-ranging coord */
	wcMsgClick,	/* pt2d16*	successful label part */
	wcMsgDblClick,	/* pt2d16*	successful label part */
	wcMsgDragMeas,	/* pt2d16*	new value */
	wcMsgMarkMeas,	/* value	marker/thumb pos. (pt2d16) */

	wcMsgDispose,	/* none */
	wcMsgCalc
};

enum
{
	winTypeDocument,
	winTypeModal,
	winTypePlainDialog,
	winTypeShadow,
	winTypeNoGrowDoc,
	winTypeMovableModal
};

enum
{
	ctlTypePushBtn,
	ctlTypeCheckBox,
	ctlTypeRadioBtn,
	ctlTypeScrollBar,
	ctlTypePopUpMenu,
	ctlTypeTab,
	ctlTypeStaticText,
	ctlTypeEditText,
	ctlTypePicture,
	ctlTypeIcon
};

#ifndef __QUICKDRAW__
enum
{
	kControlLabelPart		= 1,
	kControlEditTextPart		= 5,
	kControlUpButtonPart		= 20,
	kControlDownButtonPart,
	kPageUpButtonPart,
	kPageDownButtonPart
};

enum
{
	kControlCheckBoxCheckedValue	= 1,
	kControlCheckBoxMixedValue
};
#endif

/* ---- typedefs */

#define DEF_WIN_RTN(n)		int n(windowRef win, int msg, const void *data)
#define TYP_WIN_RTN(n)		int(*n)(struct windowRecord*, int, const void*)

typedef struct windowRecord
{
	grafPort	port;
	rect16		bounds, frame, clipSave, updateRect;
	int8		kind;
	uint8		visible; /* on the screen, takes clicks; 2=translucent */
	boolean		hilited;
	boolean		goAway;	/* has a closebox */
	long int	refCon;
	uint8		title[256]; /* Pascal string */

	struct windowRecord	*above, *below;

	struct ctlObj	*ctlList, *ctlFocus, *ctlDefault, *ctlCancel;

	TYP_WIN_RTN(rtn);
	void		*data;
} windowRecord, *windowRef;

typedef TYP_WIN_RTN(winRoutine);

/* controls */

#define DEF_CTL_RTN(n)		int32 n(ctlObj *ctl, int msg, const void *data)
#define TYP_CTL_RTN(n)		int32(*n)(struct ctlObj*, int, const void*)
#define DEF_TRK_RTN(n)		void n(ctlObj *ctl, int part, int32 value)
#define TYP_TRK_RTN(n)		void(*n)(struct ctlObj*, int, int32)

typedef TYP_CTL_RTN(ctlRoutine);

typedef struct ctlObj
{
	struct ctlObj	*next; /* in the window chain */
	windowRef	wind;
	uint8		title[256]; /* Pascal string */
	rect16		bounds;
	int32		min, max, value;
	long int	refCon;
	uint8		hilite, visible, txStyle;

	uint16		txFont, txSize;

	TYP_CTL_RTN(rtn);
	TYP_TRK_RTN(trk);
	void		*data;
} ctlObj;

/* ---- globals */

extern windowRecord	screen_desktop_window;

/* ---- prototypes/mwind.c */

int wShowWindow(windowRef wind,
	boolean visible);
int wSizeWindow(windowRef wind,
	int16 width,
	int16 height,
	boolean update);
windowRef wFrontWindow(void);
void wOpenWindow(windowRef newWin,
	windowRef behind,
	int16 procID);
void wCloseWindow(windowRef deadWin);
void wDisposeWindow(windowRef deadWin);
int wPaintOne(windowRef win,
	const rect16 *inval);
int wPaintBehind(windowRef win,
	const rect16 *inval);
void wSendBehind(windowRef win,
	windowRef behind);
void wBeginUpdate(windowRef win);
void wEndUpdate(windowRef win);
void wInvalRect(windowRef win,
	const rect16 *dirtyRect);

/* ---- prototypes/mctl.c */

ctlObj *wNewControl(windowRef wind,
	const rect16 *bounds,
	const uint8 *title, /* can now be NULL (empty) */
	boolean visible,
	int32 value,
	int32 min,
	int32 max,
	int type, /* see enum above */
	long int refCon);
int wDisposeControl(ctlObj*);
int wKillControls(windowRef wind);
int wDrawControls(windowRef wind);
int wDraw1Control(ctlObj *ctl);
int wIdleControls(windowRef wind);
ctlObj *wControlEvent(windowRef wind,
	const struct eventRecord*);

/* returns label part */
int wFindControl(const windowRef wind,
	pt2d16 loc,
	ctlObj **out);
int32 wCountControls(windowRef wind);
int32 wGetControlIndex(const windowRef wind,
	const ctlObj *ctl);

int wSetControlBounds(ctlObj *ctl,
	const rect16*);
int wSetControlValue(ctlObj *ctl,
	int32 value);
int wSetControlMinimum(ctlObj *ctl,
	int32 value);
int wSetControlMaximum(ctlObj *ctl,
	int32 value);
int wSetControlTitle(ctlObj *ctl,
	const uint8*);
int wHiliteControl(ctlObj *ctl,
	uint8 hilite);

#ifdef __MWIND_PRIVATE
int _imgContColors(cTab64 *ct,
	const rgb48 *cont,
	const rgb48 *back,
	boolean disabled);
void _defMWinModalDraw(const rect16 *box,
	const rgb48 *color,
	boolean fill);
extern const rgb48 _defMWinColors[3];
#endif

 #if !TARGET_OS_MAC && !defined(__QUICKDRAW__) /* QT 3.0+; you never know */
  typedef ctlObj *ControlRef;
  typedef windowRef WindowRef, DialogRef;
  typedef windowRecord DialogRecord;
  typedef grafPort *GrafPtr, *CGrafPtr;
  typedef mhandle DialogTHndl;

  void MacSetPort(GrafPtr);
  void GetPort(GrafPtr*);

  #define GetWindowPort(w) (&(w)->port)
  #define GetWindowFromPort(p) ((struct windowRecord*)(p))
  #define GetDialogWindow(p) (p)
  #define GetWindowDialog(p) (p)
  #define GetWRefCon(p) (0+(p)->refCon)
  #define SetWRefCon(p,v) (void)((p)->refCon=v)

  #define FrontWindow wFrontWindow
  #define DisposeWindow wDisposeWindow
  #define MacShowWindow(w) (void)wShowWindow(w,1)
  #define HideWindow(w) (void)wShowWindow(w,0)
  #define SendBehind(w) wSendBehind(w,NULL)
  #define BringToFront(w) wSendBehind(w,(windowRef)-1l)
  #define BeginUpdate wBeginUpdate
  #define EndUpdate wEndUpdate
  #define InvalWindowRect wInvalRect

  #define DisposeControl (void)wDisposeControl
  #define KillControls (void)wKillControls
  #define DrawControls (void)wDrawControls
  #define Draw1Control (void)wDraw1Control
  #define IdleControls (void)wIdleControls
  #define FindControl(l,w,c) wFindControl(w,l,c)
  #define SetControlValue (void)wSetControlValue
  #define SetControlMinimum (void)wSetControlMinimum
  #define SetControlMaximum (void)wSetControlMaximum
  #define SetControlTitle (void)wSetControlTitle
  #define HiliteControl (void)wHiliteControl
  #define GetControlValue(c) (0+(c)->value)
  #define GetControlMinimum(c) (0+(c)->min)
  #define GetControlMaximum(c) (0+(c)->max)
  #define GetControlHilite(c) (0+(c)->hilite)
  #define GetControlTitle(c,s) (void)memcpy(s,(c)->title,256)

  void update_screen_window_rect(windowRef win,
	int left,
	int top,
	int right,
	int bottom);
 #endif
#endif
