/*
 pict.c
  this was written to stop all the crashing in Classic; QD can't handle the errors
  initially based on code in AlephOne/RenderOther/images.cpp
  only image opcodes are handled

  Created: 2012.12.29
  Revised: 2014.08.22

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2013.01.01 08:12 EST
  MAX instead of MIN for w/h so all of the parchment image in ViV is rendered.
Using the slow way to finish a clipped line fixed the missing final span bug.
2013.01.04 03:14 EST
  after implementing all of the rest of the opcodes found in PictUtils.r
(not rendered, however), made the image parser its own function for PixPat codes.
  forced to use GWorld instead of raw PixMap because QT always draws to port
(the dst parameter is counter-intuitive if it can't be used alone).
2013.07.16 07:01 EDT
  final fixes, etc. for post-QT image rendering (needed for multiple QT images).
*/

#include <string.h>
#include <stdlib.h>

#ifdef TARGET_OS_MAC /* need the SDK to implement QT imaging on other platforms */
 #include "my_cseries.h"
 #define lockPixels(p) (**p).baseAddr /* alloc_image buffer w/o GWorld always a calloc */
 #define pmPitch(p) ((**(p)).rowBytes&0x3FFF)
 #define QT_IMAGE
#else
 #include "mgraf.h"
 #define DisposePixMap disposePixMap
 #define pmPitch(p) (0+(p)->pitch)
#endif

#ifdef QT_IMAGE
 #include <QuickDraw.h>
#endif

/* ---- code */

static int unpack_rle8(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	const uint8	*s = src;
	boolean		word_size = row_bytes>250;

	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z;
		const uint8	*se;
		uint8		v, *d = dst, *de = dst_pitch+d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = *s++;
				if (d+c>de)
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = v;
					while(--c);
					c = 0;
					break;
				}
				do
					*d++ = v;
				while(--c);
				c = 0;
			} else
			{ /* uncompressed run */
				++c;
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = *s++;
						else
							++s;
					while(--c);
					c = 0;
					break;
				}
				do
					*d++ = *s++;
				while(--c);
			}
		}
		s = se;
	} while(dst+=dst_pitch,--rows);

	return s-src;
}

static int unpack_rle16(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	const uint8	*s = src;
	boolean		word_size = row_bytes>250;

	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z;
		const uint8	*se;
		uint16		v, *d = (uint16*)dst, *de = d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z, ((uint8*)de)+= dst_pitch;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = va_be16(s);
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = v;
					while(--c);
					break;
				}
				do
					*d++ = v;
				while(--c);
				c = 0;
			} else
			{ /* uncompressed run */
				++c;
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = *s<<8|s[1];
					while(s+=2,--c);
					c = 0;
					break;
				}
				do
					*d++ = va_be16(s);
				while(--c);
			}
		}
		s = se;
	} while(dst+=dst_pitch,--rows);

	return s-src;
}

static int unpack_rle32(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows,
	int cmp_count)
{
	const uint8	*s = src;
	boolean		word_size = row_bytes>250;

//	assert(cmp_count>1 && cmp_count<5);
	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z, cc = cmp_count;
		const uint8	*se, *re;
		uint8		v, *d = dst, *de = dst_pitch+d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z, re = row_bytes+d, d+= BIG_ENDIAN ? 1:2;//4-cc:cc-1;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = *s++;
				do
				{
					if (d<de)
						*d = v;
					if ((d+=4)>=re)
					{
						if (cc<=1)
							{s = se; continue;}
						--cc, d-= row_bytes, d+= BIG_ENDIAN ? 1:-1;
					}
				} while(--c);
				c = 0;
				continue;
			}
			++c; /* uncompressed run */
			do
			{
				if (d<de)
					*d = *s;
				++s;
				if ((d+=4)>=re)
				{
					if (cc<=1)
						{s = se; continue;}
					--cc, d-= row_bytes, d+= BIG_ENDIAN ? 1:-1;
				}
			} while(--c);
		}
		s = se;
	} while(dst+=dst_pitch,--rows);
	return s-src;
}

static int unpack_pixels(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	uint16 depth,
	uint16 pack_type,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	int		low_pitch = MIN(row_bytes,dst_pitch), data_size = 0;

	if (low_pitch<0 || rows<=0) /* following calls expect valid values */
		return 0;
	if (row_bytes<8 || pack_type==1)
	{ /* uncompressed */
		const uint8	*p = src;
		uint8		*q;

		data_size = row_bytes*rows;
		if ((q=dst)==NULL || data_size<=0)
			goto end;
		do /* SDL AO doesn't byteswap in rb<8 case */
#if !BIG_ENDIAN
			if (depth==16)
			{
				uint16		*d = (uint16*)q;
				const uint8	*s = p;
				int		c;

				if ((c=row_bytes>>1)>0)
					do
						*d++ = va_be16(s);
					while(--c);
				c = 0;
			} else if (depth==32)
			{
				uint32		*d = (uint32*)q;
				const uint8	*s = p;
				int		c;

				if ((c=row_bytes>>2)>0)
					do
						*d++ = va_be32(s);
					while(--c);
				c = 0;
			} else
#endif
			memcpy(q,p,low_pitch);
		while(p+=row_bytes,q+=dst_pitch,--rows);
		rows = 0;
		goto end;
	}
	/* compressed */
	if (depth<=8) /* indexed */
		data_size = unpack_rle8(src,end,dst,row_bytes,dst_pitch,rows);
	else
	{ /* direct */
		if (!pack_type)
			if (depth<=16)
				pack_type = 3;
			else if (depth==24 || depth==32)
				pack_type = 4;
		if (pack_type==3) /* 15bpp */
			data_size = unpack_rle16(src,end,dst,row_bytes,dst_pitch,rows);
		else if (pack_type==4) /* 24/32bpp */
			data_size = unpack_rle32(src,end,dst,row_bytes,dst_pitch,rows,depth>>3);
		else
		{
			dprintf("unpack_pixels: bad pack type %d, depth %d;g",
				pack_type,depth);
			return -1;
		}
	}
	if (data_size<0)
		dprintf("unpack_pixels: data position exceeded end;g");
end:	return data_size;
}

static const uint8 *parse_ctab(const uint8 *src,
	platform_pixmap m)
{
	int		c, z;
#if TARGET_OS_MAC
	CTabHandle	cth = (**m).pmTable;
	CTabPtr		ct;

	c = (int16)(src[6]<<8|src[7]);
	if (cth)
	{
		z = GetHandleSize((Handle)cth)-8>>3;
		ct = *cth;
		ct->ctSize = c;
		ct->ctSeed = GetCTSeed();
	} else
		ct = NULL, z =-1;
	return parseColorsBE(ct ? (cSpec*)ct->ctTable:NULL,src+8,c+1,z,src[4]&0x80);
#else
	cTab64		*ct = m ? m->pTable:NULL;

	c = (int16)(src[6]<<8|src[7])+1;
	if (ct)
	{
		ct->flags|= 0x8000, ct->count = c, z = 1u<<m->bpp; /* expects pTable to be large enough! */
		ct->seed = GetCTSeed();
	} else
		z =-1;
	return parseColorsBE(ct ? ct->colors:NULL,src+8,c,z,src[4]&0x80);
#endif
}

/* returns new src ptr */
static const uint8 *parse_bits(const uint8 *src,
	const uint8 *end,
	platform_pixmap *image,
	int *outDepth,
	rect16 *srcRect,
	uint8 hasClip,
	boolean hasCTab)
{
	long		data_size;
	int		row_bytes, depth, width, height;
	uint16		pack_type;

	row_bytes = va_be16(src); /* data rowBytes, bounds */
	srcRect->top = *src<<8|src[1];
	srcRect->left = src[2]<<8|src[3];
	srcRect->bottom = src[4]<<8|src[5];
	srcRect->right = src[6]<<8|src[7], src+= 8;
	width = RECTANGLE_WIDTH(srcRect);
	height = RECTANGLE_HEIGHT(srcRect);
	if (row_bytes&0x8000)
	{
		src+= 2;	// pmVersion - irrelevant (stored in PICT)
		pack_type = va_be16(src);
		src+= 4+4+4+2;	// dataSize, hRes, vRes, pixelType (chunky,chunkyPlanar,planar,0x10=RGBDirect)
		depth = va_be16(src);
		if (depth!=1 && depth!=2 && depth!=4 && depth!=8 && depth!=16 && depth!=32)
		{
			dprintf("parse_bits: bad pixel size (%d);g",
				depth);
			return NULL;
		}
		//if (depth==32 && (*src<<8|src[1])==3)
		//	depth = 24; /* 4 bytes/pixel, but no alpha channel */
		src+= 2+2+4+4+4; // cmpCount, cmpSize, planeBytes, pmTable, pmReserved
	} else
		pack_type = 0, depth = 1;
	if (image && *outDepth!=depth)
	{
		if (*image)
			{DisposePixMap(*image); *image = NULL;} /* sorry to have to do this */
		if ((*image=alloc_image(depth,width,height,NULL))==NULL)
			return NULL;
		*outDepth = depth;
	}
	if (!(row_bytes&0x8000))
		hasCTab = 0;
	row_bytes&= 0x3FFF;
	if (hasCTab)
		src = parse_ctab(src,!image ? NULL:*image);
	/* would-be actual bounds */
	srcRect->top = *src<<8|src[1];
	srcRect->left = src[2]<<8|src[3];
	srcRect->bottom = src[4]<<8|src[5];
	srcRect->right = src[6]<<8|src[7], src+= 8;
	src+= 8+2; /* destination Rect, transfer mode */

	if (hasClip)
		src+= *src<<8|src[1]; /* skip clip region */
	//dprintf(" %spackBits(%d,%d,%d,%d%s), rowb=%d, depth=%d/%d%s;g",
	//	hasCTab ? "color ":"",srcRect->top,srcRect->left,srcRect->bottom,srcRect->right,
	//	hasClip ? " (clipped)":"",row_bytes,depth,pack_type,image ? "":" - not rendering");
	if (src>=end)
		return NULL;

	if ((data_size=unpack_pixels(src,end,!image ? NULL:(uint8*)lockPixels(*image),
			depth,pack_type,row_bytes,!image ? 0:pmPitch(*image),height))<0)
	{
	//	if (*image)
	//		{dispose_screen_image(*image); *image = NULL;}
		return NULL;
	}
	if (image && srcRect->top|srcRect->bottom && srcRect->left|srcRect->right)
	{
	#if TARGET_OS_MAC
		(***image).bounds = *(const Rect*)srcRect;
	#else
		platform_pixmap	m = *image;

		width = srcRect->right, height = srcRect->bottom;
		width-= m->pOrigin.x=srcRect->left;
		height-= m->pOrigin.y = srcRect->top;
		if (width>m->w)
			dprintf("parse_bits: source image suggests larger source width;g");
		else
			m->w = width;
		if (height>m->h)
			dprintf("parse_bits: source image suggests larger source height;g");
		else
			m->h = height;
	#endif
	}
	return src+data_size; /* b&w pics are off */
}

#ifdef QT_IMAGE
# include <ImageCompression.h>

static ImageDescriptionHandle make_image_description(uint32 data_size,
	uint32 codec,
//	short framecount,
	const uint8 *name, /* Pascal */
	short depth,
	short width,
	short height)
{
	ImageDescriptionHandle	h;
	ImageDescription	*d;

	if ((h=(ImageDescriptionHandle)NewHandleClear(sizeof(ImageDescription)))==NULL)
		return NULL;
	d = *h;
	d->idSize = sizeof(*d);
	d->cType = codec;
	d->width = width;
	d->height = height;
	d->dataSize = data_size;
	d->frameCount = 0;
	d->depth = depth;
	memcpy(d->name,name,32);

	return h;
}
#endif

platform_pixmap renderQDPict(const uint8 *start,
	uint32 size)
{
	rect16		src_rect, dst_rect;
	rect16		bounds;
	platform_pixmap	image = NULL;
	const uint8	*src;
	int		pic_width, pic_height, data_size, depth =-1;
	uint16		opcode = 0, type;
	uint8		vers2, renderedQT, imgHasQT;

	if ((src=start)==NULL || size<12)
		return NULL;
	src+= 2; /* low 16 bits of size */
	bounds.top = *src<<8|src[1];
	bounds.left = src[2]<<8|src[3];
	bounds.bottom = src[4]<<8|src[5];
	bounds.right = src[6]<<8|src[7], src+= 8;
	pic_width = RECTANGLE_WIDTH(&bounds);
	pic_height = RECTANGLE_HEIGHT(&bounds);

	/* the version 2 sequence is supposed to be 0x1102 (and it works), but Apple eventually went with 0x0011 so
	  an old type-1 parser would take it as 0x00,0x11 = NOP,Version, followed by a high byte outside [1,2]
	*/
	vers2 = renderedQT = imgHasQT = 0;
	while(src-start<size)
	{ /* parse the opcodes */
		if (vers2)
		{
			src = start+((src-start+1)&~1); /* word align */
			opcode = va_be16(src);
		} else
			opcode = *src++;
		//dprintf("%04x;g",
		//	opcode);
		switch(opcode)
		{
		case 0x011:	// VersionOp
			vers2 = *src++;
			if ((uint8)(vers2-1)>1u)
			{
				dprintf("renderQDPict: bad parsed version (%u);g",
					vers2);
				goto end;
			}
			vers2&= 2;
			break;
		case 0:		// NOP
		case 0x017:	// reserved
		case 0x018:
		case 0x019:
		case 0x01C:	// HiliteMode
		case 0x01E:	// DefHilite
		case 0x038:	// FrameSameRect
		case 0x039:	// PaintSameRect
		case 0x03A:	// EraseSameRect
		case 0x03B:	// InvertSameRect
		case 0x03C:	// FillSameRect
		case 0x03D:	// reserved
		case 0x03E:
		case 0x03F:
		case 0x048:	// FrameSameRRect
		case 0x049:	// PaintSameRRect
		case 0x04A:	// EraseSameRRect
		case 0x04B:	// InvertSameRRect
		case 0x04C:	// FillSameRRect
		case 0x04D:	// reserved
		case 0x04E:
		case 0x04F:
		case 0x058:	// FrameSameROval
		case 0x059:	// PaintSameROval
		case 0x05A:	// EraseSameROval
		case 0x05B:	// InvertSameROval
		case 0x05C:	// FillSameROval
		case 0x05D:	// reserved
		case 0x05E:
		case 0x05F:
		case 0x078:	// FrameSamePoly
		case 0x079:	// PaintSamePoly
		case 0x07A:	// EraseSamePoly
		case 0x07B:	// InvertSamePoly
		case 0x07C:	// FillSamePoly
		case 0x07D:	// reserved
		case 0x07E:
		case 0x07F:
		case 0x088:	// FrameSameRgn
		case 0x089:	// PaintSameRgn
		case 0x08A:	// EraseSameRgn
		case 0x08B:	// InvertSameRgn
		case 0x08C:	// FillSameRgn
		case 0x08D:	// reserved
		case 0x08E:
		case 0x08F:
			break;
		case 1:		// ClipRgn
		case 0x070:	// FramePoly
		case 0x071:	// PaintPoly
		case 0x072:	// ErasePoly
		case 0x073:	// InvertPoly
		case 0x074:	// FillPoly
		case 0x075:	// reserved
		case 0x076:
		case 0x077:
		case 0x080:	// FrameRgn
		case 0x081:	// PaintRgn
		case 0x082:	// EraseRgn
		case 0x083:	// InvertRgn
		case 0x084:	// FillRgn
		case 0x085:	// reserved rgn
		case 0x086:
		case 0x087:
			src+= *src<<8|src[1];
			break;
		case 3:		// TxFont
		case 4:		// TxFace
		case 5:		// TxMode
		case 8:		// PnMode
		case 13:	// TxSize
		case 0x015:	// PnLocHFrac - uint16
		case 0x016:	// ChExtra - int16 char spacing
		case 0x023:	// ShortLineFrom - byte dh,dv
		case 0x0A0:	// ShortComment
			src+= 2;
			break;
		case 6:		// SpExtra
		case 7:		// PnSize
		case 11:	// OvSize - Point
		case 12:	// Origin - Point
		case 14:	// FgColor
		case 15:	// BgColor
		case 0x021:	// LineFrom - Point
		case 0x068:	// FrameSameArc - angle start/end
		case 0x069:	// PaintSameArc
		case 0x06A:	// EraseSameArc
		case 0x06B:	// InvertSameArc
		case 0x06C:	// FillSameArc
		case 0x06D:	// reserved
		case 0x06E:
		case 0x06F:
			src+= 4;
			break;
		case 0x01A:	// RGBFgCol
		case 0x01B:	// RGBBkCol
		case 0x01D:	// HiliteColor
		case 0x01F:	// OpColor
			if (!vers2)
				dprintf("renderQDPict: RGB color op in version #1 PICT;g");
		case 0x022:	// ShortLine - Point/byte dh,dv
			src+= 6;
			break;
		case 2:		// BkPat
		case 9:		// PnPat
		case 10:	// FillPat
		case 0x010:	// TxRatio
		case 0x020:	// Line
		case 0x030:	// FrameRect
		case 0x031:	// PaintRect
		case 0x032:	// EraseRect
		case 0x033:	// InvertRect
		case 0x034:	// FillRect
		case 0x035:	// reserved
		case 0x036:
		case 0x037:
		case 0x040:	// FrameRRect
		case 0x041:	// PaintRRect
		case 0x042:	// EraseRRect
		case 0x043:	// InvertRRect
		case 0x044:	// FillRRect
		case 0x045:	// reserved
		case 0x046:
		case 0x047:
		case 0x050:	// FrameOval
		case 0x051:	// PaintOval
		case 0x052:	// EraseOval
		case 0x053:	// InvertOval
		case 0x054:	// FillOval
		case 0x055:	// reserved
		case 0x056:
		case 0x057:
			src+= 8;
			break;
		case 0x028:	// LongText
			src+= 4; // Point
			src+= *src++; // pstring
			break;
		case 0x02B:	// DHVText
			++src;
		case 0x029:	// DHText
		case 0x02A:	// DVText
			++src; // delta
			src+= *src++; // pstring
			break;
		case 0x0A1:	// LongComment
			src+= src[2]<<8|src[3], src+= 4;
			break;
		case 0x012:	// BkPixPat
		case 0x013:	// PnPixPat
		case 0x014:	// FillPixPat
			type = va_be16(src);
			if (type==1)
			{ // newPattern
				src+= 8; // Pattern
				if ((src=parse_bits(src,start+size,NULL,&depth,&src_rect,0,1))==NULL)
					goto end;
			} else if (type==2)
			{ // ditherPattern
				src+= 8+6; // Pattern/RGBColor
			} else
			{
				dprintf("renderQDPict: bad pattern type (%u);g",
					type);
				goto end;
			}
			break;
		case 0x02D:	// LineJustify
			type = va_be16(src);
			if (type!=8)
			{
				dprintf("renderQDPict: bad line-justify code (%u);g",
					type);
				goto end;
			}
			src+= 4+4; // spacing/extra just-spacing
			break;
		case 0x090:	// BitsRect
		case 0x091:	// BitsRgn
		case 0x098:	// PackBitsRect
		case 0x099:	// PackBitsRgn
		case 0x09A:	// DirectBitsRect
		case 0x09B:	// DirectBitsRgn
			if ((unsigned)opcode-0x09A<=1) /* whole PixMap header (includes baseAddr) */
				src+= 4;
			if ((src=parse_bits(src,start+size,renderedQT ? NULL:&image,
					&depth,&src_rect,opcode&1,opcode<0x09A))==NULL)
				goto end;
			break;
		case 0x02C:	// FontName
		case 0x02E:	// reserved
		case 0x02F:
		case 0x092:	// reserved
		case 0x093:
		case 0x094:
		case 0x095:
		case 0x096:
		case 0x097:
		case 0x09C:	// reserved
		case 0x09D:
		case 0x09E:
		case 0x09F:
			src+= va_be16(src);
			break;
		case 0x0FF:	// OpEndPic
			goto end;
		//case 0xC00:	// HeaderOp - 0xC*2, as calc'd below, is 24
		//	src+= 24; // int16 -1 = std version; -2 = extended (has h/vRes, srcRect)
		//	break;

		default:
			if (opcode>=0x300 && opcode<0x8000)
				src+= (opcode>>8)*2;
			else if (opcode==0x8200)
			{ /* QuickTime */
#ifdef QT_IMAGE
			extern long quicktime_version; /* from myMediaPlayer.c */

				ImageDescriptionHandle		desc = NULL;
				GDHandle	olddevice;
				CGrafPtr	oldport, imageport;
				int		error;
#endif
				uint32		opcode_size, opcode_start, matte_size, mask_size;
				uint32		id_start, id_size, codec;
				int		iwid, ihgt;

				opcode_size = va_be32(src);
				opcode_start = src-start;
				src+= 26;	// version/matrix - taking contained offset
				dst_rect.left = *src<<8|src[1], src+= 2+2;
				dst_rect.top = va_be16(src);
				src+= 6;
				matte_size = va_be32(src);
				src+= 10;	// matteRec/mode
				src_rect.top = *src<<8|src[1];
				src_rect.left = src[2]<<8|src[3];
				src_rect.bottom = src[4]<<8|src[5];
				src_rect.right = src[6]<<8|src[7], src+= 8;
				src+= 4;	// accuracy
				mask_size = va_be32(src);

				if (matte_size)
					data_size = va_be32(src)-4+matte_size, src+= data_size;
				src+= mask_size;

				id_start = src-start;
				id_size = va_be32(src);
				codec = *(const int32*)src;
				src+= 4+4+2+2+2+2+4+4+4; // resvd1/2/dataRefIx/ver/rev/vendor/temporalQual/spatialQual
				iwid = va_be16(src);
				ihgt = va_be16(src);
				src+= 8; //hRes/vRes
				data_size = va_be32(src);
				depth = src[34]<<8|src[35];

				if (!imgHasQT)
				{
					dprintf("QT PICT;g");
					imgHasQT = 1;
				}
				dprintf(" img @ %d,%d: width %d/%d, height %d/%d, depth %u, codec '%-4.4s'-%.*s;g",
					dst_rect.left,dst_rect.top,iwid,pic_width,ihgt,pic_height,depth,
					&codec,MIN(src[2],31),src+3);
				/* skip if the QT version is inadequate */
				/* some PICTs have a message appended in case of this */
#ifdef QT_IMAGE
				if (quicktime_version<0x2008000)
					goto qtskip;
				/* to fix Marathon Red JPEG chapter screens; 2013.07.16 */
				//if (image)
				//	dispose_screen_image(image);
				if (!image && (image=alloc_image(32,pic_width,pic_height,&imageport))==NULL
						|| (desc=make_image_description(data_size,codec,
						src+2,depth,pic_width,pic_height))==NULL)
				{
					error = MemError();
					goto qterr;
				}
				/* the QT call */
				dst_rect.bottom = dst_rect.top+ihgt;//pic_height;
				dst_rect.right = dst_rect.left+iwid;//pic_width;
				if (id_start+id_size>size)
				{
					error = eofErr;
					goto qterr;
				}
				GetGWorld(&oldport,&olddevice);
				SetGWorld(imageport,NULL);
				error = DecompressImage((Ptr)start+id_start+id_size, /* shouldn't modify it */
						desc,
						image,
						(const Rect*)&src_rect,
						(const Rect*)&dst_rect,
						ditherCopy,
						NULL);
				SetGWorld(oldport,olddevice);
				DisposeHandle((Handle)desc);
				image = NewPixMap();
				CopyPixMap(GetGWorldPixMap(imageport),image);
				DisposeGWorld(imageport);
				if (error)
				{
				qterr:	dprintf("renderQDPict: error #%d for QT image;g",
						error);
					if (error==-2003)
						dprintf(" can't find handler;g");
					else if (error==-8969)
						dprintf(" bad format;g");
				}
				renderedQT = TRUE; /* override "need to install QT" message */
				//return image;
			qtskip:
#endif
				src = start+opcode_start+opcode_size;
			} else if (opcode<0x8000 || opcode>0x8100)
			{
				dprintf("renderQDPict: unimplemented opcode 0x%04X in PICT;g",
					opcode);
				goto end;
			}
		}
	}
end:	if (opcode!=0xFF)
		dprintf("renderQDPict: abrupt termination - opcode %04X;g",
			opcode);
#if 0
#ifndef QT_IMAGE
	if (imgHasQT)
		return image; /* wide bounds invalid */
#endif
	if (image)
	#if TARGET_OS_MAC
		(**image).bounds = *(const Rect*)&bounds;
	#else
		image->h = RECTANGLE_HEIGHT(&bounds), image->w = RECTANGLE_WIDTH(&bounds);
	#endif
#endif
	return image;
}
