/*
 adpcm.c
  common ADPCM encodings

  Created: 2014.01.08 09:18 EST
  Revised: 2014.04.25

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.01.29
  Made cxd2_default const.
2014.04.25
  Fixed WAVE-based IMA ADPCM.
*/

#include <enviro64.h>

#include "snddec.h"

/* ---- macros */

#define addclamp(v,u) do{ v+=u; \
	if(v!=(int16)v) v=v>>31^0x7FFF; \
}while(0)

/* ---- BRR */

/* Yellow book values */
/* for SNES, 8 dual-sample bytes, after: [upshft.4] [filter.2] [loop.1] [end.1] */
/* for EASC CD/XA, the last coefficient pair (fifth) isn't used (or is it?) */
const adpcm_base	brr_default =
{
	{0,  0,  0, 60,-52,115,-55, 98,-60,122},
	0xE, 14 /* PSX (PlayStation) coef. pair index mask, count */
};

/* This originally routine adapted from 'mame.zip/src/emu/sound/spu.c'
  (MAME adaptation by R. Belmont.)
*/
uint8 *brr_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	int16		f0, f1, l0, l1;
	int		h;

	if (!s || !d || !f || z<=0)
		return d;
	{
		const int8	*filter;

		if (f->mx>8) /* Sony */
			filter = f->coef+(*s>>3&f->kx), h = *s&15;
		else /* SPC */
			filter = f->coef+(*s>>1&f->kx), h = 12-(*s>>4);
		f0 = filter[1], f1 = *filter;
	}
	l0 = f->l0, l1 = f->l1, m*= sizeof(int16);
	goto start;
	do
	{
		uint32		b;
		int		v;

		d+= m;
	start:	b = *++s<<8;
		v = (int16)(b<<4&0xF000)>>h;
		addclamp(v,(l0*f0+l1*f1+32)>>6);
		*(int16*)d = l1 = v;
		v = (int16)(b&0xF000)>>h;
		addclamp(v,(l1*f0+l0*f1+32)>>6);
		*(int16*)(d+=m) = l0 = v;
	} while(--z);
	f->l0 = l0, f->l1 = l1;
	return d+sizeof(int16);
}

/* ---- IMA */

/* 1.1 scale */
static const int16	IMAstep[89] =
{
	    7,    8,    9,   10,   11,   12,   13,   14,
	   16,   17,   19,   21,   23,   25,   28,   31,
	   34,   37,   41,   45,   50,   55,   60,   66,
	   73,   80,   88,   97,  107,  118,  130,  143,
	  157,  173,  190,  209,  230,  253,  279,  307,
	  337,  371,  408,  449,  494,  544,  598,  658,
	  724,  796,  876,  963, 1060, 1166, 1282, 1411,
	 1552, 1707, 1878, 2066, 2272, 2499, 2749, 3024,
	 3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484,
	 7132, 7845, 8630, 9493,10442,11487,12635,13899,
	15289,16818,18500,20350,22385,24623,27086,29794,
	32767
};

static const int8	IMAadjust[8] = {-1,-1,-1,-1, 2, 4, 6, 8};

uint8 *imanm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	int		l0;
	int16		x;

	if (!s || !d || !f || z<=0)
		return d;
	m*= sizeof(int16), x = f->kx, l0 = f->l0;
	goto start;
	do
	{
		int		v, u;
		uint8		b;

		d+= m;
	start:	b = *s;
		if (x>88u)
			x =~x>>31&88;
		u = IMAstep[x], v = u>>3;
		if (b&1)
			v+= u>>2;
		if (b&2)
			v+= u>>1;
		if (b&4)
			v+= u;
		if (b&8)
			v =-v;
		addclamp(l0,v);
		*(int16*)d = l0;
		x+= IMAadjust[b&7];
		if (x>88u)
			x =~x>>31&88;
		u = IMAstep[x], v = u>>3;
		if (b&0x10)
			v+= u>>2;
		if (b&0x20)
			v+= u>>1;
		if (b&0x40)
			v+= u;
		if (b&0x80)
			v =-v;
		addclamp(l0,v);
		*(int16*)(d+=m) = l0;
		x+= IMAadjust[b>>4&7];
	} while(++s,--z);
	f->kx = x, f->l0 = l0;
	return d+sizeof(int16);
}

uint8 *ima_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	int		i = 0, v, mz;

	if (!s || !d || !f || z<=0 || --m>1u)
		return d;
	do
	{
		if ((z-=4)<=0)
			return d;
		f[i].l0 = ((int16*)d)[i] = *s|s[1]<<8;
		v = s[2]|s[3]<<8, s+= 4;
		if (v>88u)
			return d;
		f[i].kx = v;
	} while(++i<m);
	for(mz=(m+1)*sizeof(int16),d+=mz; ; d+=mz*4-sizeof(int16))
	{
		i = 0;
		do
		{
			int		j = 4, l0 = f[i].l0, x = f[i].kx;

			do
			{
				uint8		b;

				b = *s;
				v = IMAstep[x]*((b+b+1)&15)>>3;
				if (b&8)
					v =-v;
				addclamp(l0,v);
				*(int16*)d = l0, d+= mz;
				x+= IMAadjust[b&7], b>>= 4;
				if (x>88u)
					x =~x>>31&88;
				v = IMAstep[x]*((b+b+1)&15)>>3;
				if (b&8)
					v =-v;
				addclamp(l0,v);
				*(int16*)d = l0, d+= mz;
				if (!--z)
					return d;
				x+= IMAadjust[b&7];
				if (x>88u)
					x =~x>>31&88;
			} while(++s,--j);
			f[i].l0 = l0, f[i].kx = x;
			d+= sizeof(int16)-mz*4;
		} while(++i<m);
	}
}

/* ---- OKI ADPCM */

static const int16	OKIstep[49] = /* ~12-bit precision */
{
	  256,  272,  304,  336,  368,  400,  448,  496,
	  544,  592,  656,  720,  800,  880,  960, 1056,
	 1168, 1280, 1408, 1552, 1712, 1888, 2080, 2288,
	 2512, 2768, 3040, 3344, 3680, 4048, 4464, 4912,
	 5392, 5936, 6528, 7184, 7904, 8704, 9568,10528,
	11584,12736,14016,15408,16960,18656,20512,22576,
	24832
};

uint8 *oki_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	uint32		mx, x;
	int		l0, k;

	if (!s || !d || !f || z<=0)
		return d;
	m*= sizeof(int16), x = f->kx, l0 = f->l0;
	k = f->l1, mx = f->mx;
	goto start;
	do
	{
		int		v, u;

		d+= m;
	start:	u = *s;
		if (x>mx)
			x =(signed)~x>>31&mx;
		v = ((u>>3|1)&15)*OKIstep[x]>>3&k;
		if (u&0x80)
			v =-v;
		addclamp(l0,v);
		*(int16*)d = l0;
		x+= IMAadjust[u>>4&7];
		if (x>mx)
			x =(signed)~x>>31&mx;
		v = ((u+u+1)&15)*OKIstep[x]>>3&k;
		if (u&8)
			v =-v;
		addclamp(l0,v);
		*(int16*)(d+=m) = l0;
		x+= IMAadjust[u&7];
	} while(++s,--z);
	f->kx = x, f->l0 = l0;
	return d+sizeof(int16);
}

/* ---- MS ADPCM */

static const int16	MSadapt[16] =
{
	230,230,230,230,307,409,512,614,
	768,614,512,409,307,230,230,230
};

static const int8	MScoef[14] =
{
	 64,128, 0, 48, 60,115, 98,
	  0,-64, 0, 16,  0,-52,-58
};

#define icdx(f)		(*(int16*)&(f)->kx)

int MSADPCMdecode1(adpcm_base *f,
	int u)
{
	int		v;

	v = (*(uint8*)f->coef*f->l0+f->coef[1]*f->l1)/64;
	addclamp(v,(u&8 ? u-16:u)*icdx(f));
	u = MSadapt[u]*icdx(f)>>8;
	if (u<16)
		u = 16;
	icdx(f) = u, f->l1 = f->l0, f->l0 = v;
	return v;
}

uint8 *ms_adpcm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	int		v;

	if (!s || !d || !f || z<=0 || --m>1u)
		return d;
	{ /* get the sample coefficients */
		const int8	*t;

		v = *s++;
		if (v>6u)
			return d;
		t = v+MScoef;
		f->coef[0] = *t, f->coef[1] = t[7];
		if (m)
		{
			if (!--z)
				return d;
			v = *s++;
			if (v>6u)
				return d;
			t = v+MScoef;
			f[1].coef[0] = *t, f[1].coef[1] = t[7];
		}
	}
	/* get the adaptive coeficients, and finally, the samples */
	if ((z-=3+2+2)<=0)
		return d;
	icdx(f) = *s|s[1]<<8, s+= 2;
	if (m)
	{
		if ((z-=2)<=0)
			return d;
		icdx(f+1) = *s|s[1]<<8, s+= 2;
	}
	f->l0 = *s|s[1]<<8, s+= 2;
	if (m)
	{
		if ((z-=2)<=0)
			return d;
		f[1].l0 = *s|s[1]<<8, s+= 2;
	}
	*(int16*)d = f->l1 = *s|s[1]<<8, s+= 2, d+= 2;
	if (m)
	{
		if ((z-=2)<=0)
			return d;
		*(int16*)d = f[1].l1 = *s|s[1]<<8, s+= 2, d+= 2;
	}
	*(int16*)d = f->l0, d+= 2;
	if (m)
		*(int16*)d = f[1].l0, d+= 2;
	do
	{
		uint8		b = *s++;

		*(int16*)d = MSADPCMdecode1(f,b>>4);
		((int16*)d)[1] = MSADPCMdecode1(f+m,b&15);
	} while(d+=4,--z);
	return d;
}

/* ---- Yamaha */

static const int16	YAMAHAadjust[] =
{
	230,230,230,230,307,409,512,614
};

static const int8	YAMAHAstep[] =
{
	  1,  3,  5,  7,  9, 11, 13, 15,
	 -1, -3, -5, -7, -9,-11,-13,-15
};

static inline int YAMAHAdecode1(adpcm_base *f,
	int b)
{
	int		v = f->l0;

	if (!f->kx)
		v = 0, f->kx = 127;
	addclamp(v,(f->kx*YAMAHAstep[b])/8);
	f->l0 = v;
	v = f->kx*YAMAHAadjust[b&7]>>8;
	f->kx = PIN(v,127,24567);
	return f->l0;
}

uint8 *adpcm_yamaha_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f)
{
	if (!s || !d || !f || z<=0 || --m>1u)
		return d;
	do
	{
		uint8		b = *s++;

		*(int16*)d = YAMAHAdecode1(f,b&15);
		((int16*)d)[1] = YAMAHAdecode1(f+m,b>>4);
	} while(d+=4,--z);
	return d;
}

#if 0
/* ---- EA */

static const int16	EAstep[] =
{
	   0, 240, 460, 392,
	   0,   0,-208,-220,
	   0,   1,   3,   4,
	   7,   8,  10,  11,
	   0,  -1,  -3,  -4
};

static const int8	SWFadjust[4][16] =
{
	{-1, 2},
	{-1,-1, 2, 4},
	{-1,-1,-1,-1, 2, 4, 6, 8},
	{-1,-1,-1,-1,-1,-1,-1,-1, 1, 2, 4, 6, 8,10,13,16}
};
#endif
