/*
 snddec.c
  sound decoder - main level

  Created: 2014.01.06 06:05 EST
  Revised: 2014.08.07
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.01.28
  added IMA-OKI, seek starting milliseconds.
  proper 2-byte alignment in AIFF(-C) reading.
2014.01.29
  added A-law and u-law; now all built-in Sound Manager 3.4 types are here.
  fixed categorization, splitting PCM into two (linear and non-linear PCM).
  removed the functions in xxlaw.c & just went w. a smaller, faster unpackNPCM8_16.
2014.01.30
  I've yet to get src on the ACE codecs.  disassemble IIGS Tool29 it is...some day.
2014.01.31
  got MS ADPCM to work; too bad the expected AIFF-C block size isn't compatible...
  added millseconds duration; sndTimeToPack is now sndTimeToFrame.
  fixed frame size calculation - added mSampleSize.
  made maxRead 32 bits; read the whole frickin' file for all I care...
2014.02.01
  added sound playback (Win32), info print-out.  worth it.
2014.03.23
  revamped sound playback to always use the virtual sound manager, which now works.
2014.03.24
  optimized byteswapping.
*/

#include <stdlib.h>
#include <string.h>

#include "enviro64.h"
#include "snddec.h"

/* ---- macros */

#define be24read(s) (((*(s)<<8|(s)[1])<<8|(s)[2])<<8)
#define le24read(s) ((((s)[2]<<8|(s)[1])<<8|*(s))<<8)

#if BIG_ENDIAN
 #define bs32read(s) ((((s)[3]<<8|(s)[2])<<8|(s)[1])<<8|*(s))
 #define bs16read(s) ((s)[1]<<8|*(s))
#else
 #define bs32read(s) (*(s)<<24|(s)[1]<<16|(s)[2]<<8|(s)[3])
 #define bs16read(s) (*(s)<<8|(s)[1])
 #if __POCC__ || _MSC_VER>=1300
  #undef bs32read
  #define bs32read(s) _reverse32(*(int32*)(s))
 #endif
#endif

/* ---- globals */

const static int32	cmpID2type[7] =
{
	0,
	CODEC_ACE_2,CODEC_ACE_8,
	CODEC_MACE_3,CODEC_MACE_6,
	CODEC_CDXA_2,CODEC_CDXA_4
};

/* ---- code */

static uint8 *unpackNPCM8_16(uint8 *d,
	const uint8 *s,
	uint32 z,
	const int16 *t)
{
	do
#if BIG_ENDIAN
		*((int16*)d)++ = t[*s++];
#else
		*(int16*)d = t[*s], d+= 2, ++s;
#endif
	while(--z);
	return d;
}

static void andeorcpy(uint8 *d,
	const uint8 *s,
	int c,
	uint32 a,
	uint32 m)
{
	int             x;

	if ((x=c>>2)!=0)
		do
#if BIG_ENDIAN
			*((int32*)d)++ = (*((int32*)s)++&a)^m;
#else
			*(int32*)d = (*(int32*)s&a)^m, s+= 4, d+= 4;
#endif
		while(--x);
	if (c&=3)
		do
			*d++ = (*s++&a)^m;
		while(--c);
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
static void __fastcall eorcpy(int c,
	uint32 m,
	uint8 *d,
	const uint8 *s)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		push	edi
		mov	edi,[esp+12]
		mov	esi,[esp+16]
		shr	ecx,1
		jnc	wo

		mov	al,[esi+ecx*2]
		xor	al,dl
		mov	[edi+ecx*2],al
	wo:	shr	ecx,1
		jnc	lo

		mov	ax,[esi+ecx*4]
		xor	ax,dx
		mov	[edi+ecx*4],ax
	lo:	shr	ecx,1
		jnc	qo

		mov	eax,[esi+ecx*8]
		xor	eax,edx
		mov	[edi+ecx*8],eax
	qo:	test	ecx,ecx
		jz	eo

	qc:	mov	eax,[esi]
		xor	eax,edx
		mov	[edi],eax
		mov	eax,[esi+4]
		xor	eax,edx
		mov	[edi+4],eax
		add	esi,8
		add	edi,8
		loop	qc

	eo:	pop	edi
		pop	esi
		ret	8
	}
#else
	int             x;

	if ((x=c>>3)!=0)
		do
		{
			*(uint32*)d = *(uint32*)s^m;
			((uint32*)d)[1] = ((uint32*)s)[1]^m;
		} while(s+=8,d+=8,--x);
	if (c&4)
		*(uint32*)d = *(uint32*)s^m, s+= 4, d+= 4;
	if (c&2)
		*(uint16*)d = *(uint16*)s^(uint16)m, d+= 2, s+= 2;
	if (c&=1)
		*d = *s^(uint8)m;
#endif
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
static void __fastcall bseorcpy2(uint32 c,
	uint32 m,
	int16 *d,
	const uint8 *s)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		push	edi
		mov	edi,[esp+12]
		mov	esi,[esp+16]
		shr	ecx,1
		jnc	lo

		mov	ax,[esi+ecx*4]
		xchg	al,ah
		xor	ax,dx
		mov	[edi+ecx*4],ax
	lo:	shr	ecx,1
		jnc	qo

		mov	eax,[esi+ecx*8]
		bswap	eax
		rol	eax,16
		xor	eax,edx
		mov	[edi+ecx*8],eax
	qo:	test	ecx,ecx
		jz	eo

	qc:	mov	eax,[esi]
		bswap	eax
		rol	eax,16
		xor	eax,edx
		mov	[edi],eax
		mov	eax,[esi+4]
		bswap	eax
		rol	eax,16
		xor	eax,edx
		mov	[edi+4],eax
		add	esi,8
		add	edi,8
		loop	qc

	eo:	pop	edi
		pop	esi
		ret	8	
	}
#else
	if (c&1)
		d[c-1] = bs16read(s+c*2-2)^m;
	if (c>>=1)
		do
			*(uint32*)d = (bs16read(!BIG_ENDIAN*2+s)<<16
				|bs16read(BIG_ENDIAN*2+s))^m;
		while(d+=2,s+=4,--c);
#endif
}

static uint8 *unpackPCM(uint8 *d,
	const uint8 *s,
	uint32 z,
	const sndFormat *f)
{
	int		m = f->type<<24&0x80000000u;

	switch(f->depth)
	{ /* AIFF supports smaller depths */
	case 1:	case 2: case 3: case 4: case 5: case 6: case 7:
		/* lower bits are padding; they are not valid */
		m|= m>>16, m|= m>>8;
		andeorcpy(d,s,z,(uint8)(~0xFF>>f->depth)*0x1010101,m);
		break;
	case 8:	if (f->type&STYPE_ISGN)
			goto direct;
		eorcpy(z,0x80808080u,d,s);
		break;
	case 9:	case 10: case 11: case 12: case 13: case 14: case 15:
		/* lower bits are padding; they are not valid */
	case 16:
		if ((!BIG_ENDIAN<<6)==(int8)(f->type&~STYPE_FP))
			goto direct;
		m|= m>>16;
		if ((!BIG_ENDIAN<<6)==(f->type&STYPE_LSB))
			eorcpy(z,m,d,s);
		else
			bseorcpy2(z>>1,m,(int16*)d,s);
		break;
	case 17: case 18: case 19: case 20: case 21: case 22: case 23:
		/* lower bits are padding; they are not valid */
	case 24: /* unpacked to 32-bit */
		{
			int32		*w = (int32*)d;
			uint32		x = z/3;

			if ((z=x<<2)<=0)
				break;
			if (f->type&STYPE_LSB)
			{
				if (x&1)
					*w = le24read(s)^m;
				if (x>>=1)
					do
						*w++ = le24read(s)^m, *w++ = le24read(s+3)^m;
					while(s+=6,--x);
			} else
			{
				if (x&1)
					*w = be24read(s)^m;
				if (x>>=1)
					do
						*w++ = be24read(s)^m, *w++ = be24read(s+3)^m;
					while(s+=6,--x);
			}
		}
		break;
	case 25: case 26: case 27: case 28: case 29: case 30: case 31:
		/* lower bits are padding; they are not valid */
	case 32:
		if ((!BIG_ENDIAN<<6)==(int8)(f->type&~STYPE_FP))
			goto direct;
		if ((!BIG_ENDIAN<<6)==(f->type&STYPE_LSB))
			eorcpy(z,m,d,s);
		else
		{
			int32		*w = (int32*)d;
			uint32		x = z>>2;

			if (x>0)
				do
					*w = bs32read(s)^m, w[1] = bs32read(s+4)^m;
				while(w+=2,s+=8,--x);
			if (z&4)
				*w = bs32read(s)^m;
		}
		break;
	case 64:
		if ((!BIG_ENDIAN<<6)==(int8)(f->type&~STYPE_FP))
		{
		direct:	memcpy(d,s,z);
			break;
		}
		if (z>7)
		{
			int32	*w = (int32*)d, x = z>>3;

			if (f->type&STYPE_LSB)
				do
#if BIG_ENDIAN
					*w++ = bs32read(s+4)^m, *w++ = bs32read(s);
				while(--x);
#else
					*w++ = *(int32*)s, *w++ = *(int32*)(s+4)^m;
				while(s+=8,--x);
#endif
			else
				do
#if BIG_ENDIAN
					*w++ = *(int32*)s^m, *w++ = *(int32*)(s+4);
				while(--x);
#else
					*w++ = bs32read(s), *w++ = bs32read(s+4)^m;
				while(s+=8,--x);
#endif
		}
	}
	d+= z;
	return d;
}

static void reverseChannels(uint8 *d,
	int32 z,
	const sndFormat *f)
{
	int32		c, x;
	int		chc, chm, chk;

	if ((chc=f->chanCount&127)<2 || (c=z/f->mSampleSize)<=0)
		return;
	chm = chc>>1, chk = chm+(chc&1);
	switch(f->mSampleSize)
	{
	case 1:
		do
		{
			uint8	*s, v;

			x = chm, s = d+chc;
			do
				v = *d, *d = *--s, *s = v;
			while(++d,--x);
		} while(d+=chk,--c);
		break;
	case 2:
		do
		{
			int16	*s, v;

			x = chm, s = (int16*)d+chc;
			do
				v = *(int16*)d, *(int16*)d = *--s, *s = v;
			while(d+=2,--x);
		} while(d+=chk,--c);
		break;
	case 4:
		do
		{
			int32	*s, v;

			x = chm, s = (int32*)d+chc;
			do
				v = *(int32*)d, *(int32*)d = *--s, *s = v;
			while(d+=4,--x);
		} while(d+=chk,--c);
		break;
	case 8:
		do
		{
			double	*s, v;

			x = chm, s = (double*)d+chc;
			do
				v = *(double*)d, *(double*)d = *--s, *s = v;
			while(d+=8,--x);
		} while(d+=chk,--c);
	}
}

uint8 *decompressSndFrame(uint8 *_d,
	const uint8 *s,
	sndFormat *f,
	int z)
{
	uint8		t[4], *d, *r, i, chc, stype;
	int		c;

	if ((d=_d)==NULL || !f || (chc=f->chanCount&127)==0)
		return NULL;
	if (z<=0 && s)
		return d;
	i = 0, r = d;
	if ((stype=f->type&15)!=0)
	{
		if (stype!=CTYPE_NPCM)
			goto start;
		if (!s)
			goto end;
		if (f->codec==CODEC_ALAW)
			r = unpackNPCM8_16(d,s,z,Alaw16btable);
		else if (f->codec==CODEC_ULAW)
			r = unpackNPCM8_16(d,s,z,ulaw16btable);
		else
			goto na;
		goto end;
	}
	if (s)
		if ((f->depth==8 && f->type&STYPE_ISGN) || (!BIG_ENDIAN<<6)==f->type)
			memcpy(d,s,z), r = d+z;
		else
			r = unpackPCM(d,s,z,f);
	goto end;
	do
	{
		adpcm_base	*ch; /* ADPCM base by default */

		if (s)
			s+= f->packSize, d+= f->mSampleSize, z-= f->packSize;
	start:	c = MIN(z,f->packSize);
		ch = i+(adpcm_base*)f->variables;
		if (stype==CTYPE_ADPCM)
		{
			if (f->codec==CODEC_CDXA_4 || f->codec==CODEC_CDXA_2)
			{
				if (!s)
				{ /* init */
					memcpy(ch,&brr_default,16);
					ch->kx = 6;
					continue; /* stereo fix (was break); 2014.01.27 */
				}
				if (!--c)
					c = 1, *t = *s, t[1] = 0, s = t;
				if (f->codec==CODEC_CDXA_2)
					c = 1; /* unsure about this, CD/XA 2:1 */
				r = brr_decode(d,s,chc,MIN(14,c),ch);
			} else if (f->codec==CODEC_AIFC_IMA)
			{
				if (!s) /* IMA frames define the variables */
					{ch->mx = 88; break;}
				ch->l0 = *s<<8, ch->kx = 0;
				if (--c)
					--c, ch->l0|= s[1]&0x80, ch->kx = s[1]&0x7F;
				r = imanm_decode(d,s+2,chc,MIN(32,c),ch);
			} else if (f->codec==CODEC_DVI_IMA)
			{
				if (!s) /* IMA frames define the variables */
					{ch->mx = 88; break;}
				/* decoder will handle both channels */
				r = ima_decode(d,s,chc,MIN(1020,c),ch);
				break;
			} else if (f->codec==CODEC_OKI_ADPCM)
			{
				if (!s)
				{ /* stored OKI frames have no initializers */
					ch->mx = 48, ch->l0 = 0, ch->l1 =~0xFFFF>>12;
					continue;
				}
				r = oki_decode(d,s,chc,MIN(256,c),ch);
			} else if (f->codec==CODEC_MS_ADPCM)
			{ /* stereo packets are done all-in-one */
				if (!s)
				{
					//ch->l0 = ch->l1 = ch[1].l0 = ch[1].l1 = 0;
					break;
				}
				/* decoder will handle both channels...everything */
				r = ms_adpcm_decode(d,s,chc,z,ch);
				break;
			} else
				goto na;
		} else if (stype==CTYPE_TDPCM)
		{
			if (f->codec==CODEC_MACE_3)
			{
				if (!s)
				{ /* init */
					*(int32*)ch = 0;
					continue; /* stereo fix; 2014.01.27 */
				}
				r = MACE3_decode(d,s,chc,c,(int16*)ch);
			} else if (f->codec==CODEC_MACE_6)
			{
				if (!s)
				{ /* init */
					memset(ch,0,12);
					continue; /* stereo fix; 2014.01.27 */
				}
				r = MACE6_decode(d,s,chc,c,(int16*)ch);
			} else
				goto na;
		} else
			{na: s = NULL; break;}
	} while(++i<chc);
end:	if (s)
	{
		if (f->maxRead>f->frameCount)
			f->frameCount = 0;
		else
			f->frameCount-= f->maxRead;
		if (f->chanCount&128)
			reverseChannels(_d,r-_d,f);
	}
	return r;
}

/* ---- parsing, identification */

int sndFixCmpType(uint32 c,
	sndFormat *f)
{
	if (!f)
		return -1;
	/* first, check for alternate types */
	if (c==CODEC_MS_ALAW || c==MK4CC('A','L','A','W'))
		{c = CODEC_ALAW; goto alaw;}
	else if (c==CODEC_MS_ULAW || c==MK4CC('U','L','A','W'))
		{c = CODEC_ULAW; goto alaw;}
	else if (c==MK4CC('F','L','3','2')) /* for CSound, SoundHack, etc. */
		c = CODEC_PCM_BE_F32;
	else if (c==CODEC_MS_G726_ADPCM)
		c = CODEC_G726_ADPCM;
	/* now make sure the type, depth, packet, etc. values are correct */
	if (c==CODEC_ALAW || c==CODEC_ULAW)
	alaw: f->frameSize = 1, f->packSize = 1, f->depth = 12, f->type = CTYPE_NPCM;
	else if (c==CODEC_MACE_3)
		f->frameSize = 6, f->packSize = 2, f->depth = 9, f->type = CTYPE_TDPCM; /* was 2:ADPCM */
	else if (c==CODEC_MACE_6)
		f->frameSize = 6, f->packSize = 1, f->depth = 10, f->type = CTYPE_TDPCM; /* was 2:ADPCM */
	else if (c==CODEC_ACE_2)
		f->frameSize = 2, f->packSize = 1, f->depth = 8, f->type = CTYPE_ADPCM;
	else if (c==CODEC_ACE_8)
		f->frameSize = 8, f->packSize = 3, f->depth = 8, f->type = CTYPE_ADPCM;
	else if (c==CODEC_DVI_IMA)
	{
		if (!f->packSize)
			f->packSize = 1024;
		if (!f->frameSize)
			f->frameSize = f->packSize*2-8; /* WAVE: packSize*2-7 */
		f->depth = 16, f->type = STYPE_LSB|CTYPE_ADPCM;
	} else if (c==CODEC_MS_ADPCM) /* Microsoft ADPCM - usu. 1012/512 in WAVEs */
	{
		if (!f->packSize)
			f->packSize = 256;
		f->frameSize = f->packSize*2-12, f->depth = 16, f->type = STYPE_LSB|CTYPE_ADPCM;
	} else if (c==CODEC_AIFC_IMA)
		f->frameSize = 64, f->packSize = 34, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_CDXA_4)
		f->frameSize = 28, f->packSize = 15, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_CDXA_2) /* unsure about this */
		f->frameSize = 2, f->packSize = 2, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_OKI_ADPCM)
		f->frameSize = 512, f->packSize = 256, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_MEDIASPACE || c==CODEC_YAMAHA_ADPCM)
		f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_GSM) /* MS packSize=65; MSN=41 */
		f->packSize = 33, f->depth = 16, f->type = 15;
	else if (c==CODEC_QUALCOMM)
		f->packSize = 35, f->depth = 16, f->type = 15;
	else if (c==CODEC_MPEG_L3 || c==CODEC_MPEG_L3P) /* an MPEG reader is needed for this */
		f->depth = 16, f->type = CTYPE_MPEG; /*f->frameSize = 1152; // max. frame size */
	else if (c==CODEC_PCM_BE_S24)
		f->depth = 24;
	else if (c==CODEC_PCM_BE_S32)
		f->depth = 32;
	else if (c==CODEC_PCM_LE_S32)
		f->type = STYPE_LSB, f->depth = 32;
	else if (c==CODEC_PCM_BE_F32)
		f->type = STYPE_FP, f->depth = 32;
	else if (c==CODEC_PCM_LE_F32)
		f->type = STYPE_FP|STYPE_LSB, f->depth = 32;
	else if (c==CODEC_PCM_BE_F64)
		f->type = STYPE_FP, f->depth = 64;
	else if (c==CODEC_PCM_LE_F64)
		f->type = STYPE_FP|STYPE_LSB, f->depth = 64;
	else if (c==CODEC_PCM_BE_S16 && f->depth!=8)
		f->depth = 16;
	else if (c==CODEC_PCM_LE_S16 && f->depth!=8) /* was missing; 2014.08.07 */
		f->type = STYPE_LSB, f->depth = 16;
	if (f->depth<8)
		if (c==CODEC_PCM_U8 || !c || c==CODEC_PCM_DEFAULT)
			f->depth = 8;
		else
		{ /* default: 16 bits */
			if (c!=CODEC_PCM_BE_S16 && c!=CODEC_PCM_LE_S16)
				return -8;
			f->depth = 16;
		}
	if (!f->packSize && (f->packSize=(f->depth+7)>>3)==0)
		f->packSize = 1;
	if (c==CODEC_MS_PCM)
		if (f->depth==8)
			c = CODEC_PCM_U8;
		else
			c = CODEC_PCM_LE_S16, f->type = STYPE_LSB;
		/* other depths are not WAVE-supported */
	else if (c==CODEC_MS_FLOAT)
	{
		f->type = STYPE_FP|STYPE_LSB;
		if (f->depth==64)
			c = CODEC_PCM_LE_F64;
		else
			c = CODEC_PCM_LE_F32;
	} else if (!c)
		switch(f->depth)
		{ /* get appropriate AIFC type */
		case 8: break;
		default: c = CODEC_PCM_BE_S16, f->depth = 16; break;
		case 24: c = CODEC_PCM_BE_S24; break;
		case 32: c = CODEC_PCM_BE_S32; break;
		case 64: c = CODEC_PCM_BE_F64; break; /* ...whatever... */
		}
	f->codec = c;
	return 0;
}

void sndFixCodecFrame(sndFormat *f)
{
	int		a, b;

	if (!f)
		return;
	if (f->codec==CODEC_PCM_U8)
		f->type = STYPE_ISGN;
	if (!f->blockSize)
		f->blockSize = f->packSize*(f->chanCount&127);
	if (!f->frameSize)
		f->frameSize = 1;
	for(a=(f->depth+7)>>3,b=1; b<a; )
		b+= b;
	f->mSampleSize = b;
	f->frameSize*= f->chanCount*b;
	f->maxRead = (f->type&15)>CTYPE_NPCM ? 1:512; //144
}

int32 parseBufferCmdHeader(const uint8 *s,
	sndFormat *f,
	uint8 *n)
{
	uint32		c, o = -(int32)s;
	uint8		h;
	/* must read header type first */
	/* samplePtr should be NULL! */
	if (!s || !f || ((h=s[20])-1u<(0xFEu-1u)) || va_be32(s))
		return -1;
	memset(f,0,sizeof(*f));
	if ((signed)(c=va_be32(s))<0)
		return -3; /* bit #31 is invalid for a stored bufferCmd */
	if (n)
		*n = s[21-8];
	if (!h)
	{ /* stdSH: count is #mono frames, uint8 */
		f->frameCount = c, f->codec = CODEC_PCM_U8;
		f->depth = 8, f->packSize = 1;
		c = 1;
	} /* extSH/cmpSH: count is #channels, frame count follows basenote */
	if (c-1u>1u)
		return -1;
	f->chanCount = c;
	c = va_be32(s);
	/* rate: max(c,1Hz); QuickTime doesn't even allow rates below 1KHz */
	if (c<(1<<16))
		return -1;
	f->rate = c>>16; /* ignoring float80 rate, as Sound Manager 3.4 does */
	f->rateFrac = c<<16;
	s+= 10; /* loopStart.l/loopEnd.l/enc.b/baseNote.b */
	if (!h)
		goto end;
	f->frameCount = va_be32(s);
	s+= 10+4; /* AIFFRate.t/markerChunk.p */
	if (h==0xFF)
	{
		s+= 8; /* insChunks.p/AESRecording.p/reserved.. */
		//f->packSize = 0;
		c = 0; /* figure out type by depth */
	} else
	{
		c = *(int32*)s;
		s+= 4+14; /* reserved.l/stateVars.p/leftOverSamples.p/cmpID.w */
		if (!c)
		{ /* cmpID values: 0=none/implied */
		  /* -1=fixed (default MAC3 for 8-bit) */
		  /* -2=variable-sized */
			if ((signed)(c=(int16)be16_native(((const int16*)s)[-1]))<0)
				c = s[5]==8 ? 3:0;
			if (c>6)
				return -8;
			c = cmpID2type[c];
		}
		f->packSize = be16_native(*(const int16*)s)>>3;
		s+= 4;
	}
	f->depth = s[1];
	/* without a codec, 8 bit is always unsigned */
	if ((!c || c==CODEC_PCM_DEFAULT) && f->depth==8)
		c = CODEC_PCM_U8;
	if (sndFixCmpType(c,f) || *s)
		return -1; /* bad codec or depth */
	s+= h==0xFF ? 16:2;
end:	sndFixCodecFrame(f);
	return o+(int32)s;
}

int parseSoundSampleHeader(const uint8 *s,
	sndFormat *f,
	int32 z)
{
	int8		stereo;

	if (!s || !f)
		return -1;
	if (z<=0)
		return -5;
	memset(f,0,sizeof(*f));
	stereo = !!(s[6]&15);
	if ((f->rate=s[9]<<8|s[8])==0)
		f->rate = 11127; /* default */
	f->chanCount = stereo ? 0x82:1; /* ordering reversed in IIgs AIFF-C files */
	f->frameCount = z>>stereo;
	sndFixCmpType(CODEC_PCM_U8,f);
	sndFixCodecFrame(f);
	return 10;
}

int parseExt80Rate(const uint8 *s,
	sndFormat *f)
{
	uint64		v;
	uint32		a, b;
	int16		x;

	if (!s)
		return -1;
	x = *s<<8|s[1], s+= 2;
	a = va_be32(s), b = va_be32(s);
	/* normalize */
	if (!a)
	{
		if ((a=b)==0)
			if (x!=0x7FFF)
				goto zero;
			else /* INF */
				goto ovfl;
		b = 0, x-= 32;
	}
	int64sethl(v,a,b);
	if (!(~x&0x7FFF)) /* NaN */
		return 0;
	if (x<0) /* fixed-point value is unsigned */
		goto zero;
	for(; (signed)int64hi(v)>=0; --x)
		uint64add(v,v);
	x =-x, x+= 0x401E;
	if (x<0)
		{ovfl: f->rate = f->rateFrac =-1u; return 0;}
	else if (x>63u)
	zero:	int64xset(v,0); /* zero or underflow */
	uint64shr(v,x);
	f->rate = int64hi(v), f->rateFrac = int64lo(v);
	return !!f->rate;
}

const char *sndGetCodecName(int32 c)
{
	switch(c)
	{
	case CODEC_MS_PCM: return "MS PCM";
	case CODEC_MS_ADPCM: return "MS ADPCM";
	case CODEC_MS_FLOAT: return "MS float";
	case CODEC_MS_VSELP: return "MS VSELP";
	case CODEC_IBM_CVSD: return "IBM CVSD";
	case CODEC_OKI_ADPCM: return "OKI";
	case CODEC_DVI_IMA: return "DVI IMA 4:1";
	case CODEC_MEDIASPACE: return "MediaSpace";
	case CODEC_SIERRA_ADPCM: return "Sierra";
	case CODEC_G723_ADPCM: return "Antex G723";
	case CODEC_DSP_DIGISTD: return "DSP DigiStd";
	case CODEC_DSP_DIGIFIX: return "DSP DigiFix";
	case CODEC_DIALOGIC_OKI: return "Dialogic OKI";
	case CODEC_MEDIAVISION: return "Media Vision";
	case CODEC_CU_HP: return "Hewlett-Packard CU";
	case CODEC_YAMAHA_ADPCM: return "Yamaha";
	case CODEC_SONARC: return "SC Sonarc";
	case CODEC_TRUESPEECH: return "DSP TrueSpeech";
	case CODEC_ECHOSC1: return "Echo Speech SC1";
	case CODEC_AUDIOFILE_AF36: return "AudioFile AF36";
	case CODEC_APTX: return "APTX";
	case CODEC_AUDIOFILE_AF10: return "AudioFile AF10";
	case CODEC_PROSODY_1612: return "Aculab Prosody 1612";
	case CODEC_LRC: return "Merging Technologies LRC";
	case CODEC_DOLBY_AC2: return "Dolby AC2";
	case CODEC_MS_GSM_610: return "MS GSM 6.10";
	case CODEC_MSNAUDIO: return "MSN Audio";
	case CODEC_ANTEX_ADPCME: return "Antex ADPCME";
	case CODEC_CONTROLRES_VQLPC: return "Control Resources VQLPC";
	case CODEC_DIGIREAL: return "DSP DigiReal";
	case CODEC_DIGIADPCM: return "DSP DigiADPCM";
	case CODEC_CONTROLRES_CR10: return "Control Resources CR10";
	case CODEC_NMS_VBXADPCM: return "Natural MicroSystems VBXADPCM";
	case CODEC_ROLAND_RDAC: return "Roland RDAC";
	case CODEC_ECHOSC3: return "Echo Speech Codec 3";
	case CODEC_ROCKWELL_ADPCM: return "Rockwell ADPCM";
	case CODEC_ROCKWELL_DIGITALK: return "Rockwell DigiTalk";
	case CODEC_XEBEC: return "Xebec Multimedia";
	case CODEC_G721_ADPCM: return "Antex G721";
	case CODEC_G728_CELP: return "Antex G728 CELP";
	case CODEC_G723_1: return "MS G723.1";
	case CODEC_MS_MPEG: return "MS MPEG";
	case CODEC_RT24: return "InSoft RT24";
	case CODEC_PAC: return "InSoft PAC";
	case CODEC_MPEG_L3: return "MPEG L3";
	case CODEC_LUCENT_G723: return "Lucent G723";
	case CODEC_CIRRUS: return "Cirrus Logic";
	case CODEC_ESPCM: return "ESS ESPCM";
	case CODEC_MS_G726_ADPCM: return "APICOM G726";
	case CODEC_MS_G722_ADPCM: return "APICOM G722";
	case CODEC_MS_DSAT: return "MS DSAT";
	case CODEC_DSAT_DISPLAY: return "MS DSAT Display";
	case CODEC_VOXWARE_ALIGNED: return "Voxware Aligned";
	case CODEC_VOXWARE_AC8: return "Voxware AC8";
	case CODEC_VOXWARE_AC10: return "Voxware AC10";
	case CODEC_VOXWARE_AC16: return "Voxware AC16";
	case CODEC_VOXWARE_AC20: return "Voxware AC20";
	case CODEC_VOXWARE_RT24: return "Voxware RT24";
	case CODEC_VOXWARE_RT29: return "Voxware RT29";
	case CODEC_VOXWARE_RT29HW: return "Voxware RT29HW";
	case CODEC_VOXWARE_VR12: return "Voxware VR12";
	case CODEC_VOXWARE_VR18: return "Voxware VR18";
	case CODEC_VOXWARE_TQ40: return "Voxware TQ40";
	case CODEC_SOFTSOUND: return "SoftSound";
	case CODEC_VOXWARE_TQ60: return "Voxware TQ60";
	case CODEC_MS_RT24: return "MS RT24";
	case CODEC_G729A: return "AT&T G729A";
	case CODEC_DF_GSM610: return "DataFusion GSM 610";
	case CODEC_SBC24: return "SBC24";
	case CODEC_ZYXEL_ADPCM: return "ZyXEL ADPCM";

	default: switch(c)
		{
		case CODEC_MS_ALAW:
		case CODEC_ALAW: return "A-law";
		case CODEC_MS_ULAW:
		case CODEC_ULAW: return "mu-law"; /* micro character difficult to make here */
		case CODEC_MACE_3: return "MACE 3:1";
		case CODEC_MACE_6: return "MACE 6:1";
		case CODEC_ACE_2: return "ACE 2:1";
		case CODEC_ACE_8: return "ACE 8:3";
		case CODEC_AIFC_IMA: return "IMA 4:1 (64:17)";
		case CODEC_CDXA_4: return "CD/XA 4:1";
		case CODEC_CDXA_2: return "CD/XA 2:1";
		case CODEC_DVCA: return "DV Audio";
		case CODEC_QDESIGN: return "QDesign music";
		case CODEC_QDESIGN2: return "QDesign2 music";
		case CODEC_QUALCOMM: return "QUALCOMM PureVoice";
		case CODEC_DWVW: return "DWVW";
		case CODEC_GSM: return "Apple GSM";
		case CODEC_G726_ADPCM: return "G726 ADPCM";
		case CODEC_ADP4: return "ADP4";
		case CODEC_MPEG_L3P: return "MPEG Layer II/III";
		case CODEC_PCM_DEFAULT: return "none";
		case CODEC_PCM_U8: return "offset binary";
		case CODEC_PCM_BE_S16:
		case CODEC_PCM_LE_S16:
		case CODEC_PCM_BE_S24:
		case CODEC_PCM_BE_S32:
		case CODEC_PCM_LE_S32: return "integer";
		case CODEC_PCM_BE_F32:
		case CODEC_PCM_BE_F64:
		case CODEC_PCM_LE_F32:
		case CODEC_PCM_LE_F64: return "floating point";
		case CODEC_RHETOREX_ADPCM: return "Rhetorex ADPCM";
		case CODEC_WMA1: return "WMA v1";
		case CODEC_WMA2: return "WMA v2";
		case CODEC_WMAPRO: return "WMA Pro";
		case CODEC_WMALOSSLESS: return "WMA Lossless";
		case CODEC_ADPCM_CT: return "ADPCM CT";
		case CODEC_IMC: return "IMC";
		case CODEC_IAC_2: return "Indeo(R) IAC v2";
		case CODEC_VME_VMPCM: return "VME VMPCM";
		case CODEC_LH: return "Lernout & Hauspie";
		case CODEC_AAC_LATM: return "AAC LATM";
		case CODEC_DTS: return "DTS";
		case CODEC_SPEEX: return "Speex";
		case CODEC_FLAC: return "FLAC";

		default: return NULL;
		}
	}
}

/* ---- size and time evaluation */

int sndMaxFrameSize(const sndFormat *f)
{
	if (!f)
		return 0;
	return f->frameSize*f->maxRead;
}

int32 sndTimeToFrame(uint32 ms,
	const sndFormat *f)
{
	int32		u;

	if (!f)
		return -1;
	if ((f->type&15)==CTYPE_MPEG)
		return -8; /* MPEG functions not implemented here */
	u = (f->chanCount&127)*f->mSampleSize;
	if (u<=0) /* avoid dbz */
		return -1;
	u = f->frameSize/u*1000; /* faster multiplying 1000 here then dividing */
	if (u<=0)
		return -1;
	/* wanted to do tbyte loading (and wrote _asm) */
	/* but unlike M68K the Intel FPU doesn't auto-normalize! */
	/* fma also sucks - the function does more testing than math */
	return (f->rateFrac*(1/4294967296.)+f->rate)*ms/u+.5;
}

uint32 sndFrameToTime(uint32 x,
	const sndFormat *f,
	int rounded)
{
	long double	r, n;
	int32		u;

	if (!f || !f->rate || !f->frameSize || (u=(f->chanCount&127)*f->mSampleSize)<=0)
		return 0;
	if (x>f->frameCount)
		x = f->frameCount;
	n = x*(f->frameSize/u);
	if (!int64z(f->totalCount) && n>(r=int64tox(f->totalCount)))
		n = r;
	r = (f->rateFrac*(1/4294967296.)+f->rate);
	if (rounded)
		r = (uint32)(r+.5);
	return n*1000./r+.5;
}

static int store16bs32(void *d,
	const uint8 *r,
	int c,
	int8 t)
{
	int		x;

	if (t&STYPE_MAX16)
	{ /* 16-bit maximum */
		int16	*w = (int16*)d;
		int	a = (t>>6&1)+2*!BIG_ENDIAN, b = a^1;

		if ((x=c>>1)!=0)
			do
				*w++ = r[a]<<8|r[b],
					*w++ = r[4+a]<<8|r[4+b];
			while(r+=8,--x);
		if (c&1)
			*w = r[a]<<8|r[b], r+= 4;
		c+= c;
	} else
	{
		int32	*w = (int32*)d;

		if ((x=c>>1)!=0)
			do
				*w++ = bs32read(r),
					*w++ = bs32read(4+r);
			while(r+=8,--x);
		if (c&1)
			*w = bs32read(r), r+= 4;
		c<<= 2;
	}
	return c;
}

static void f32i32(int32 *d,
	const float *s,
	int x)
{
	do
	{
		double		v = *s*2147483648.-.5;

		if (v>2147483647)
			*d = 2147483647;
		else
			*d = v;
	} while(++s,++d,--x);
}

static void f64i32(int32 *d,
	const float *s,
	int x)
{
	do
	{
		double		v = *s*2147483648.-.5;

		if (v>2147483647)
			*d = 2147483647;
		else
			*d = v;
	} while(++s,++d,--x);
}

long writeSndSamples(const void *p,
	long z,
	void *outs,
	const sndFormat *f,
	uint8 t,
	sndwfunc_t wf)
{
	long		oz, tz, c;
	uint8		b[64];

	if (!p || !f)
		return -1;
	if (z<=0)
		return 0;
	if (((t&STYPE_LSB)==STYPE_ENDIAN_NATIVE /* desired/compatible endianess? */
			|| (f->depth<9 && (int8)(t&STYPE_ISGN))) /* allow S8? */
			&& !(int8)(f->type&~t&STYPE_FP) /* fp allowed? */
			&& (f->depth<17 || !(t&STYPE_MAX16))) /* max. 16-bit? */
		return wf(p,z,outs);
	tz = 0;
	if (f->depth<9)
	{ /* we'll only get here if S8 was forced (invert sign bits) */
		while((c=MIN(64,z))>0)
		{
			z-= c;
			eorcpy(c,0x80808080u,b,p);
			if ((c=wf(b,c,outs))==0)
				break;
			tz+= c, p = c+(const int8*)p;
		}
	} else if (f->depth<17)
	{ /* 16-bit (int16 only, differing endian) */
		z>>= 1;
		while((c=MIN(32,z))>0)
		{
			int		x;

			z-= c;
			bseorcpy2(c,0,(int16*)b,p);
			x = c+c;
			if ((c=wf(p,x,outs))==0)
				break;
			tz+= c, p = x+(const int8*)p;
		}
	} else if (f->depth<33)
	{ /* 32-bit: int32 or float */
		z>>= 2;
		while((c=MIN(16,z))>0)
		{
			const uint8	*r = p;

			z-= c;
			if ((int8)(f->type&~t&STYPE_FP))
			{ /* float not allowed; convert to int32 */
				f32i32((int32*)b,p,c);
				r = (const uint8*)b;
			}
			if ((oz=wf(b,store16bs32(b,r,c,t),outs))==0)
				break;
			tz+= oz, p = c+(float*)p;
		}
	} else
	{ /* 64-bit (doubles only) */
		z>>= 3;
		while((c=MIN(8,z))>0)
		{
			const uint8	*r = p;

			z-= c;
			if ((int8)((f->type^t)&STYPE_FP) || t&STYPE_MAX16)
			{ /* float not allowed; convert to int32 */
				f64i32((int32*)b,p,c);
				r = (const uint8*)b;
				oz = store16bs32(b,r,c,t);
			} else
			{
				uint8	*w = b;
				int	x;

				x = c;
				do
				{
					*w = r[7*!BIG_ENDIAN];
					w[1] = r[1+5*!BIG_ENDIAN];
					w[2] = r[2+3*!BIG_ENDIAN];
					w[3] = r[3+1*!BIG_ENDIAN];
					w[4] = r[4-1*!BIG_ENDIAN];
					w[5] = r[5-3*!BIG_ENDIAN];
					w[6] = r[6-5*!BIG_ENDIAN];
					w[7] = r[7*BIG_ENDIAN];
				} while(w+=8,r+=8,--x);
				oz = c<<3, p = r;
			}
			if ((oz=wf(b,oz,outs))==0)
				break;
			tz+= oz, p = c+(double*)p;
		}
	}
	return tz;
}

/* ---- header setup */

static uint8	WAVEheader[44] =
/*x00*/	"RIFF\0\0\0\0WAVEfmt " /* 0x004:uint32 total LSB */
/*x10*/	"\20\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0"
/*x20*/	"\0\0\0\000data"; /* 0x028:uint32 data size LSB */

const uint8 *setupWAVEHeader(const sndFormat *f,
	long *t)
{
	uint8		*d = WAVEheader;
	int		c, z, m;

	if (!f)
		return NULL;
	/* WAVEFORMAT: tag.w, chnCnt.w, rate.l, sfsize.l, balgn.w */
	/* PCMWAVEFORMAT: WAVEFORMAT, depth.w */

	/* floating point not really supported in 32-bit WAVE */
	//d[20] = f->type&STYPE_FP ? (uint8)CODEC_MS_FLOAT:(uint8)CODEC_MS_PCM;
	z = 1+(f->depth>8);
	if ((m=f->chanCount&127)==0)
		m = 1;
	d[22] = m;
	c = f->rate+(f->rateFrac>>31);
	c+= !c;
	d[24] = (uint8)c;
	d[25] = (uint8)(c>>8);
	d[26] = (uint8)(c>>16);
	d[27] = (uint8)(c>>24);
	d[34] = (uint8)(z<<3);
	d[35] = (uint8)(z>>5);
	z*= m, c*= z;
	d[28] = (uint8)c;
	d[29] = (uint8)(c>>8);
	d[30] = (uint8)(c>>16);
	d[31] = (uint8)(c>>24);
	d[32] = (uint8)z;
	d[33] = (uint8)(z>>8);
	z = t ? *t:0;
	d[40] = (uint8)z;
	d[41] = (uint8)(z>>8);
	d[42] = (uint8)(z>>16);
	d[43] = (uint8)(z>>24);
	z+= 36;
	d[4] = (uint8)z;
	d[5] = (uint8)(z>>8);
	d[6] = (uint8)(z>>16);
	d[7] = (uint8)(z>>24);
	if (t)
		*t = 44;
	/* form: componentbits.l, expbytes.l */
	return d;
}

/* ---- basic write functions */

uint32 sndwrite_memcpy(const void *p,
	uint32 z,
	void *d)
{
	memcpy(*(uint8**)d,p,z);
	*(uint8**)d+= z;
	return z;
}

uint32 sndwrite_null(const void *p,
	uint32 z,
	void *d)
{
#pragma unused(p,z,d)
	return z;
}
