/*
 sndfile.c
  file reading/writing

  Revised: 2014.09.15

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.24/5
  Added WAVE file parsing.
*/

#include <stdio.h>
#include <string.h>
#include "enviro64.h"
#include "snddec.h"

/* ---- basic functions */

int readSndFrame(uint8 *d,
	FILE *s,
	const sndFormat *f)
{
	if (!d || !s || !f)
		return -1;
	return fread(d,1,f->blockSize*MIN(f->frameCount,f->maxRead),s);
}

uint32 sndwrite_fwrite(const void *p,
	uint32 z,
	void *s)
{
	return fwrite(p,1,z,s);
}

/* ---- header parsing */

static long int parseFileFORM(FILE *s,
	sndFormat *f)
{ /* FORM: AIFx */
	uint8		b[40];
	long int	o =-1;
	const int32	*r;
	int32		t, z, c;
	uint8		cmp;

	if (fread(b,8,1,s)!=1)
		return -3;
	r = (const int32*)&b;
	if ((t=va_be32(r))<=0)
		return 0;
	cmp = *r==MK4CC('A','I','F','C');
	if (!cmp && *r!=MK4CC('A','I','F','F'))
		return 0;
	for(t+=4; (t-=8)>0; t-=z)
	{
		int32		a = 0;

		if (fread(b,8,1,s)!=1)
			return -3;
		if ((z=be32_native(((const int32*)b)[1]))<0)
			return -1;
		switch(*(const int32*)b)
		{
		case MK4CC('C','O','M','M'):
			if (z<18)
				return -1;
			if ((a=fread(b,1,40,s))<18)
				return -3;
			if (*b|b[6] || !b[7]) /* bad channel count, depth */
				return -1;
			/* chnCnt.w/frameCnt.l/depth.w/rate.x80 */
			f->chanCount = b[1];
			r = (const int32*)(b+2);
			f->frameCount = va_be32(r);
			/* more than one COMM, bad frame count or rate */
			if (f->depth || !parseExt80Rate(b+8,f))
				return -1;
			f->depth = b[7];
			c = 0;
			if (cmp && z>18) /* codec.l cname.pstr */
				c = *(const int32*)(b+18);
			if ((c=sndFixCmpType(c,f))!=0)
				return c;
			f->blockSize = f->packSize*f->chanCount;
			goto next;
		case MK4CC('S','S','N','D'):
			if (!f->blockSize || o>0) /* no format or more than one SSND */
				return -1;
			if (z<8 || (a=fread(b,1,8,s))<8)
				return -1;
			/* data offset */
			r = (const int32*)b, o = va_be32(r), c = va_be32(r);
			if (o<0 || c<0) /* QT returns -208 on bad offsets */
				return -1;
			/* I/O block size (for alignment) ignored (QT does) */
			c = (z-o+f->blockSize-1)/f->blockSize;
			o+= ftell(s);
			/* may have to adjust frame count if too short */
			if (c<f->frameCount)
				f->frameCount = c; //return -1;
		default:
		next:	++z, z&=~1; /* must align! (AIFF-C.9.26.91.pdf); 2014.01.27 */
			fseek(s,z-a,SEEK_CUR);
		}
	}
	if (t<-1) /* -1 valid in regards to alignment */
		return -3;
	sndFixCodecFrame(f);
	//mulu64u32(f->totalCount,f->frameCount,f->frameSize);
	return o;
}

long int parseFileRIFF(FILE *s,
	sndFormat *f,
	int is64b)
{ /* RIFF: WAVE (32-bit) */
	uint8		b[32];
	int64		t, dz, z;
	long int	o =-1;
	size_t		a;
	int32		c;

	if (fread(b,8,1,s)!=1)
		return -3;
	if ((a=le32_native(*(const int32*)b))==0)
		return 0;
	int64setl(t,a);
	if (((const int32*)b)[1]!=MK4CC('W','A','V','E'))
		return 0;
	int64qset(z,0); int64qset(dz,0); int64qset(f->totalCount,0); a = 0;
	if (is64b)
	{ /* 64-bit - need signature; the first tag */
		if ((a=fread(b,1,32,s))<32)
			return -3;
		if (*(const int32*)b!=MK4CC('d','s','6','4')
				|| (c=le32_native(((const int32*)b)[1]))<16)
			return -1;
		int64setl(z,c);
		int64sethl(t,le32_native(((const int32*)b)[3]),le32_native(((const int32*)b)[2]));
		int64sethl(dz,le32_native(((const int32*)b)[5]),le32_native(((const int32*)b)[4]));
		int64sethl(f->totalCount,le32_native(((const int32*)b)[7]),
			le32_native(((const int32*)b)[6]));
	}
	for(int64addu32(t,4); int64subu32(t,8),int64sub(t,z),int64p(t)>0;)
	{
	#if LONG_LONG
		fpos_t		pos; /* POCC 5+ makes has this 64-bit */

		fgetpos(s,&pos);
		pos.off+= z-a;
		fsetpos(s,&pos);
	#else
		fseek(s,int64tol(z)-a,SEEK_CUR);
	#endif
		if ((a=fread(b,1,8,s))!=8)
		{
			if (a-1u<4u)
				goto end; /* some WAVEs have extra benign bytes */
			return -3;
		}
		a = le32_native(((const int32*)b)[1]);
		int64setl(z,a);
		a = 0;
		switch(*(const int32*)b)
		{
		case MK4CC('f','m','t',' '):
			if (!int64gehl(z,0,16))
				return -1;
			if ((a=fread(b,1,20,s))<16)
				return -3;
			if ((b[11]&0x80)|b[3]|b[7]|b[15] || !b[2])
				return -1; /* bad stored frame size, chan-count, rate or depth */
			/* fmt.w, chnCnt.w, rate.l, sfsize.l, balgn.w, depth.w */
			f->chanCount = b[2];
			f->rate = le32_native(((const int32*)b)[1]);
			/* more than one 'fmt ', bad frame count or rate */
			if (f->depth) // || (f->blockSize=le32_native(((const int32*)b)[2]))<0)
				return -1;
			f->blockSize = b[12]|b[13]<<8, f->depth = b[14];
			f->packSize = f->blockSize/f->chanCount;
			if (int64gthl(z,0,19))
				f->frameSize = b[18]|b[19]<<8;
			if ((c=sndFixCmpType(MK4CC('m','s',b[1],*b),f))!=0)
				return c;
			continue;
		case MK4CC('d','a','t','a'):
			if (!f->blockSize || o>0) /* no format or more than one 'data' */
				return -1;
			if (is64b)
				z = dz;
			else
				dz = z; /* size as a last resort (no 'fact' or 'ds64') */
			o = ftell(s);
			continue;
		case MK4CC('f','a','c','t'):
			if (!int64z(f->totalCount))
				return -1; /* multiple instances */
			if ((a=fread(b,1,4,s))<4)
				return -3;
			int64sethl(f->totalCount,0,le32_native(*(const int32*)b));
		default:
			; /* anything about alignment? */
		}
	}
	if (int64n(t)<0)
		return -3;
end:	sndFixCodecFrame(f);
	if (int64z(dz) || f->frameSize<=0 || !f->blockSize || !f->mSampleSize || !f->chanCount)
		return -1;
	int64addu32(dz,f->blockSize-1);
	int64divu32(dz,f->blockSize);
	if (!int64z(f->totalCount))
	{
		int64mulu32(dz,f->frameSize);
		t = f->totalCount; int64mulu32(t,f->mSampleSize*f->chanCount);
		if (int64gt(dz,t))
			dz = t;
		int64addu32(dz,f->frameSize-1);
		int64divu32(dz,f->frameSize);
	}
	f->frameCount = int64tou(dz);
	return o;
}

long int readAudioHeader(FILE *s,
	sndFormat *f)
{
	int32		b;
	long int	o;

	if (!s || !f)
		return -1;
	if (fread(&b,sizeof(b),1,s)!=1)
		return -3;
	memset(f,0,sizeof(*f));
	o = 0;
	if (b==MK4CC('F','O','R','M'))
		o = parseFileFORM(s,f);
	else if (b==MK4CC('R','I','F','F')
			|| b==MK4CC('R','F','6','4'))
		o = parseFileRIFF(s,f,b^MK4CC('R','I','F','F'));
	else /* riff: 64-bit WAVE (16B marker, 8B aligned) */
		return 0;
	if (o<=0)
		return o;
	if (!f->chanCount || !f->depth || !f->rate || !o)
		return -1;
	return o;
}
