/*
 sndmgr.c
  sound channel player, command processor

  Created: 2014.03.21 02:09 EDT
  Revised: 2014.12.24
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.21
  separated from smixer.
2014.03.22
  finished working prototype; gaps (mixer); not everything's been tested.
2014.03.23
  success!!  multi-threaded OR single (manual sndMgrProcess calls).
2014.03.24
  seamless interpolation, end-sample silence (instead of zeroes).
2014.04.04
  pause now repeats current sample at ->pos (instead of zeroes).
deactivation can still be performed by setting bit #1 of actFlag.
2014.08.09
  sndFilePlay callback now uses double-buffering, as it should have.
2014.08.14
  sndPlayFileStop clears userInfo first (as it's supposed to).
not doing so promotes crashes, hangs.
2014.12.05
  using the delayed buffering approach (2ms timer) used in mpegdecmac.c
to stop all the sndFilePlay callback crashing.
*/

#include <stdio.h>

#include <enviro64.h>
#include "../ftrig.h"
#include "snddec.h"

#if defined(WIN32) || defined(WIN64)
 #include <windows.h>
 #include <mmsystem.h>
 #include <process.h>	//threads

 #define CloseSemaphore		CloseHandle
 #define WaitForSemaphore	WaitForSingleObject
 /* these call for win_ce_semaphore.c ... CleanUp has bugs with its mask conditionals */
 #ifdef _WIN32_WCE
  #if _WIN32_WCE<300
   #define CreateSemaphore	CreateSemaphoreCE
   #define ReleaseSemaphore	ReleaseSemaphoreCE
   #undef CloseSemaphore
   #undef WaitForSemaphore
   #define CloseSemaphore	CloseSynchHandle
   #define WaitForSemaphore	WaitForSemaphoreCE
  #endif
 #endif
 #define MgrMutexLock(m)	EnterCriticalSection(&(m)->criticalSection)
 #define MgrMutexUnlock(m)	LeaveCriticalSection(&(m)->criticalSection)
#endif

/* ---- constants */

/* the thread loop is so lean it ends up most often calling Sleep(1) */
/* the Semphore method can also break up easily */
#define NO_SEMAPHORE
#define QUEUED_WAVE
//#define WAVE_DUMP

#ifdef __MT__
 #undef WAVE_DUMP
#endif
#ifdef WAVE_DUMP
 #include <stdio.h>
static FILE		*fwout;
static int32		_lastFmtSize;
static sndFormat	format;
 #define lastFmtSize		&_lastFmtSize
#else
 #define lastFmtSize		NULL
#endif

/*
 The Apple Sound Chip and EASC have a 1024-stereo sample buffer,
but has also a 'half-full' indicator - 512 samples, that is, the ideal.
 However, the Sound Manager buffers 4096 samples in loading; this weakens
loading latency, but makes stereo DVI_IMA possible (2040*2).  But that's
not the potential max. here.  64-bit samples (doubles) can trump that;
maxRead(512)*8 makes 4096.  4096*stereo 16-bit(4) makes 16384.
*/
/* 2014.07.27: 4096 doesn't seem to work w. stereo 8-bit (out-of-bounds dcmp writes) */
#define chBufCount		(MAX(sndEndFrameCount,8192)+1)
#if defined(__MT__) && defined(SNDMGR_PRECISION)
 #define sndEndFrameCount	1024	/* can afford smaller at above-normal... */
#else
 #define sndEndFrameCount	2048
#endif
#define sndBufferCount		4	/* Semaphore method with small buffer needs 3+ */
#define defaultVolume		0x1000100
/* not meant to handle rates above 96000 */
#define defaultOutRate		44100

/* ---- manager structure */

typedef struct sndMgrData
{
	sndChannel	*sndChannels; /* a 'next' chain, until NULL */
	uint32		rflag;
 #ifdef WINUSERAPI
  #if defined(__MT__) && !defined(NO_SEMAPHORE)
	HANDLE		hSemaphore;
  #endif
	HWAVEOUT	hWAudio;
	WAVEHDR		hWHeader[sndBufferCount];

	CRITICAL_SECTION	criticalSection;
 #endif
	int32		out[sndBufferCount][sndEndFrameCount*2];
} sndMgrData;

/* ---- code */

static inline void updateChMixerAmp(sndChannel *ch)
{
	int32		v = ch->LRVolume;
	uint8		flag = ((sndMgrData*)ch->private)->rflag>>24&ch->initFlag;

	if (!(flag&sndEnableLeft))
		v&= 0xFFFF;
	if (!(flag&sndEnableRight))
		v&=~0xFFFF;
	ch->sndFrame.LRAmp = v;
}

/* commands */

static inline int enQueue(sndChannel *ch,
	const sndCommand *cmd,
	int8 reverse)
{
	sndMgrData	*mgr;
	int		x, r = 0;

	if ((x=ch->qLength)<=0 || !cmd)
		return -1;
	if ((mgr=ch->private)!=0)
		MgrMutexLock(mgr);
	if (ch->qTail+0u>=x)
		ch->qTail = 0;
	if (ch->qHead>0)
		x = ch->qHead;
	if (ch->qTail==--x)
		r =-203; /* full */
	else if (reverse) /* pre-incr. from head */
		ch->qHead = x, ch->queue[x] = *cmd;
	else /* post-incr. from tail */
		ch->queue[x=ch->qTail] = *cmd, ch->qTail = ++x>=ch->qLength ? 0:x;
	if (mgr)
		MgrMutexUnlock(mgr);
	return r;
}

static inline int deQueue(sndChannel *ch,
	sndCommand *cmd) /* must be valid */
{
	int		z, x, r = 0;

	if ((z=ch->qLength)<=0)
		return -1;
	/* post increment from head */
	/* first, correct head/tail if out of bounds */
	if ((x=ch->qHead)+0u>=z)
		ch->qHead = 0;
	if (ch->qTail+0u>=z)
		ch->qTail = 0;
	if (x!=ch->qTail)
		*cmd = ch->queue[x], ch->qHead = ++x>=z ? 0:x, r = 1;
	return r;
}

int sndChannelBusy(sndChannel *ch)
{
	return ch->qHead!=ch->qTail || ch->waitCnt
		|| ch->sndFrame.cmpSamples
		|| (ch->sndFrame.dcmpSize && ch->sndFrame.pos);
}

static int processCommands(sndChannel *ch,
	int c)
{
//	sndMgrData	*mgr;
	sndCommand	cmd;
	int32		r;

//	if ((mgr=ch->private)!=0)
//		MgrMutexLock(mgr);
	do
	{
		if ((r=deQueue(ch,&cmd))<=0)
			goto end;
		switch(cmd.cmd)
		{
		case sndAllocCmd:
			if (!(ch->actFlag&0x80) && !ch->sndFrame.dcmpBuf)
				ch->sndFrame.dcmpBuf = calloc(sizeof(int16)*2,chBufCount);
			break;
		case sndFreeCmd:
			if (!(ch->actFlag&0x80) && ch->sndFrame.dcmpBuf)
				{free(ch->sndFrame.dcmpBuf); ch->sndFrame.dcmpBuf = NULL;}
			break;
		case sndQuietCmd:
			ch->sndFrame.pos = ch->sndFrame.dcmpSize = 0;
			ch->sndFrame.format.frameCount = 0;
			ch->sndFrame.cmpSamples = NULL;
			ch->sndFrame.mixerRtn = NULL;
		case sndNullCmd:
			break;
		case sndFlushCmd:
			ch->sndFrame.cmpSamples = NULL;
			ch->sndFrame.mixerRtn = NULL;
			ch->waitCnt = 0;
			ch->qHead = ch->qTail = 0;
			r = 0; /* nothing to process now */
			goto end;
		case sndReInitCmd:
			ch->initFlag = (uint8)cmd.param2&~sndInitNoNearDropSample;
			if (ch->initFlag&sndEnableRight) /* initial bit */
				ch->initFlag|= sndInitStereo; /* mono-left or right */
			ch->initFlag^= sndInitPanMask;
			goto upmix;
		case sndFormatCmd:
			if (!cmd.param2)
				goto inv;
			ch->sndFrame.cmpSamples = NULL;
			memcpy(&ch->sndFrame.format,cmd.param2,sizeof(ch->sndFrame.format));
			ch->sndFrame.format.frameCount = 0;
			if (sndFixCmpType(ch->sndFrame.format.codec,&ch->sndFrame.format)<0)
				{ch->sndFrame.format.codec = 0; goto inv;}
			r = 0;
			ch->sndFrame.format.maxRead =
				(ch->sndFrame.format.type&15)>CTYPE_NPCM ? 1:512;
			if (!ch->sndFrame.format.blockSize)
				sndFixCodecFrame(&ch->sndFrame.format);
			goto fmt;
		case sndRawCmd:
			if (!cmd.param1)
				break;
			if (!cmd.param2 || !ch->sndFrame.format.rate)
				goto inv;
			ch->sndFrame.format.frameCount = cmd.param1;
			ch->sndFrame.cmpSamples = cmd.param2;
			r = 1;
			goto end;
		case sndBufferCmd:
			if ((r=parseBufferCmdHeader(cmd.param2,&ch->sndFrame.format,NULL))<0)
				goto end;
			ch->sndFrame.cmpSamples = r+(uint8*)cmd.param2;
			r = 1;
			/* initialize decompression variables */
			decompressSndFrame(ch->sndFrame.dcmpBuf,NULL,&ch->sndFrame.format,0);
		fmt:	ch->sndFrame.mixerRtn = NULL;
			goto end;
		case sndRateMultiplierCmd:
			if (!ch->sndFrame.dcmpSize)
				ch->sndFrame.cnvFrc = (uint16)cmd.param1;
			ch->sndFrame.rateMul = (ufixed)cmd.param2;
			ch->sndFrame.mixerRtn = NULL;
			break;
		case sndWaitCmd:
			ch->waitCnt+= ufixmuldiv((uint32)cmd.param2,
				((sndMgrData*)ch->private)->rflag&0xFFFFFF,2000);
			r = 1;
			goto end;
		case sndPauseCmd:
			ch->actFlag|= 1;
			r = 1;
			goto end;
		case sndResumeCmd:
			ch->actFlag&=~1;
			break;
		case sndCallBackCmd:
			if (cmd.param2)
				((sndCallBack)cmd.param2)(ch,&cmd);
			break;
		case sndRLVolumeCmd:
			ch->LRVolume = (uint16)((uint32)cmd.param2>>16)|(uint32)cmd.param2<<16;
			goto upmix;
		case sndLRVolumeCmd:
			ch->LRVolume = (uint32)cmd.param2;
		upmix:	updateChMixerAmp(ch);
			break;
		inv:
		default:
			r =-50;
			goto end;
		}
	} while(!c || --c);
end:	//if (mgr)
	//	MgrMutexUnlock(mgr);
	return r;
}

int sndDoCommand(sndChannel *ch,
	const sndCommand *cmd,
	int8 immediate)
{
	int		err;

	if (!ch || !cmd)
		return -1;
	if ((err=enQueue(ch,cmd,immediate))==0 && immediate)
		err = processCommands(ch,1);
	return err;
}

/* fill channel packet using sndMixSamples;
  if still unable to fill packet, advance queue;
  all else fails, silence
*/
static inline void sndChRender(int32 *out,
	sndChannel *ch)
{
	uint32		rflag, rem;

	if (ch->actFlag&2)
		return;
	rem = sndEndFrameCount;
	ch->actFlag|= 0x80;
	if (ch->actFlag&1)
	{
		rflag = ((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF);
		goto fill;
	}
	goto start;
	do
	{
		if (ch->sndFrame.format.frameCount<=0 && !ch->sndFrame.cmpSamples
				&& !ch->sndFrame.pos)
		{
			if (!processCommands(ch,0))
				ch->waitCnt+= rem; /* nothing left */
			/* must check for wait period(s) */
		start:	/* because of reInit, rflag must be recalculated */
			rflag = ((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF);
			if (ch->waitCnt)
			{
				if (ch->waitCnt>=rem)
				{
					ch->waitCnt-= rem;
				fill:	sndMixFill(rem,rflag,
						out+2*(sndEndFrameCount-rem),
						&ch->sndFrame);
					break;
				}
				sndMixFill(ch->waitCnt,rflag,
					out+2*(sndEndFrameCount-rem),
					&ch->sndFrame);
				rem-= ch->waitCnt, ch->waitCnt = 0;
			}
			continue;
		}
		rem = sndMixSamples(out+2*(sndEndFrameCount-rem),&ch->sndFrame,
			rflag,sndEndFrameCount,rem);
	} while(rem>0);
	ch->actFlag&=~0x80;
}

/* channel creation/disposal */

sndChannel *sndNewChannel(sndManager *smgr,
	uint8 initFlags)
{
	sndMgrData	*mgr;
	sndChannel	**p, *ch;

	if ((mgr=(sndMgrData*)smgr)==NULL)
		return NULL;
	for(p=&mgr->sndChannels; *p;)
		if (*(p=&(**p).next)==mgr->sndChannels)
			return NULL; /* recursion! */
	if ((ch=calloc(sizeof(sndChannel),1))==NULL)
		return NULL;
	if ((ch->sndFrame.dcmpBuf=calloc(sizeof(int16)*2,chBufCount))==NULL)
		{free(ch); return NULL;}
	ch->initFlag = initFlags|sndInitPanMask;
	ch->private = mgr;
	ch->qLength = GET_ARRAY_COUNT(ch->queue);
	ch->sndFrame.rateMul = 0x10000;
	ch->LRVolume = defaultVolume;
	updateChMixerAmp(ch);
	*p = ch;
	return ch;
}

int sndDisposeChannel(sndChannel *ch)
{
	sndMgrData	*mgr;
	sndChannel	**p;

	if (!ch || (mgr=ch->private)==NULL)
		return -1;
	for(p=&mgr->sndChannels; *p;)
		if (*p==ch)
			goto found;
		else if (*(p=&(**p).next)==mgr->sndChannels)
			return -50; /* recursion! */
	return -205; //-7;
found:	*p = ch->next;
	MgrMutexLock(mgr);
	ch->actFlag|= 2;
	if (ch->sndFrame.dcmpBuf)
		{free(ch->sndFrame.dcmpBuf); ch->sndFrame.dcmpBuf = NULL;}
	MgrMutexUnlock(mgr);
	ch->private = NULL;
	free(ch);
	return 0;
}

/* sound file playback */

#ifdef __MT__
 #include "mevent.h"
extern timerThread	*global_timer_thread;
#endif

typedef struct playFileData
{
	FILE		*in;
	sndFormat	format;
	int32		mxc, bz;
#ifdef __MT__
	timerEntry	*task;
	sndChannel	*chan;
#endif
	uint8		*wh, *alt;
	uint8		buf[1];
} playFileData;

static int loader(playFileData *pfd)
{
	sndCommand	cmd;
	uint32		act;

	if (pfd->wh)
	{
		cmd.cmd = sndRawCmd;
		cmd.param1 = pfd->bz;
		pfd->wh = pfd->wh==pfd->buf ? pfd->alt:pfd->buf; /* swap buffer */
		cmd.param2 = pfd->wh;
		sndDoCommand(pfd->chan,&cmd,0);
	} else
		pfd->wh = pfd->buf;
	if (!pfd->format.frameCount || (act=fread(pfd->wh,pfd->format.blockSize,
			MIN(pfd->mxc,pfd->format.frameCount),pfd->in))==0)
	{
		pfd->wh = NULL;
		return -1;
	}
	pfd->bz = act;
	pfd->format.frameCount-= act;
	return 0;
}

#ifdef __MT__
static void autoloader(timerEntry *param)
{
	playFileData	*pfd;

	if ((pfd=param->userInfo)==NULL)
		return;
	if (!pfd->chan)
		return;
	if (!pfd->chan->userInfo)
		return;
	loader(pfd);
}
#endif

static void playFileCallBack(sndChannel *ch,
	sndCommand *_cmd)
{
#pragma unused(_cmd)
	sndCommand	cmd;
	playFileData	*pfd;

	if ((pfd=ch->userInfo)==NULL || !pfd->format.blockSize)
		return;
	cmd.cmd = sndCallBackCmd;
	cmd.param1 = 0;
	cmd.param2 = playFileCallBack;
#ifdef __MT__
	PrimeTimer(global_timer_thread,pfd->task,2000);
#else
	loader(pfd);
#endif
	sndDoCommand(ch,&cmd,0);
}

int sndPlayFile(FILE *in,
	sndChannel *ch,
	const sndFormat *f,
	int32 k)
{
	playFileData	*pfd;
	sndCommand	cmd;
	int32		z, v;

	if (!in || !f || !ch || !f->blockSize || ch->userInfo)
		return -1;
	/* 32KB is the largest power of two that will fit in a uint16 */
#ifdef __MT__
	for(z=f->maxRead*f->blockSize; z<16384; z+=z)
#else
	for(z=f->maxRead*f->blockSize; z<32768; z+=z)
#endif
		;
	if (z<=0)
		return -1;
	v = f->maxRead*f->frameSize>>1;
	v = MAX(z,v);
	if ((pfd=ch->userInfo=calloc(1,v+sizeof(playFileData)))==NULL)
		return -2;
	/* alternate must be in a separate allocation
	  (Access Exception can occur if any pages cross writing when sndMgr reads)
	*/
	if ((pfd->alt=calloc(1,v))==NULL)
	{
		free(pfd);
		return -2;
	}
#ifdef __MT__
	setvbuf(in,NULL,_IOFBF,32768);
	pfd->chan = ch;
	if ((pfd->task=InstallTimer(global_timer_thread,autoloader))!=NULL)
		pfd->task->userInfo = pfd;
#endif
	pfd->in = in;
	pfd->format = *f;
	pfd->mxc = z/f->blockSize;
	/* skip any leading frames now */
	if ((v=k)<0)
		k =~k;
	for(; k; --k)
		if (decompressSndFrame(pfd->buf,pfd->buf,&pfd->format,
				readSndFrame(pfd->buf,in,&pfd->format))-pfd->buf<=0)
			return 0; /* ended before we even started */
	if (v<0)
	{
		cmd.cmd = sndPauseCmd;
		sndDoCommand(ch,&cmd,1);
	}
	cmd.cmd = sndFlushCmd;
	sndDoCommand(ch,&cmd,1);
	cmd.cmd = sndFormatCmd;
	cmd.param2 = &pfd->format;
	sndDoCommand(ch,&cmd,1);
#ifdef __MT__
	if (loader(pfd))
		return 0;
	if (pfd->wh)
	{
		cmd.cmd = sndRawCmd;
		cmd.param1 = pfd->bz;
		cmd.param2 = pfd->wh;
		sndDoCommand(pfd->chan,&cmd,0);
	}
#endif
	playFileCallBack(ch,NULL);
	return 1;
}

int sndPlayFileStop(sndChannel *ch)
{
	sndCommand	cmd;
	playFileData	*pfd;
	int		r;

	if (!ch || !ch->private)
		return -1;
	 if ((pfd=ch->userInfo)==NULL)
		return 0;
	r = pfd->format.frameCount>0;
	ch->userInfo = NULL;
	pfd->format.frameCount = 0;
	cmd.cmd = sndFlushCmd;
	sndDoCommand(ch,&cmd,1);
	free(pfd->alt);
	free(pfd);
	return r;
}

/* sound manager top level */

int sndMgrMute(sndManager *smgr,
	uint8 lrdisable)
{
	sndMgrData	*mgr;
	sndChannel	*ch;

	if ((mgr=(sndMgrData*)smgr)==NULL)
		return -1;
	MgrMutexLock(mgr);
	mgr->rflag&=~(sndInitPanMask<<24);
	mgr->rflag|= (lrdisable&=sndInitPanMask)<<24;
	if ((ch=mgr->sndChannels)!=NULL)
		do
			updateChMixerAmp(ch);
		while((ch=ch->next)!=NULL);
	MgrMutexUnlock(mgr);
	return lrdisable;
}

static inline void sndMgrRender(sndMgrData *mgr,
	int i)
{
	sndChannel	*ch = mgr->sndChannels;
	int32		*b = mgr->out[i];

	memset(b,0,sizeof(*mgr->out));
	do
	{
		if (ch->actFlag&2)
			continue;
		sndChRender(b,ch);
	} while((ch=ch->next)!=NULL);
	s32to16(sndEndFrameCount,mgr->rflag,b);
}

#ifdef WINUSERAPI
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
/* this should work on Win32 and up */
/* ReleaseSemaphore will not work from a callback on older systems */
static void CALLBACK sndWaitSync(HWAVEOUT hwa,
	UINT msg,
	DWORD_PTR dwu,
	DWORD dw1,
	DWORD dw2)
{
	if (msg!=WOM_DONE)
		return;
	ReleaseSemaphore(((sndMgrData*)dwu)->hSemaphore,1,NULL);
}
 #elif defined(QUEUED_WAVE)
static void CALLBACK sndWaitSync(HWAVEOUT hwa,
	UINT msg,
	DWORD_PTR dwu,
	DWORD dw1,
	DWORD dw2)
{
	sndMgrData	*mgr = (sndMgrData*)dwu;

	if (msg!=WOM_DONE)
		return;
	MgrMutexLock(mgr);
	--*(int32*)&mgr->hWHeader->dwUser;
	MgrMutexUnlock(mgr);
}
 #endif

HWAVEOUT sndMgrGetHWO(sndManager *smgr)
{
	sndMgrData	*mgr;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0xFFFFFF))
		return NULL;
	return mgr->hWAudio;
}

static int sndMgrProcess(sndMgrData *mgr)
{
	WAVEHDR		*hdr;
	int		x;

	if (!mgr->sndChannels)
		return 0;
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
	x = (int32)mgr->hWHeader[1].dwUser;
	hdr = mgr->hWHeader+x;
	MgrMutexLock(mgr);
	sndMgrRender(mgr,x);
	MgrMutexUnlock(mgr);
	if (mgr->hWHeader->dwUser)
		WaitForSemaphore(mgr->hSemaphore,INFINITE);
	waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr));
	*(uint32*)&mgr->hWHeader[1].dwUser = (x+1u)%sndBufferCount;
	mgr->hWHeader->dwUser = (DWORD_PTR)1;
	return 1;
 #elif defined(QUEUED_WAVE)
	if ((uint32)mgr->hWHeader->dwUser>=sndBufferCount)
		return 1;
	x = (int32)mgr->hWHeader[1].dwUser;
	hdr = mgr->hWHeader+x;
	MgrMutexLock(mgr);
	++*(int32*)&mgr->hWHeader->dwUser;
	sndMgrRender(mgr,x);
	MgrMutexUnlock(mgr);
	*(uint32*)&mgr->hWHeader[1].dwUser = (x+1u)%sndBufferCount;
  #ifdef WAVE_DUMP
	fwrite(hdr->lpData,hdr->dwBufferLength,1,fwout);
  #endif
	if (waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr))!=MMSYSERR_NOERROR)
		hdr->dwFlags|= WHDR_DONE;
	return 2;
 #else
	{ /* basic loader (no callback) */
		int		i = 0;

		x = 2, hdr = mgr->hWHeader;
		do
			if (hdr->dwFlags&WHDR_DONE)
			{
				MgrMutexLock(mgr);
				hdr->dwFlags&=~WHDR_DONE;
				sndMgrRender(mgr,i);
				MgrMutexUnlock(mgr);
			#ifdef WAVE_DUMP
				fwrite(hdr->lpData,hdr->dwBufferLength,1,fwout);
			#endif
				if (waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr))!=MMSYSERR_NOERROR)
					hdr->dwFlags|= WHDR_DONE;
				x = 1;
			}
		while(++hdr,++i<sndBufferCount);
	}
	return x;
 #endif
}

int sndMgrWait(sndManager *smgr)
{
	sndMgrData	*mgr;
	uint32		rate;

	if ((mgr=(sndMgrData*)smgr)==NULL || (rate=mgr->rflag&0xFFFFFF)==0)
		return -1;
	if (!(mgr->rflag&0x80000000u))// || !mgr->sndChannels)
		return 0;
	Sleep(1+sndEndFrameCount*1000u/rate);
	return 1;
}

 #ifdef __MT__
static void sndMgrThread(void *smgr)
{
	while(((sndMgrData*)smgr)->rflag&0x80000000u)
	#ifdef NO_SEMAPHORE
		Sleep(((sndMgrProcess(smgr)-1)&3)+1);
	#else
		if (!sndMgrProcess(smgr))
			Sleep(2);
	#endif
	_endthread();
}

void sndMgrProcWait(sndManager *smgr)
{
	if (smgr)
		Sleep(2);
}
 #else
void sndMgrProcWait(sndManager *smgr)
{
	if (smgr)
		Sleep(2-!!sndMgrProcess((sndMgrData*)smgr));
}
 #endif

int sndMgrClose(sndManager *smgr)
{
	sndMgrData	*mgr;
	WAVEHDR		*hdr;
	int		i, r = 0;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u))
		return -1;
	MgrMutexLock(mgr);
	mgr->rflag = 0;
	MgrMutexUnlock(mgr);
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
	if (mgr->hSemaphore)
	{
		HANDLE		sem = mgr->hSemaphore;

		ReleaseSemaphore(sem,1,NULL);
		Sleep(8);
		mgr->hSemaphore = NULL;
		CloseSemaphore(sem);
		++r;
	}
 #endif
	for(i=0,hdr=mgr->hWHeader; i<sndBufferCount; ++i,++hdr)
		while(hdr->dwFlags&WHDR_PREPARED
				&& waveOutUnprepareHeader(mgr->hWAudio,hdr,sizeof(*hdr))
				==WAVERR_STILLPLAYING)
			Sleep(2);
	if (mgr->hWAudio)
	{
		while(waveOutClose(mgr->hWAudio)==WAVERR_STILLPLAYING)
			Sleep(2);
		mgr->hWAudio = NULL, ++r;
	}
 #ifdef WAVE_DUMP
	{
		const uint8	*h;
		int32		z = ftell(fwout)-_lastFmtSize;

		fseek(fwout,0,SEEK_SET);
		h = setupWAVEHeader(&format,&z);
		fwrite(h,z,1,fwout);
		fclose(fwout);
	}
 #endif
	MgrMutexLock(mgr);
	while(mgr->sndChannels)
		sndDisposeChannel(mgr->sndChannels);
	MgrMutexUnlock(mgr);
	DeleteCriticalSection(&mgr->criticalSection);
	free(mgr);
	return r;
}

int sndMgrMixRate(sndManager *smgr,
	int rate)
{
	sndMgrData	*mgr;
	int		r;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u) || rate<=0)
		return -1;
	rate = PIN(rate,1000,100000);
	MgrMutexLock(mgr);
	r = mgr->rflag;
	mgr->rflag = (r&0xFF000000u)|rate;
	MgrMutexUnlock(mgr);

	return r;
}

sndManager *sndMgrOpen(int outRate,
	int stereo,
	int *err)
{
	WAVEFORMATEX	wfmt;
 #ifndef WAVE_DUMP
	sndFormat	format;
 #endif
	const uint8	*h;
	sndMgrData	*mgr = NULL;
	WAVEHDR		*hdr;
	int32		*d;
	MMRESULT	mmr;
	int		i, r =-1;

	memset(&format,0,sizeof(format));
	format.chanCount = stereo ? 2:1;
	format.frameCount = sndEndFrameCount;
	format.rate = outRate<=0 ? defaultOutRate:PIN(outRate,1000,100000);
	sndFixCmpType(CODEC_PCM_LE_S16,&format);
	if ((h=setupWAVEHeader(&format,lastFmtSize))==NULL)
		goto end;
	memcpy(&wfmt,h+20,sizeof(wfmt)); /* assuming native LE... */
	wfmt.cbSize = sizeof(WAVEFORMATEX); //0;
	/* allocate */
	if ((mgr=calloc(sizeof(sndMgrData),1))==NULL)
		{r =-2; goto end;}
	mgr->rflag = format.rate|(stereo ? -1<<24:~sndInitStereo<<24);
	if ((mmr=waveOutOpen(&mgr->hWAudio,WAVE_MAPPER,&wfmt,
 #if (defined(__MT__) && !defined(NO_SEMAPHORE)) || defined(QUEUED_WAVE)
			(DWORD_PTR)sndWaitSync,(DWORD_PTR)mgr,CALLBACK_FUNCTION
 #else
			0,(DWORD_PTR)mgr,CALLBACK_NULL
 #endif
			))!=MMSYSERR_NOERROR)
		goto close;
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
	if ((mgr->hSemaphore=CreateSemaphore(NULL,sndBufferCount-1,sndBufferCount,NULL))==NULL)
		goto close;
 #endif
	InitializeCriticalSection(&mgr->criticalSection);

	i = 0, hdr = mgr->hWHeader, d = *mgr->out;
	do
	{
		hdr->lpData = (LPSTR)d;
		hdr->dwBufferLength = sndEndFrameCount<<format.chanCount;
		mmr = waveOutPrepareHeader(mgr->hWAudio,hdr,sizeof(*hdr));
		hdr->dwFlags|= WHDR_DONE;
		if (mmr!=MMSYSERR_NOERROR)
		{
		close:	sndMgrClose((sndManager*)mgr);
			goto end;
		}
	} while(++hdr,d+=sndEndFrameCount*2,++i<sndBufferCount);
 #ifdef WAVE_DUMP
	if ((fwout=fopen("fwout.wav","wb"))!=NULL)
		fwrite(h,_lastFmtSize,1,fwout);
 #endif
 #ifdef __MT__
	if ((r=_beginthread(sndMgrThread,0,mgr))==-1)
		goto end;
	SetThreadPriority((HANDLE)r,
		outRate>=96000 ? THREAD_PRIORITY_HIGHEST:THREAD_PRIORITY_ABOVE_NORMAL);
 #endif
	r = 0;
end:	if (err)
		*err = r;
	return (sndManager*)mgr;
}

 #undef CreateSemaphore
 #undef ReleaseSemaphore
 #undef CloseSemaphore
 #undef WaitForSemaphore
#endif /* defined(WINUSERAPI) */
