/*
 sndmixer.c
  sample mixing code, stereo output

  Created: 2014.01.26 09:29 EST
  Revised: 2014.08.07
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.23
  interpolation end-sample problem solved: writes final sample to front.
2014.03.25
  mix routines no longer return remaining count;
all length clipping is done with sndMixSamples now.
  more 1.0 volume detection (for performance).
*/

#include <stdlib.h>
#include <string.h>
#include <enviro64.h>
#include "snddec.h"
#include "../ftrig.h"

/* ---- typedefs */

typedef struct lri16
{
	int16		l, r;
} lri16;

/* ---- macros */

#define monoMix(v,u,l,r)	((v)*(l)+(u)*(r))>>9

/* ---- code */

/*
 Process:
  1.	Decode (via decompressSndFrame).
  2.	Conform buffer to 16-bit samples.
  3.	Add to output, performing rate conversion & L/R amp. if necessary.
  4.    Repeat steps 1-3 until output block or cmpSamples are finished.

  Caller would make the mixing calls, and clip the result (s32to16).
*/

/* self-depth conversion routines */

/* to-larger overwrite: works backward from end */
#if defined(__POCC__) && envx86
__declspec(naked)
#endif
static uint32 wrChannelu8(const void *p,
	uint32 z,
	void *stream)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		mov	ecx,[esp+12]
		mov	esi,[esp+16]
		mov	esi,[esi]
		lea	edx,[esi+ecx*2]
		add	esi,ecx
		shr	ecx,1
		jnc	ao

		dec	esi
		mov	al,[esi]
		mov	ah,al
		xor	ax,0x8000
		sub	edx,2
		mov	[edx],ax
		test	ecx,ecx
		jz	eo

	ao:	sub	esi,2
		mov	ax,[esi]
		mov	al,ah
		shl	eax,16
		mov	al,[esi]
		mov	ah,al
		xor	eax,0x80008000
		sub	edx,4
		mov	[edx],eax
		loop	ao

	eo:	mov	eax,[esp+12]
		add	eax,eax
		pop	esi
		ret
	}
#else
	const uint8	*s;
	int16		*d = *(int16**)stream;
	int		c;
	unsigned	v;

	s = (uint8*)d+z, d+= z;
	if ((c=z>>1)!=0)
		do
		{
			v = *--s, *--d = (v|v<<8)^0x8000;
			v = *--s, *--d = (v|v<<8)^0x8000;
		} while(--c);
	if (z&1)
		v = s[-1], d[-1] = (v|v<<8)^0x8000;
	/* No need to update the stream; sndMixSamples only makes one call;
	  so does writeSndSamples; in-stack downsampling only occurs for 16-bit.
	*/
	return z+z;
#endif
}

/* ---- mixer routines */

/* rate conversion */

#define endRateCnv \
	m->pos=(const uint8*)s-m->dcmpBuf, m->cnvFrc=rf;
#define startRateCnv(t) \
	const t*s=p; \
	uint32 rt=((uint16*)&m->cnvRate)[!BIG_ENDIAN]; \
	uint32 rl=((uint16*)&m->cnvRate)[BIG_ENDIAN]; \
	uint32 rf=m->cnvFrc; \
	int32 v, lv=m->LRAmp;

/* interpolation: linear (standard) */

static __forceinline int lerp(int m,
	int a,
	int b)
{
	return a+((b-a)*m>>15);
}

/* stereo source */

#if envx86
extern DEF_MIX_RTN(rateCnvIntrp2);
extern DEF_MIX_RTN(rateCnvIntrp2m);
#else
static DEF_MIX_RTN(rateCnvIntrp2)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = rf>>1;
			*d+= lerp(v,s->l,s[1].l), d[1]+= lerp(v,s->r,s[1].r);
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = rf>>1;
			*d+= lerp(v,s->l,s[1].l)*lv>>8, d[1]+= lerp(v,s->r,s[1].r)*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrp2m)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = rf>>1;
			v = (lerp(v,s->l,s[1].l)+lerp(v,s->r,s[1].r))>>1;
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = rf>>1;
			v = monoMix(lerp(v,s->l,s[1].l),lerp(v,s->r,s[1].r),lv,rv);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}
#endif

/* mono source */

static DEF_MIX_RTN(rateCnvIntrp1)
{
	startRateCnv(int16)
	if (lv==0x1000100)
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v*lv>>8, d[1]+= v*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrp1m)
{
	startRateCnv(int16)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
		do
		{
			v = lerp(rf>>1,*s,s[1])*lv>>8;
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	endRateCnv
}

/* no interpolation */

/* stereo source */

static DEF_MIX_RTN(rateCnvNoInterp2)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			*d+= s->l, d[1]+= s->r;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		v = (int16)lv, lv>>= 16;
		do
		{
			*d+= s->l*lv>>8, d[1]+= s->r*v>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvNoInterp2m)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = (s->l+s->r)>>1, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = monoMix(s->l,s->r,lv,rv), *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

/* mono source */

static DEF_MIX_RTN(rateCnvNoInterp1)
{
	startRateCnv(int16)
	if (lv==0x1000100)
		do
		{
			v = *s, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = *s, *d+= v*lv>>8, d[1]+= v*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvNoInterp1m)
{
	startRateCnv(int16)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
		do
		{
			v = *s, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
		do
		{
			v = *s*lv>>8, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	endRateCnv
}

/* no rate conversion */

#define startOneCnv(h) \
	const int16 *s=p; \
	int32 v, lv=m->LRAmp; \
	m->pos+=c<<h;

static DEF_MIX_RTN(mixNoRateCnv2)
{
	startOneCnv(2)
	if (lv==0x1000100)
	{
		if (c&1)
			*d+= *s, d[1]+= s[1], s+= 2, d+= 2;
		if (c>>=1)
			do
				*d+= *s, d[1]+= s[1], d[2]+= s[2], d[3]+= s[3];
			while(s+=4,d+=4,--c);
	} else if (lv)
	{
		v = (int16)lv, lv>>= 16;
		do
			*d+= *s*lv>>8, d[1]+= s[1]*v>>8;
		while(s+=2,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv2m)
{
	startOneCnv(2)
	if (lv==0x1000100)
	{
		if (c&1)
			v = (*s+s[1])>>1, *d+= v, d[1]+= v, s+= 2, d+= 2;
		if (c>>=1)
			do
			{
				v = (*s+s[1])>>1, *d+= v, d[1]+= v;
				v = (s[2]+s[3])>>1, d[2]+= v, d[3]+= v;
			} while(s+=4,d+=4,--c);
	} else if (lv)
	{
		int32		rv = (int16)lv;

		lv>>= 16;
		do
			v = monoMix(*s,s[1],lv,rv), *d+= v, d[1]+= v;
		while(s+=2,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv1)
{
	startOneCnv(1)
	if (lv==0x1000100)
	{
		if (c&1)
			v = *s, *d+= v, d[1]+= v, ++s, d+= 2;
		if (c>>=1)
			do
			{
				v = *s, *d+= v, d[1]+= v;
				v = s[1], d[2]+= v, d[3]+= v;
			} while(s+=2,d+=4,--c);
	} else if (lv)
	{
		int32		rv = (int16)lv;

		lv>>= 16;
		do
			v = *s, *d+= v*lv>>8, d[1]+= v*rv>>8;
		while(++s,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv1m)
{
	startOneCnv(1)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
	{
		if (c&1)
			v = *s, *d+= v, d[1]+= v, ++s, d+= 2;
		if (c>>=1)
			do
			{
				v = *s, *d+= v, d[1]+= v;
				v = s[1], d[2]+= v, d[3]+= v;
			} while(s+=2,d+=4,--c);
	} else if (lv)
		do
			v = *s*lv>>8, *d+= v, d[1]+= v;
		while(++s,d+=2,--c);
}

/* the decompression and mixer routine manager */

int sndMixSamples(int32 *d,
	sndMixerFrame *m,
	uint32 rflag,
	int req,
	int rem)
{
	uint32		z;
	int		st;

	if (!m->dcmpBuf || !m->format.blockSize)
		return -1;
	st = (m->format.chanCount&127)>1;
	if (!m->pos)
	{ /* decompression phase */
		uint8		*p, *s = m->dcmpSize+m->dcmpBuf, *a = s;
		const uint8	*e;

		for(e=a+(req<<(m->format.chanCount&127)); a<e; m->cmpSamples+=z)
		{
			z = m->format.blockSize*MIN(m->format.frameCount,
				m->format.maxRead);
			if ((p=decompressSndFrame(a,m->cmpSamples,&m->format,z))<=a)
				{m->cmpSamples = NULL; break;}
			a = p;
		}
		/* sndwrite_null if already in native s16 format */
		/* endian conversion is done in decompression; no need to check that */
		m->dcmpSize+= writeSndSamples(s,a-s,&s,&m->format,
			STYPE_STDMAX16,m->format.mSampleSize==2 ? sndwrite_null
			:m->format.mSampleSize==1 ? wrChannelu8:sndwrite_memcpy); 
	}
	/* resampling phase, including mono/stereo and rate conversion */
	if (!m->mixerRtn)
	{
		m->cnvRate = ufix32muldiv(m->format.rate,m->format.rateFrac,
			m->rateMul,rflag&0xFFFFFF);
		if (m->cnvRate!=0x10000)
		{
			m->cnvRate+= !m->cnvRate; /* zero reserved for non-conversion */
			/* interpolating when too close to 1.0 causes unwanted smoothing */
			/* (this is optional) */
			/* also check the 'no interpolation' init flag - in high byte */
			if ((!(rflag&sndInitNoNearDropSample) && m->cnvRate-0xFFF8u<16u)
					|| rflag&sndInitNoInterpolation<<24)
				m->mixerRtn = rflag&sndInitStereo<<24 ?
					st ? rateCnvNoInterp2:rateCnvNoInterp1
					:st ? rateCnvNoInterp2m:rateCnvNoInterp1m;
			else
				m->mixerRtn = rflag&sndInitStereo<<24 ?
					st ? rateCnvIntrp2:rateCnvIntrp1
					:st ? rateCnvIntrp2m:rateCnvIntrp1m;
		} else
		{
			m->cnvRate = 0;
			m->mixerRtn = rflag&sndInitStereo<<24 ?
				st ? mixNoRateCnv2:mixNoRateCnv1
				:st ? mixNoRateCnv2m:mixNoRateCnv1m;
		}
	}
	/* Must clip the count, adjusting with the rate, so the last read is in bounds. */
	/* This must be done so the interpolation routines get the correct samples. */
	/* This can also be a tricky process, but using division is faster than conditional
	   checking per iteration, as used previously.
	*/
	z = (m->dcmpSize-m->pos)>>st>>1;
	if (m->cnvRate)
		z = rfix16div(z,m->cnvRate,0x1FFFF); /* required: one whole + any fraction */
	if (rem>z)
	{
		int32	dif = rem-z;

		rem = z, z = dif;
	} else
		z = 0;
	if (rem>0)
		m->mixerRtn(d,m->pos+m->dcmpBuf,m,rem);
	rem = z;
	if (m->pos>=m->dcmpSize)
		m->pos = m->dcmpSize = 0;
	else if (z)
	{ /* must do a full copy due to possible rateMultiplier changes */
		m->dcmpSize-= m->pos;
		memcpy(m->dcmpBuf,m->dcmpBuf+m->pos,m->dcmpSize);
		m->pos = 0;
	}
	return rem;
}

void __fastcall sndMixFill(int c,
	uint32 rflag,
	int32 *d,
	const sndMixerFrame *m)
{
	int32		v, u, lv, rv;
	const int16	*s;

	if (!m->dcmpBuf || m->dcmpSize<2 || c<=0)
		return;
	v = m->pos ? m->pos:m->dcmpSize-2;
	s = (int16*)(m->dcmpBuf+v);
	lv = m->LRAmp, rv = (int16)lv, lv>>= 16;
	if (rflag&sndInitStereo<<24)
		if ((m->format.chanCount&127)>1)
			v = s[-1]*lv>>8, u = *s*rv>>8;
		else
			v = *s, u = v*rv>>8, v = v*lv>>8;
	else
	{
		if ((m->format.chanCount&127)>1)
			v = monoMix(s[-1],*s,lv,rv);
		else
			v = *s*((lv+rv)>>1)>>8;
		u = v;
	}
	if (!(v|u))
		return;
	if (c&1)
		*d+= v, d[1]+= u, d+= 2;
	if ((c>>=1)==0)
		return;
	do
		*d+= v, d[1]+= u, d[2]+= v, d[3]+= u;
	while(d+=4,--c);
}

/* clipping */

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
void __fastcall s32to16(int32 c,
	uint32 rflag,
	int32 *b)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		push	edi
		test	ecx,ecx
		jle	eo

		mov	esi,[esp+12]
		mov	edi,esi
	#if sndInitStereo==0x40
		add	edx,edx
		jns	mo
	#else
		test	edx,sndInitStereo<<24
		jz	mo
	#endif
	so:	mov	eax,[esi]
		movsx	edx,ax
		cmp	eax,edx
		je	s1

		sar	eax,31
		xor	ax,0x7FFF
	s1:	mov	[edi],ax
		mov	eax,[esi+4]
		movsx	edx,ax
		cmp	eax,edx
		je	s2

		sar	eax,31
		xor	ax,0x7FFF
	s2:	mov	[edi+2],ax
		add	esi,8
		add	edi,4
		loop	so
		pop	edi
		pop	esi
		ret	4

	mo:	mov	eax,[esi]
		movsx	edx,ax
		cmp	eax,edx
		je	m1

		sar	eax,31
		xor	ax,0x7FFF
	m1:	mov	[edi],ax
		add	esi,8
		add	edi,2
		loop	mo

	eo:	pop	edi
		pop	esi
		ret	4
	}
#else
	int16		*d;
	int32		v;

	if (c<=0)
		return;
	d = (int16*)b;
	if (rflag&sndInitStereo<<24)
		do
		{
			v = *b, *d = v==(int16)v ? v:v>>31^0x7FFF;
			v = b[1], d[1] = v==(int16)v ? v:v>>31^0x7FFF;
		} while(b+=2,d+=2,--c);
	else /* already mono - just take one and skip the other */
		do
			v = *b, *d++ = v==(int16)v ? v:v>>31^0x7FFF;
		while(b+=2,--c);
#endif
}
